


#include <vga.h>
#include <vgagl.h>
#include "conf.h"
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "screendef.h"
#include "caldef.h"

#define ERRLINE 5
#define WAITMENU_LINE 4
#define MAX_PULSE_ERRORS 5



int final_filtercorr_init(void)
{
int i, j, k, mm, line, mask;
int ia,ib,ja,jb,ka,kb;
int siz_n, siz;
float t1, t2, t3;
double dt1, dt2;
double *dbuf;
double pwrinteg[17], sum[4], isum[4];
char s[80];
double dc[4];
// The correction function fft1_filtercorr is available in the same 
// number of points as we currently have for fft1.
// During the calibration process the number of points is usually
// larger than during normal operation.
// Since the Q of the hardware is limited there is no need to use
// too many points in the filter function.
// This routine will reduce the number of points used for the filtercorr
// function while removing erronous points and reducing noise.
//
// Get the transform of the current filtercorr function.
cal_initscreen();
line=0;
dbuf=(void *)(buf2);
mm=twice_rxchan;
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=fft1_filtercorr[mm*i+j  ];
    fft1_tmp[2*i+1]=fft1_filtercorr[mm*i+j+1];
    }
  fftforward(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
  for(i=0; i<fft1_size; i++)
    {
    buf1[mm*i+j  ]=fft1_tmp[2*i  ]/fft1_size;
    buf1[mm*i+j+1]=fft1_tmp[2*i+1]/fft1_size;
    }
  }  
// We now have the pulse response of the hardware filter chain, possibly
// including the approximate digital real to complex conversion.
// The pulse is centered at point 0.
// Calculate the integrated power of the pulse.
dt1=0;
for(i=0; i<fft1_size; i++)
  {
  for(j=0; j<mm; j+=2)
    {
    dt1+=pow(buf1[mm*i+j],2.)+pow(buf1[mm*i+j+1],2.0);
    }
  dbuf[i]=dt1;
  }  
dt2=0.5*(dbuf[fft1_size/2-8]+dbuf[fft1_size/2+8]);
// Add an integration constant for the integral to become zero
// in the center region, as far away as possible from the pulse.
for(i=0; i<fft1_size; i++)
  {
  dbuf[i]-=dt2;
  }
// Find out how fast the integral falls when the size is doubled.
i=4;
k=0;
xytext(10,line,"Part of pulse energy lost outside range.");
while(i<fft1_size/2)
  {
  pwrinteg[k]=(dbuf[fft1_size-i]-dbuf[i])/dt1;
  sprintf(s,"Range=%d",i);
  line++;
  xytext(12,line,s);
  sprintf(s,"Lost=%.25f",pwrinteg[k]);
  xytext(25,line,s);
  i*=2;
  k++;
  }
k--;  
// Step siz_n until the remaining energy content of the integral
// is reduced by half the initial ratio.    
siz_n=1;
t1=2*pwrinteg[1]/pwrinteg[0];
incn:;
while(siz_n < k && pwrinteg[siz_n]/pwrinteg[siz_n-1] < t1)siz_n++;    
// In case siz_n is small, allow the energy content to drop less
// at one step if it drops fast again the next one.
if(siz_n < fft1_n - 5  && siz_n < k-1)
  {
  if(pwrinteg[siz_n+1]/pwrinteg[siz_n] < t1)
    {
    siz_n++;
    goto incn;
    } 
  }
siz_n+=2;
line+=2;
siz=1<<siz_n;
sprintf(s,"New size for calibration function %d  (old = %d)",siz,fft1_size);
xytext(5,line,s);
xytext(5,line+1,"Save modified calibration function ? (Y/N)");
kbdinp:;
await_keyboard();
clean_vga_inkey();
if(vga_inkey == 'N') return 0;
if(vga_inkey != 'Y') goto kbdinp;
if(siz_n < 4)
  {
  xytext(5,line+2,"Calibration data seems incorrect");
  xytext(5,line+3,"Nothing changed");
skip:;
  xytext(5,line+5,"Press any key");
  await_keyboard();
  return 0;
  }
if( siz_n >= fft1_n)
  {
  xytext(5,line+2,"Filter response needs current number of data points");
  xytext(5,line+3,"Data will be used without change");
  goto skip;
  }
// When limiting the pulse response in the time domain, we may
// introduce discontinuities. 
// Differentiate the pulse response and do the limitation in
// the derivative.

mask=fft1_size-1;
ia=fft1_size/2;
ib=ia+1;
for(i=1; i<fft1_size; i++)
  {
  for(j=0; j<mm; j++)
    {
    buf1[ia*mm+j] = buf1[ia*mm+j] - buf1[ib*mm+j];
    }
  ia=(ia+1)&mask;
  ib=(ib+1)&mask;
  }  
for(j=0;j<mm; j++)
  {
  buf1[ia*mm+j]=0;
  sum[j]=0;
  }
// Clear the region we will not use.
ja=mm*siz/2;
jb=mm*(fft1_size-siz/2);
while(ja<jb)
  {
  buf1[ja]=0;
  ja++;
  }
// Get the DC component so we can remove it from the derivative
for(i=0; i<fft1_size; i++)
  {
  for(j=0;j<mm; j++)
    {
    sum[j]+=buf1[i*mm+j];
    }
  }
for(j=0;j<mm; j++)
  {
  sum[j]/=fft1_size;
  isum[j]=0;
  }
// Integrate to get back the pulse response.
for(i=0; i<fft1_size; i++)
  {
  for(j=0; j<mm; j++)
    {
    isum[j]+=buf1[ia*mm+j]-sum[j];
             buf1[ia*mm+j]=isum[j];
    }
  ia=(ia+mask)&mask;
  }  
// Now that we have removed the tails of the impulse response 
// which is the same as to remove narrow band components
// we take the back transform to get back the filter correction
// function.
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=buf1[mm*i+j  ];
    fft1_tmp[2*i+1]=buf1[mm*i+j+1];
    }
  fftback(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
  for(i=0; i<fft1_size; i++)
    {
    buf1[mm*i+j  ]=fft1_tmp[2*i  ];
    buf1[mm*i+j+1]=fft1_tmp[2*i+1];
    }
  }  
// Get the amplitude of the difference between the old and the new
// frequency responses, the error function.
for(i=0; i<fft1_size; i++)
  {
  t1=0;
  for(j=0; j<mm; j++)
    {
    t1+=pow(buf1[mm*i+j]-fft1_filtercorr[mm*i+j],2.0);
    }
  buf2[i]=sqrt(t1);  
  }
// Get narrow band components of the error function into buf3.
// Use forward fft, filter, backwards fft.
for(i=0; i<fft1_size; i++)
  {
  fft1_tmp[2*i  ]=buf2[i];
  fft1_tmp[2*i+1]=0;
  }
fftforward(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
k=fft1_size/2-fft1_size/siz;
for(i=0; i<k; i++)
  {
  j=fft1_size-1-i;
  fft1_tmp[2*i  ]=0;
  fft1_tmp[2*i+1]=0;
  fft1_tmp[2*j  ]=0;
  fft1_tmp[2*j+1]=0;
  }
fftback(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
for(i=0; i<fft1_size; i++)
  {
  buf3[i]=fft1_tmp[2*i  ]/fft1_size;
  }
// Locate regions where narrow band components are strong
// and use smoothed data from buf1 in those regions.
// Use the original filtercorr functions otherwise.
// Shift the buf1 data by a constant value to make the average
// of the replaced points equal to the average of the original
// points at the end points.
for(i=0; i<mm*fft1_size; i++)buf4[i]=fft1_filtercorr[i];
ja=0;
repl1:;
while(ja<fft1_size && buf2[ja] > 2*fabs(buf3[ja]))ja++;
if(ja < fft1_size)
  {
  jb=ja;
  while(jb<fft1_size && buf2[jb] <= 2*fabs(buf3[jb]))jb++;
  if(jb < fft1_size)
    {
    if(jb-ja < 3)
      {
      ja=jb+1;
      goto repl1;
      }
    ka=mm*(ja-k/4);  
    kb=mm*(jb+1+k/4);
    if(ka<0)ka=0;
    if(kb > mm*fft1_size)kb=mm*fft1_size;
    for(i=ka; i<kb; i++)      
      {
      buf4[i]=buf1[i];
      } 
    ja=jb+1;
    goto repl1;
    }
  }  
// Get the transform of the cleaned up filtercorr function.
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=buf4[mm*i+j  ];
    fft1_tmp[2*i+1]=buf4[mm*i+j+1];
    }
  fftforward(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
  for(i=0; i<fft1_size; i++)
    {
    buf1[mm*i+j  ]=fft1_tmp[2*i  ];
    buf1[mm*i+j+1]=fft1_tmp[2*i+1];
    }
  }  
// We already decided to save siz points.
// Make sure the DC component is zero within this range and clear points
// outside.
for(j=0; j<mm; j++)dc[j]=0;
k=siz/2;
ia=0;
ib=fft1_size;
while(ia < k)
  {
  ib--;
  for(j=0; j<mm; j++)
    {
    dc[j]+=buf1[ia*mm+j]+buf1[ib*mm+j];
    }
  ia++;  
  }  
for(j=0; j<mm; j++)dc[j]/=2*k;
ia=0;
ib=fft1_size;
while(ia < k)
  {
  ib--;
  for(j=0; j<mm; j++)
    {
    buf1[ia*mm+j]-=dc[j];
    buf1[ib*mm+j]-=dc[j];
    }
  ia++;  
  }  
// We have now cut the time function abruptly at +/- siz points from
// the pulse peak.
// Apply a sine wave attenuation over 1/8 of the total time span
// at each end to avoid oscillations in case there is some low
// frequency component present.
t2=8*PI_L/siz;
t1=0;
t3=sin(t1);
while(t1 <= PI_L/2)
  {
  t3=sin(t1);
  ia--;
  for(j=0; j<mm; j++)
    {
    buf1[ia*mm+j]*=t1;
    buf1[ib*mm+j]*=t1;
    }
  ib++;  
  t1+=t2;
  }  
// Store the pulse response in siz points
ja=siz/2;
jb=fft1_size-ja;
while(ja < siz)
  {
  for(j=0; j<mm;j++)
    {
    buf1[ja*mm+j]=buf1[jb*mm+j];
    }
  ja++;
  jb++;
  }  
// Store fft1_desired in siz points in buf2
k=fft1_size/siz;
j=k/2;
for(i=0; i<siz; i++)
  {
  buf2[i]=fft1_desired[j];
  j+=k;
  }
write_filcorr(siz);
sprintf(s,"Filter correction function saved in %d points",siz);
xytext(5,line+2,s);
sprintf(s,"File size reduced from %d to %d bytes",
           5*fft1_size*sizeof(float)+20*sizeof(int), 
           5*siz*sizeof(float)+20*sizeof(int));
xytext(5,line+3,s);
xytext(5,line+5,"Press any key");
await_keyboard();
return 0;
}






void calibrate_fft1(void)
{
audio_buf_info ad_info;
int j,jj,totbytes;    
char s[80];
int i,k,m,n,interval,width,domain;
int ia,ib,ic,ja,jb,ka,kb,max_pulpos;
float collect_noiselevel[MAX_ADCHAN/2];
float collect_powerlevel[MAX_ADCHAN/2];
float collect_powermax[MAX_ADCHAN/2];
float old_trig_power,trig_power,dc[2*MAX_ADCHAN],ampmax;
float t1,t2,t3,ygain,xscale,avg_power,summed_power;
int pulse_error[MAX_PULSE_ERRORS]; 
int siz128, mm, collected_pulses;
int pulpos;
int mask;
double sum[4],isum[4];
double dsum1, dsum2;
double dt1,dt2,dt3,dt4,dt5;
calibrate_flag = 1;
// **************************************************************
// Set fft1_direction positive.
// The calibration routine does not want to know if the fft1 routine
// will invert the frequency scale.
fft1_direction=1; 
// Get normal buffers with minimum for all parameters.
// Everything selectable becomes deselected.
// Save fft1 version, window and bandwidth.
ia=genparm[FIRST_FFT_VERNR];
ib=genparm[FIRST_FFT_SINPOW];
ic=genparm[FIRST_FFT_BANDWIDTH];
for(i=0; i<MAX_GENPARM; i++)genparm[i]=genparm_min[i];
// Select the correct fft version for the current rx_mode
genparm[FIRST_FFT_VERNR]=ia;
// Force fft1_size to be 4 times larger than specified for the current
// rx_mode by dividing bandwidth by 4.
t2=ic/4;
// Compensate for not using the specified window
// We use sin power 4 to suppress wideband noise that otherwise would
// be produced by discontinuities in matching between transform ends.
if(genparm_default[rx_mode][FIRST_FFT_SINPOW] != 0)
  {
  t1=pow(0.5,1.0/ib);
  t2*=(1-2*asin(t1)/PI_L);
  }
t1=pow(0.5,1.0/4);
t2/=(1-2*asin(t1)/PI_L);
genparm[FIRST_FFT_SINPOW]=4;
i=t2+0.5;  
if(i < genparm_min[FIRST_FFT_BANDWIDTH])i=genparm_min[FIRST_FFT_BANDWIDTH];
genparm[FIRST_FFT_BANDWIDTH]=i;
vga_clear();
i=5;
xytext(5,5,"Checking FFT speed");
reduce_time:;
get_buffers(0);
if(lir_status != LIR_OK)return;
i++;
sprintf(s,"Size %d (bw=%d)", fft1_size,genparm[FIRST_FFT_BANDWIDTH]);
xytext(5,i,s);
wg.first_xpoint=0;
wg.xpoints=fft1_size;
set_fft1_endpoints();
if(fft1afc_flag == 0)fft1afc_flag=-1;
// Check that the fft routine is fast enough with current parameters.
fft1_pa=0;
fft1_pb=0;
open_audio_input();
timchk:;
make_fft1(0);
if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(7);
if(ad_info.fragments > 0.9*ad_info.fragstotal)
  {
  close_audio_input();
  if(genparm[FIRST_FFT_BANDWIDTH]==0)
    {
    genparm[FIRST_FFT_BANDWIDTH]=1;
    }
  else
    {  
    genparm[FIRST_FFT_BANDWIDTH]*=2;
    }
  free_buffers();
  goto reduce_time;
  }
if( fft1_pb < 5*fft1_block)goto timchk;
close_audio_input();
cal_graph=chk_malloc((2*MAX_ADCHAN)*WIDTH*sizeof(short int),1049);
cal_table=chk_malloc(fft1_size*sizeof(COSIN_TABLE )/2,1049);
cal_permute=chk_malloc(fft1_size*sizeof(short int),1049);
cal_win=chk_malloc((1+fft1_size/2)*sizeof(float),1049);
ygain=1;
xscale=1;
domain=0;
if(ui.rx_channels == 1)
  {
  cal_ymax=.25;
  cal_yzer=.65;
  }
else
  {
  cal_ymax=.15;
  cal_yzer=.5;
  }
init_fft(0,fft1_n, fft1_size, cal_table, cal_permute);
make_window(2,fft1_size/2, 4, cal_win);
mm=twice_rxchan;
iqbeg:;  
buf4[0]=-1;
cal_initscreen();
if( (ui.input_mode&IQ_DATA) != 0)
  {
  xytext(0, 5,"Running in IQ mode (direct conversion receiver)");
  xytext(0, 6,"The mirror image spurs should be balanced before");
  xytext(0, 7,"the amplitude and phase response is calibrated");
  xytext(5, 9,"A=> Balance spurs");
  xytext(5,10,"B=> Calibrate amplitude and phase");
  xytext(5,11,"C=> Remove center discontinuity");
  xytext(5,12,"D=> Refine amplitude and phase correction");
selcal:;
  await_keyboard();
  clean_vga_inkey();
  if(vga_inkey == 'X')goto cal_skip;
  if(vga_inkey == 'A')
    {
    if(balance_spurs() == 0)goto cal_skip;
exit(0);
    goto iqbeg;
    }    
  if(vga_inkey == 'C')
    {
    if(remove_iq_notch() == 0)goto cal_skip;
    goto iqbeg;
    }
  if(vga_inkey == 'D')
    {
    if(final_filtercorr_init() == 0)goto cal_skip;
    goto iqbeg;
    }
  if(vga_inkey != 'B')goto selcal;
  }    
else
  {
  xytext(5, 9,"A=> Calibrate amplitude and phase");
  xytext(5,10,"B=> Refine amplitude and phase correction");
selcal_2:;
  await_keyboard();
  clean_vga_inkey();
  if(vga_inkey == 'X')goto cal_skip;
  if(vga_inkey == 'B')
    {
    if(final_filtercorr_init() == 0)goto cal_skip;
    goto iqbeg;
    }
  if(vga_inkey != 'A')goto selcal_2;
  }
interval=raw_calibrate_pulses();
if(interval == 0)goto cal_skip;
// **********************************************************************
// To get the pulses on one format independently of the hardware and
// fft implementation (some contain filters) we use the standard routine
// for the first fft to get fourier transforms of the input data.
// By back transformation the input data is converted to floating point
// complex format regardless of the input format.
// **********************************************************************
restart_cal:;
cal_initscreen();
summed_power=0;
for(i=0; i<mm*fft1_size-1; i++)buf5[i]=0;
for(j=0; j<ui.rx_channels; j++)
  {
  collect_noiselevel[j]=0;
  collect_powerlevel[j]=0;
  }
totbytes=0;
collected_pulses=0;
old_trig_power=0;
for(i=0; i<MAX_PULSE_ERRORS; i++)pulse_error[i]=0;
get_data:;
// ******************************************************
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = amplitude and phase (normalised) if buf4[0]>=0
// ******************************************************
xytext(0,WAITMENU_LINE,
             "Wait until curves are stable.  +,-,E,C => Change scale");
xytext(0,WAITMENU_LINE+1,"U => Update filter corrections in RAM.");
xytext(0,WAITMENU_LINE+2,"S => Save to disk");
xytext(0,WAITMENU_LINE+3,"T => Toggle time/frequency domain");
xytext(0,WAITMENU_LINE+4,"Z => Clear.");      
open_audio_input();
fft1_pa=0;
fft1_pb=0;
fft1_na=0;
fft1_nb=0;
fft1_px=0;
fft1_nx=0;
newdat_flag=0;
timf1p_pa=0;
timf1p_px=0;
j=0;
press:;
vga_inkey=vga_getkey();
if(vga_inkey != 0)
  {
  close_audio_input();
press1:;
  clean_vga_inkey();
  if(vga_inkey == 'X')goto iqbeg;
  if(vga_inkey == '+')
    {
    ygain*=2;
    }
  if(vga_inkey == '-')
    {
    ygain/=2;
    }
  if(vga_inkey == 'C')
    {
    xscale*=2;
    }
  if(vga_inkey == 'E')
    {
    xscale/=2;
    }
  if(vga_inkey == 'S')write_filcorr(0);
  if(vga_inkey == 'T')
    {
    domain^=1;
    for(j=0; j<mm; j+=2)
      {
      for(i=0; i<WIDTH; i++)
        {
errno=10200;
        ck_setpixel(i, cal_graph[WIDTH*j+i], 0);  
errno=10201;
        ck_setpixel(i, cal_graph[WIDTH*(j+1)+i], 0);  
        }
      }
    }  
  if(vga_inkey == 'U')
    {
    if(buf4[0]>=0)
      {
      if(cal_update_ram() == 0)goto cal_skip;
      vga_clear();
      }
    goto restart_cal;
    }
  if(vga_inkey == 'Z') 
    {
    clear_fft1_filtercorr();
    goto restart_cal;
    }
  goto get_data;
  }
make_fft1(0);
if( fft1_pa <3*fft1_block)goto press;
// Using a window function we have to get two buffers.
// The first buffer has partial data only.
// We are stopping the A/D after each collect cycle to avoid overruns
// so some extra time is lost here.
close_audio_input();
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=buf2[mm*i+j  ];
    fft1_tmp[2*i+1]=buf2[mm*i+j+1];
    }
  if( (ui.input_mode&IQ_DATA) != 0)fft_iqshift(fft1_size, fft1_tmp);
  fftback(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
  for(i=0; i<fft1_size; i++)
    {
    buf1[mm*i+j  ]=fft1_tmp[2*i  ];
    buf1[mm*i+j+1]=fft1_tmp[2*i+1];
    }
  }  
// The time function we got in buf1 is the time function after multiplication
// with the sin pow4 window.
// Correct the time function by a division by cal_win - but use only the
// center half so we avoid division by small numbers.
for(j=0; j<mm; j+=2)
  {
  t1=0;
  t2=0;
  for(i=fft1_size/4; i<=fft1_size/2; i++)
    {
    buf1[i*mm+j  ]/=cal_win[i];
    t1+=buf1[i*mm+j  ];
    buf1[i*mm+j+1]/=cal_win[i];
    t2+=buf1[i*mm+j+1];
    }                               
  for(i=fft1_size/2+1; i<=3*fft1_size/4; i++)
    {
    buf1[i*mm+j  ]/=cal_win[fft1_size-1-i];
    t1+=buf1[i*mm+j  ];
    buf1[i*mm+j+1]/=cal_win[fft1_size-1-i];
    t2+=buf1[i*mm+j+1];
    }
// Remove any DC component to make pulses more visible.
  t1/=fft1_size/2+1;  
  t2/=fft1_size/2+1;  
  for(i=fft1_size/4; i<=3*fft1_size/4; i++)
    {
    buf1[i*mm+j  ]-=t1;
    buf1[i*mm+j+1]-=t2;
    }
  }
// Get the total power summed over all channels and store in timf1_float    
ampmax=0;
for(i=fft1_size/4; i<=3*fft1_size/4; i++)
  {
  timf1_float[i]=pow(buf1[i*mm  ],2.0);
  for(j=1; j<mm; j++)
    {
    timf1_float[i]+=pow(buf1[i*mm+j],2.0);
    }
  if(ampmax < timf1_float[i])ampmax=timf1_float[i];
  }            
if(old_trig_power == 0)
  {
  old_trig_power=ampmax;
  }
else
  {
  old_trig_power=(old_trig_power*(INIT_PULSENUM-1)+ampmax)/INIT_PULSENUM;         
  }
// Make sure we have a reasonable value for trig power by not
// going further until we looked at a few pulses.
totbytes+=fft1_size/2;
if(totbytes < INIT_PULSENUM*interval) goto get_data;
totbytes=INIT_PULSENUM*interval;
// Set trig power at -15dB (30 times)
// Pulses must be separated well enough for the power level to
// have dropped well below -15dB after 70% of the interval time.
trig_power=old_trig_power/30;
// Look for a pulse, at least interval/2 points into our data.
pulpos=interval/2+fft1_size/4;
max_pulpos=3*fft1_size/4-interval;
while( timf1_float[pulpos] > trig_power/4 && pulpos<max_pulpos)pulpos++;
if(pulpos >= max_pulpos)
  {
  xytext(1,1,"Noise level too high");
  goto get_data;
  }
sprintf(s,"Search range %d to %d",pulpos,max_pulpos);
xytext(1,1,s);
while( timf1_float[pulpos] < trig_power && pulpos<max_pulpos)pulpos++;
find_pulse:;
if(pulpos >= max_pulpos)goto get_data;
sprintf(s,
"Accepted %4d    Wide %4d    Weak %4d   S/N %4d  Spur %4d   y-Gain %.2f   x-Gain %.2f",
                          pulse_error[0],pulse_error[1],pulse_error[2],
                          pulse_error[3],pulse_error[4],ygain,1/xscale);
xytext(0,2,s);
vga_inkey=vga_getkey();
if(vga_inkey != 0)goto press1;

ia=pulpos-interval/2;
ib=pulpos+interval/2;
if(ib >= max_pulpos)goto get_data;
ampmax=0;
for(i=ia; i<=ib; i++)
  {
  if(timf1_float[i] > ampmax)
    {
    ampmax=timf1_float[i];
    pulpos=i;
    }
  }  
// Copy data into buf2 so the pulse becomes placed at location 0
// Also copy the power function to fft1_sumsq with the pulse at location 0
k=pulpos;
pulpos+=interval;
for(i=0; i<fft1_size; i++)
  {
  for(j=0; j<mm; j++)buf2[i*mm+j]=buf1[k*mm+j];
  fft1_sumsq[i]=timf1_float[k];
  if( k<fft1_size/4 || k>3*fft1_size/4 )fft1_sumsq[i]=-1; 
  k=(k+1)&(fft1_size-1);
  }
// Find the width of the current pulse.
// First find the maximum power.
if( ampmax < old_trig_power/20)
  {
  pulse_error[2]++;
  goto find_pulse;
  }
// Now step from the peak position until we reach power/8  
ampmax/=8;
ja=0;
jb=fft1_size-1;    
while( fft1_sumsq[ja] > ampmax && ja<fft1_size)ja++;
while( fft1_sumsq[jb] > ampmax && jb>ja)jb--;
width=ja+fft1_size-jb;
if(width > interval/4)
  {
  pulse_error[1]++;
  goto find_pulse;
  }
// Filter the power function to reduce noise. Place in fft1_slowsum.
width/=2;
if(width < interval/64)width=interval/64;
if(width < 4)width=4;
ja=fft1_size-width;
jb=width;
width=2*width+1;
dsum1=0;
for(i=0;    i<=jb;     i++)dsum1+=fft1_sumsq[i];
for(i=ja; i<fft1_size; i++)dsum1+=fft1_sumsq[i];
for(i=0; i<fft1_size; i++)
  {
  jb=(jb+1)&(fft1_size-1);
  if(jb >= fft1_size)jb-=fft1_size;
  fft1_slowsum[i]=dsum1/width;
  dsum1+=fft1_sumsq[jb]-fft1_sumsq[ja];
  ja=(ja+1)&(fft1_size-1);
  } 
// Locate the surrounding pulses.
ja=0.5*interval;
jb=1.5*interval;
t1=0;
for(i=ja; i<jb; i++)
  {
  if(fft1_slowsum[i] > t1)
    {
    t1=fft1_slowsum[i];
    ib=i;
    }
  }  
ja=fft1_size-1.5*interval;
jb=fft1_size-0.5*interval;
t1=0;
for(i=ja; i<jb; i++)
  {
  if(fft1_slowsum[i] > t1)
    {
    t1=fft1_slowsum[i];
    ic=i;
    }
  }
// ib and ic point to the peaks of surrounding pulses.
// skip 25% after the previous pulse and 15% before the next one.
ia=ib/2;
ib=0.85*ib;
ic=fft1_size-(fft1_size-ic)*0.75;
// When we pick the range ic to ib we may create a discontinuity.
// Remove it by making the first derivative zero.
// Differentiate the pulse response starting at ic:
mask=fft1_size-1;
ja=ic;
jb=ic+1;
while(ja != ib)
  {
  for(j=0; j<mm; j++)
    {
    buf2[ja*mm+j] = buf2[ja*mm+j] - buf2[jb*mm+j];
    }
  ja=(ja+1)&mask;
  jb=(jb+1)&mask;
  }  
for(j=0;j<mm; j++)
  {
  buf2[ja*mm+j]=0;
  sum[j]=0;
  }
isum[0]=0;
// Get the DC component so we can remove it from the derivative
jb=ic;
while(jb != ib)
  {
  for(j=0;j<mm; j++)
    {
    sum[j]+=buf2[jb*mm+j];
    }
  jb=(jb+1)&mask;
  }
for(j=0;j<mm; j++)
  {
  sum[j]/=(ib-ic)&mask;
  isum[j]=0;
  }
// Integrate the pulse response:
while(ja != ic)
  {
  for(j=0; j<mm; j++)
    {
    isum[j]+=buf2[ja*mm+j]-sum[j];
    buf2[ja*mm+j]=isum[j];
    }
  ja=(ja+mask)&mask;
  }  
// Attenuate before and after the pulse with a cos squared window.
k=fft1_size-ic;
if(ib < k)k=ib;
k/=2;
ja=ic+k;
jb=ib-k;
t1=0;
for(i=0; i<k; i++)
  {
  t2=pow(cos(t1),2.0);
  for(j=0; j<mm; j++)
    {
    buf2[ja*mm+j]*=t2;
    buf2[jb*mm+j]*=t2;
    }
  t1+=0.5*PI_L/k;
  ja--;
  jb++;
  }
for(i=jb; i<=ja; i++)
  {
  for(j=0; j<mm; j++)
    {
    buf2[i*mm+j]=0;
    }
  }
if(domain == 1)
  {
  for(j=0; j<mm; j+=2)
    {
    for(i=0; i<WIDTH; i++)
      {
      k=i-WIDTH/3;
      k*=xscale;
      k&=(fft1_size-1);
errno=10202;
      ck_setpixel(i, cal_graph[WIDTH*j+i], 0);  
      t2=0.00001*ygain*buf2[mm*k+j];   
      if(t2 <-cal_ymax)t2=-cal_ymax;
      if(t2 >cal_ymax)t2=cal_ymax;
      if(j > 1)t2-=0.32;
      cal_graph[WIDTH*j+i]=HEIGHT*(cal_yzer-t2);
errno=10203;
      ck_setpixel(i, cal_graph[WIDTH*j+i], 13);  
errno=10204;
      ck_setpixel(i, cal_graph[WIDTH*(j+1)+i], 0);  
      t2=0.00001*ygain*buf2[mm*k+j+1];   
      if(t2 <-cal_ymax)t2=-cal_ymax;
      if(t2 >cal_ymax)t2=cal_ymax;
      if(j > 1)t2-=0.32;
      cal_graph[WIDTH*(j+1)+i]=HEIGHT*(cal_yzer-t2);
errno=10205;
      ck_setpixel(i, cal_graph[WIDTH*(j+1)+i], 10);  
      }
    }
  }
for(j=0; j<ui.rx_channels; j++)collect_powermax[j]=0;  
// Now that we have a single pulse (hopefully), get the
// fourier transform of it.
siz128=fft1_size/128;
for(j=0; j<mm; j+=2)
  {
  jj=j/2;
  for(i=0; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=buf2[mm*i+j  ];
    fft1_tmp[2*i+1]=buf2[mm*i+j+1];
    }
  fftforward(fft1_size, fft1_n, fft1_tmp, cal_table, cal_permute);
  if( (ui.input_mode&IQ_DATA) != 0)fft_iqshift(fft1_size, fft1_tmp);
  for(i=0; i<fft1_size; i++)
    {
    buf2[mm*i+j  ]=fft1_tmp[2*i  ];
    buf2[mm*i+j+1]=fft1_tmp[2*i+1];
    }
// The complex transform must not contain high frequencies because the
// filter function must vary slowly with frequency.
// Store a low pass filtered version in buf3 as complex numbers
// with the filtered powers in buf6
  dsum1=dsum2=0;
  n=2*siz128;
  for(i=2; i<n+2; i++)
    {
    dsum1+=fft1_tmp[2*i];
    dsum2+=fft1_tmp[2*i+1];
    }
  for(i=0; i<siz128; i++)
    {
    buf3[mm*i+j  ]=dsum1/n;
    buf3[mm*i+j+1]=dsum2/n;
    buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
    }
  ja=fft1_size-siz128-1;  
  if( (ui.input_mode&IQ_DATA) == 0)
    {
    for(i=siz128; i<ja; i++)
      {
      buf3[mm*i+j  ]=dsum1/n;
      buf3[mm*i+j+1]=dsum2/n;
      buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
      if(buf6[mm*i+j  ]>collect_powermax[jj])
            collect_powermax[jj]=buf6[mm*i+j  ];
      dsum1+=fft1_tmp[2*(i+siz128)]-fft1_tmp[2*(i-siz128)];
      dsum2+=fft1_tmp[2*(i+siz128)+1]-fft1_tmp[2*(i-siz128)+1];
      }
    }
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = accumulated power and phase (normalised)
//buf2 = current pulse: complex amplitude
//buf3 = current pulse: smoothed complex amplitude
//buf6 = current pulse: smoothed power
// ******************************************************
// In direct conversion mode there may be a discontinuity at fft1_size/2
// because the A/D converters are AC coupled.      
  else
    {
    jb=fft1_size/2-siz128-1;
    for(i=siz128; i<jb; i++)
      {
      buf3[mm*i+j  ]=dsum1/n;
      buf3[mm*i+j+1]=dsum2/n;
      buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
      if(buf6[mm*i+j  ]>collect_powermax[jj])
            collect_powermax[jj]=buf6[mm*i+j  ];
      dsum1+=fft1_tmp[2*(i+siz128)]-fft1_tmp[2*(i-siz128)];
      dsum2+=fft1_tmp[2*(i+siz128)+1]-fft1_tmp[2*(i-siz128)+1];
      }
    for(i=jb; i<fft1_size/2; i++)
      {
      buf3[mm*i+j  ]=dsum1/n;
      buf3[mm*i+j+1]=dsum2/n;
      buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
      }
    dsum1=dsum2=0;
    jb=fft1_size/2+2*siz128+1;
    for(i=fft1_size/2+1; i<jb; i++)
      {
      dsum1+=fft1_tmp[2*i];
      dsum2+=fft1_tmp[2*i+1];
      }
    jb=fft1_size/2+siz128+1;
    for(i=fft1_size/2+1; i<jb; i++)
      {
      buf3[mm*i+j  ]=dsum1/n;
      buf3[mm*i+j+1]=dsum2/n;
      buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
      }
    for(i=jb; i<ja; i++)
      {
      buf3[mm*i+j  ]=dsum1/n;
      buf3[mm*i+j+1]=dsum2/n;
      buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
      if(buf6[mm*i+j  ]>collect_powermax[jj])
            collect_powermax[jj]=buf6[mm*i+j  ];
      dsum1+=fft1_tmp[2*(i+siz128)]-fft1_tmp[2*(i-siz128)];
      dsum2+=fft1_tmp[2*(i+siz128)+1]-fft1_tmp[2*(i-siz128)+1];
      }
    buf3[ui.rx_channels*fft1_size+j  ]=
                          (buf3[ui.rx_channels*(fft1_size+2)+j  ]+
                           buf3[ui.rx_channels*(fft1_size-2)+j  ])/2;
    buf3[ui.rx_channels*fft1_size+j+1]=
                          (buf3[ui.rx_channels*(fft1_size+2)+j+1]+
                           buf3[ui.rx_channels*(fft1_size-2)+j+1])/2;
    buf6[ui.rx_channels*fft1_size+j]=
                          (buf6[ui.rx_channels*(fft1_size+2)+j]+
                           buf6[ui.rx_channels*(fft1_size-2)+j])/2;
    }
  for(i=ja; i<fft1_size; i++)
    {
    buf3[mm*i+j  ]=dsum1/n;
    buf3[mm*i+j+1]=dsum2/n;
    buf6[mm*i+j  ]=pow(dsum1/n,2.0)+pow(dsum2/n,2.0);
    }
  }
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = accumulated power and phase (normalised)
//buf2 = current pulse: complex amplitude
//buf3 = current pulse: smoothed complex amplitude
//buf6 = current pulse: smoothed power
// ******************************************************
// Collect noise weighted by power spectrum 
for(j=0; j<mm; j+=2)
  {
  dc[j]=dc[j+1]=0;
  for(i=siz128; i<fft1_size/2-siz128; i++)
    {
    t1=buf6[mm*i+j  ];
    t2=pow(buf2[mm*i+j  ]-buf3[mm*i+j  ],2.0)+
       pow(buf2[mm*i+j+1]-buf3[mm*i+j+1],2.0);
    dc[j]+=t1*t2;
    dc[j+1]+=t1;
    }
  for(i=fft1_size/2+siz128; i<fft1_size-siz128; i++)
    {
    t1=buf6[mm*i+j  ];
    t2=pow(buf2[mm*i+j  ]-buf3[mm*i+j  ],2.0)+
       pow(buf2[mm*i+j+1]-buf3[mm*i+j+1],2.0);
    dc[j]+=t1*t2;
    dc[j+1]+=t1;
    }
  dc[j]/=dc[j+1];
  }
for(j=0; j<ui.rx_channels; j++)
  {
  t1=-10*log10(dc[2*j]/collect_powermax[j]);
  sprintf(s,"(%.1f)",t1);
  xytext(8+20*(j+ui.rx_channels),3,s);   
  }
if(collected_pulses < START_PULSES)
  {
  collected_pulses++;
  for(j=0; j<ui.rx_channels; j++)
    {
    collect_noiselevel[j]+=dc[2*j];
    collect_powerlevel[j]+=collect_powermax[j];
    t1=-10*log10(collect_noiselevel[j]/collect_powerlevel[j]);
    sprintf(s,"S/N  %.1fdB",t1);
    xytext(20*j,3,s);   
    }
  if(collected_pulses < START_PULSES)goto get_data;
  if(t1 < STON_LIMIT)
    {
    clear_lines(3,13);
    sprintf(s,"Pulse S/N not good enough (%.1f) Limit %ddB",t1,STON_LIMIT);
    xytext(0,ERRLINE,s);
    xytext(0,ERRLINE+1,"Amplify pulses, reduce noise or increase PRF");
    xytext(0,ERRLINE+2,"Reason may also be old bad correction function");
    xytext(5,ERRLINE+4,"I => Ignore (=try anyway)");
    xytext(5,ERRLINE+5,"Z => Clear old correction function.");
errkbd:;
    await_keyboard();
    clean_vga_inkey();
    clear_lines(ERRLINE,ERRLINE+5);
    if(vga_inkey=='X')goto cal_skip;
    if(vga_inkey == 'Z')
      {
      clear_fft1_filtercorr();
      goto restart_cal;
      }
    if(vga_inkey !='I')goto errkbd;
      
    }
  for(j=0; j<ui.rx_channels; j++)
    {
    collect_noiselevel[j]/=START_PULSES;
    collect_powerlevel[j]/=START_PULSES;
    }
  }
for(j=0; j<ui.rx_channels; j++)
  {
  t1=collect_noiselevel[j]/collect_powerlevel[j];
  t2=dc[2*j]/collect_powermax[j];
  if(t2<0.25*t1)
    {
    pulse_error[3]++;
    goto find_pulse;
    }
  if(1.5*collect_powermax[j]<collect_powerlevel[j])
    {
    pulse_error[2]++;
    goto find_pulse;
    }
  }
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = accumulated power and phase (normalised)
//buf2 = current pulse: complex amplitude
//buf3 = current pulse: smoothed complex amplitude
//buf6 = current pulse: smoothed power
// ******************************************************
// This pulse is probably ok.
// Convert the complex data in buf2 to power and phase.
// Store noise as square of amplitude differences between raw and filtered data.
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    buf6[mm*i+j+1]=
          pow(buf2[mm*i+j  ]-buf3[mm*i+j  ],2.)+
          pow(buf2[mm*i+j+1]-buf3[mm*i+j+1],2.);
    t1=pow(buf2[mm*i+j  ],2.0)+
       pow(buf2[mm*i+j+1],2.0);
    if(t1 > 0.01)
      {
      buf2[mm*i+j+1]=atan2(buf2[mm*i+j+1],buf2[mm*i+j  ]);
      }
    else  
      {
      buf2[mm*i+j+1]=0;
      }
    buf2[mm*i+j  ]=t1;
    }
  } 
// Get the first derivative of the phase function
for(i=0; i<fft1_size-1; i++)
  {
  for(j=0; j<mm; j+=2)
    {
    buf2[mm*i+j+1]=buf2[mm*(i+1)+j+1]-buf2[mm*i+j+1];
    while(buf2[mm*i+j+1] >  PI_L)buf2[mm*i+j+1] -= 2*PI_L;
    while(buf2[mm*i+j+1] < -PI_L)buf2[mm*i+j+1] += 2*PI_L;
    }
  }  
// Get the second derivative of the phase function
for(i=0; i<fft1_size-2; i++)
  {
  for(j=0; j<mm; j+=2)
    {
    buf2[mm*i+j+1]=buf2[mm*(i+1)+j+1]-buf2[mm*i+j+1];
    while(buf2[mm*i+j+1] >  PI_L)buf2[mm*i+j+1] -= 2*PI_L;
    while(buf2[mm*i+j+1] < -PI_L)buf2[mm*i+j+1] += 2*PI_L;
    }
  }
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = accumulated power and phase (normalised)
//buf2 = current pulse: power and d2phase/df2
//buf3 = current pulse: smoothed complex amplitude
//buf6 = current pulse: smoothed power and noise power
// ******************************************************
// Step through the data and look for points with excessive noise.
// Make sure the integral is smaller than PI over the bad interval.
for(j=0; j<mm; j+=2)
  {
  dc[j]=0;
  t1=0;
  t2=0;
  for(i=siz128; i<fft1_size-siz128; i++)
    {
    t3=buf6[mm*i+j  ];
// In case the filter slope is large we made an error neglecting the
// curvature of the power function.
// Correct by subtracting the second derivative error from the noise.
// This may be a too large correction, make sure the result is not negative.
    buf6[mm*i+j+1]-=fabs(2*t3-buf6[mm*(i+siz128)+j  ]-
                              buf6[mm*(i-siz128)+j  ])/32;    
    if(buf6[mm*i+j+1] < 0)buf6[mm*i+j+1]=0;
    dc[j]+=t3*buf6[mm*i+j+1];
    t2+=t3;
    if(t1 < t3 ) t1=t3;
    }
  dc[j]*=10./t2;
  dc[j+1]=0.1*t1;
  }
n=0;
ja=siz128;
spur_remove:;
k=0;
while(ja < fft1_size-siz128  && k==0)
  {
  ja++;
  for(j=0; j<mm; j+=2)
    {
    if(dc[j]<buf6[mm*ja+j+1])k=1;
    if(fabs(buf2[mm*ja+j+1]-buf2[mm*(ja+1)+j+1])>0.75*PI_L)k=1;
    if(fabs(buf2[mm*ja+j+1]-buf2[mm*(ja-1)+j+1])>0.75*PI_L)k=1;
    }
  }
// ja is first bad point  
jb=ja;
while(jb < fft1_size-siz128  && k==1)
  {
  k=0;
  jb++;
  for(j=0; j<mm; j+=2)if(dc[j]<buf6[mm*jb+j+1])k=1;
  if(fabs(buf2[mm*jb+j+1]-buf2[mm*(jb+1)+j+1])>0.75*PI_L)k=1;
  if(fabs(buf2[mm*jb+j+1]-buf2[mm*(jb-1)+j+1])>0.75*PI_L)k=1;
  }
// jb-1 is last bad point.
// Check that we are within the passband.
k=0;
for(j=0; j<mm; j+=2)
  {
  if(buf6[mm*ja+j] > dc[j+1] || buf6[mm*jb+j] > dc[j+1])k=1;
  }
if(k==1)
  {
  n+=jb-ja;
  if(jb-ja-1>fft1_size/128 || n > fft1_size/32)
    {
spurerr:;
    pulse_error[4]++;
    goto find_pulse;
    }
  }  
ja-=fft1_size/256;
if(ja<0)ja=0;
jb+=fft1_size/256;
if(jb>=fft1_size)jb=fft1_size-1;
for(j=0; j<mm; j+=2)
  {
  m=0;  
  ka=kb=0;
rem_tupi:;
  t1=0;
  t2=BIG;
  t3=-BIG;
// Calculate the integral across the region ja to jb
  for(i=ja; i<jb; i++)
    {
    t1+=buf2[mm*i+j+1];
    if(buf2[mm*i+j+1]<t2)
      {
      t2=buf2[mm*i+j+1];
      ka=i;
      }
    if(buf2[mm*i+j+1]>t3)
      {
      t3=buf2[mm*i+j+1];
      kb=i;
      }
    }
  m++;
// If abs of the integral is larger than PI, change the largest point. 
  if(m < 5)
    {
    if(t1 < -PI_L)
      {
      buf2[mm*ka+j+1]+=2*PI_L;
      goto rem_tupi;
      }
    if(t1 > PI_L)
      {
      buf2[mm*kb+j+1]-=2*PI_L;
      goto rem_tupi;
      }
    }
  if(fabs(t1) > PI_L/4 && buf6[mm*ja+j] > dc[j+1] && buf6[mm*jb+j] > dc[j+1])
    {
    goto spurerr;
    }
  }
ja=jb-fft1_size/256+1;
if(ja < fft1_size-1-fft1_size/256)goto spur_remove;
// *******************************************************************
//The pulse seems ok.
//Add the new power spectrum to the powers already collected in buf5.
avg_power=0;
for(i=0; i<fft1_size; i++)
  {
  for(j=0; j<mm; j+=2)
    {
    buf5[mm*i+j]+=buf2[mm*i+j];
    avg_power+=buf2[mm*i+j];
    }
  }
avg_power/=fft1_size;
summed_power+=avg_power;
// Add the new d2phase weighted by power into buf5
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size-2; i++)
    {
    buf5[mm*i+j+1]+=buf2[mm*i+j+1]*avg_power;
    }
  }
pulse_error[0]++;
if(pulse_error[0] == 0)goto get_data;
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
// ******************************************************
// Place the normalised average power and d2phase in buf4 
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    buf4[mm*i+j  ]=buf5[mm*i+j  ]/summed_power;
    buf4[mm*i+j+1]=buf5[mm*i+j+1]/summed_power;
    }
  }
// Integrate d2phase so we get d1phase (backwards so the sign becomes wrong).
for(j=1; j<mm; j+=2)
  {
  buf4[mm*(fft1_size-2)+j]=0;
  for(i=fft1_size-2; i>0; i--)buf4[mm*(i-1)+j]+=buf4[mm*i+j];
  }
// Remove the integration constant.
// Choose it for (d1phase weighted by power) = 0.
// Exclude points near the spectrum ends. 
k=fft1_size/16;
if( (ui.input_mode&IQ_DATA) == 0)
  {
  for(j=0; j<mm; j+=2)
    {
    dt1=0;
    dt2=0;
    for(i=k; i<fft1_size-k; i++)
      {
      dt3=sqrt(buf4[mm*i+j]*buf4[mm*(i+1)+j]);
      dt1+=dt3*buf4[mm*i+j+1];
      dt2+=dt3;
      }
    dt1/=dt2;  
    for(i=0; i<fft1_size-1; i++) buf4[mm*i+j+1]-=dt1;
    }
  }
else
  {  
// for complex input we may have a discontinuity at the center
  m=fft1_size/2-k;
  for(j=0; j<mm; j+=2)
    {
    dt1=0;
    dt2=0;
    dt3=0;
    dt4=0;
    n=fft1_size-k;
    for(i=k; i<m; i++)
      {
      n--;
      dt5=sqrt(buf4[mm*i+j]*buf4[mm*(i+1)+j]);
      dt1+=dt5*buf4[mm*i+j+1];
      dt2+=dt5;
      dt5=sqrt(buf4[mm*n+j]*buf4[mm*(n+1)+j]);
      dt3+=dt5*buf4[mm*n+j+1];
      dt4+=dt5;
      }
    dt1/=dt2;
    dt3/=dt4;
    for(i=0; i<fft1_size/2; i++) buf4[mm*i+j+1]-=dt1;
    for(i=fft1_size/2; i<fft1_size-1; i++) buf4[mm*i+j+1]-=dt3;
    }
  }
// ******************************************************
//Valid spectra:
//buf5 = accumulated power and d2phase/df2 (not normalised)
//buf4 = accumulated power and dphase/df (normalised)
// ******************************************************
// Integrate d1phase so we get the phase. 
// Backwards so the sign becomes right aygain.
  for(j=1; j<mm; j+=2)
    {
    buf4[mm*(fft1_size-1)+j]=0;
    for(i=fft1_size-1; i>0; i--)
      {
      buf4[mm*(i-1)+j]+=buf4[mm*i+j];
      }
    }
// Remove the integration constant.
// Choose it for (phase weighted by power) = 0.
if( (ui.input_mode&IQ_DATA) == 0)
  {
  for(j=0; j<mm; j+=2)
    {
    dt1=0;
    dt2=0;
    for(i=1; i<fft1_size-1; i++)
      {
      dt3=buf4[mm*i+j];
      dt1+=dt3*buf4[mm*i+j+1];
      dt2+=dt3;
      }
    dt1/=dt2; 
    for(i=0; i<fft1_size; i++) buf4[mm*i+j+1]-=dt1;
    }
  }
else
// In direct conversion mode there is a discontinuity at the
// center frequency because the A/D converter is not DC coupled
// Treat the two halfs on either side independently
  {
  m=fft1_size/16;
  for(j=0; j<mm; j+=2)
    {
    dt1=0;
    dt2=0;
    dt3=0;
    dt4=0;
    for(i=m; i<fft1_size/2-m; i++)
      {
      k=fft1_size-i;
      dt5=buf4[mm*i+j];
      dt1+=dt5*buf4[mm*i+j+1];
      dt2+=dt5;
      dt5=buf4[mm*k+j];
      dt3+=dt5*buf4[mm*k+j+1];
      dt4+=dt5;
      }
    dt1/=dt2;
    dt3/=dt4;  
    for(i=0; i<fft1_size/2; i++)
      {
      k=fft1_size-1-i;
      buf4[mm*i+j+1]-=dt1;
      buf4[mm*k+j+1]-=dt3;
      }
    }
  }
// Limit phase angle to +/- pi
// Filters with large phase shifts will be difficult to see 
// on screen otherwise.
for(j=0; j<mm; j+=2)
  {
  for(i=0; i<fft1_size; i++)
    {
    t1=buf4[mm*i+j+1]/(2*PI_L);
    t1-=(int)(t1);
    if(t1 < -0.5)t1+=1;
    if(t1 >  0.5)t1-=1;
    buf4[mm*i+j+1]=t1*2*PI_L;
    }
  }
if(domain == 0)
  {
  for(j=0; j<mm; j+=2)
    {
    for(i=0; i<WIDTH; i++)
      {
      k=i*fft1_size/WIDTH;
      k*=xscale;
      while(k>fft1_size)k-=fft1_size;
      while(k<0)k+=fft1_size;
errno=10206;
      ck_setpixel(i, cal_graph[WIDTH*j+i], 0);  
      t2=0.1*ygain*buf4[mm*k+j];   
      if(t2 <-cal_ymax)t2=-cal_ymax;
      if(t2 >cal_ymax)t2=cal_ymax;
      if(j > 1)t2-=0.32;
      cal_graph[WIDTH*j+i]=HEIGHT*(cal_yzer-t2);
errno=10207;
      ck_setpixel(i, cal_graph[WIDTH*j+i], 13);  
errno=10208;
      ck_setpixel(i, cal_graph[WIDTH*(j+1)+i], 0);  
      t2=0.03*ygain*buf4[mm*k+j+1];   
      if(t2 <-cal_ymax)t2=-cal_ymax;
      if(t2 >cal_ymax)t2=cal_ymax;
      if(j > 1)t2-=0.32;
      cal_graph[WIDTH*(j+1)+i]=HEIGHT*(0.5-t2);
errno=10209;
      ck_setpixel(i, cal_graph[WIDTH*(j+1)+i], 10);  
      }
    }
  }
goto find_pulse;
cal_skip:;
free(cal_graph);
fflush(NULL);
free(cal_table);
free(cal_permute);
free(cal_win);
close_all_except_vga();
}
