
#include <ctype.h>
#include <vga.h>
#include <vgagl.h>
#include <unistd.h>
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "caldef.h"

#define INFOLINE 3
#define ERRLINE 5

void cal_initscreen(void)
{
int i;
char s[80];
vga_clear();
sprintf(s,"Calibration routine for %s",rxmodes[rx_mode]);
for(i=0; i<WIDTH*(2*MAX_ADCHAN); i++)cal_graph[i]=HEIGHT-1;
settextcolor(14);
xytext(10,0,s);
settextcolor(7);
}

void normalise_fft1_filtercorr(float *out)
{
int i, j, k;
int mm;
float t1, t2, sum[2];
mm=2*ui.rx_channels;
for(i=0; i<mm*fft1_size; i++)out[i]=fft1_filtercorr[i];
//Make the average gain equal to one (weighted by desired spectrum)
t1=0;
sum[0]=0;
sum[1]=0;
for(i=0; i<fft1_size; i++)
  {
  t2=pow(fft1_desired[i],2.0);
  t1+=t2;
  k=0;
  for(j=0; j<mm; j+=2)
    {
    sum[k]+=t2*(pow(out[mm*i+j  ],2.0)+
                pow(out[mm*i+j+1],2.0));
    k++;
    }
  }  
sum[0]=sqrt(sum[0]/t1);
sum[1]=sqrt(sum[1]/t1);
for(i=0; i<fft1_size; i++)
  {
  k=0;
  for(j=0; j<mm; j+=2)
    {
    out[mm*i+j  ]/=sum[k];
    out[mm*i+j+1]/=sum[k];
    k++;
    }
  }
}

void write_filcorr(int type)
{
FILE *file;
int i;
int wrbuf[20];
char s[80];
make_filfunc_filename(s);
file = fopen(s, "wb");
if (file == NULL)end_program(7689);
wrbuf[0]=fft1_n;
wrbuf[1]=fft1_size;
wrbuf[2]=rx_mode;
wrbuf[3]=ui.input_mode;
wrbuf[4]=genparm[FIRST_FFT_VERNR];
wrbuf[5]=ui.input_speed;
wrbuf[6]=ui.rx_channels;
wrbuf[7]=type;
for(i=8; i<10; i++)wrbuf[i]=0;
fwrite(wrbuf, sizeof(int),10,file);
if(type == 0)
  {
  normalise_fft1_filtercorr(fft1_tmp);
  fwrite(fft1_tmp, 2*ui.rx_channels*sizeof(float),fft1_size, file);  
  fwrite(fft1_desired,sizeof(float),fft1_size,file);
  }
else
  {
  fwrite(buf1, 2*ui.rx_channels*sizeof(float),type, file);  
  fwrite(buf2,sizeof(float),type,file);
  }  
fwrite(wrbuf, sizeof(int),10,file);
fclose(file);
}


int raw_calibrate_pulses(void)
{
int i, j, k, l, m, n, color,wid,first_pulse;
int ia, ib, ic, ja, jb, ka, kb, skipflag;
int imode, rdp, rawbytes, lim, overflow_flag;
int maxval[MAX_ADCHAN],minval[MAX_ADCHAN];
int old_end;
int points; 
int textend,ysiz,ytop;
char s[80];
float *filpwr;
float old_trig_power,trig_power,dc[MAX_ADCHAN],ampmax;
float signal_level,interval; 
float t1, t2;
float xgain, ygain, noise_floor;
skipflag=0;
signal_level=0;  
cal_initscreen();
xytext(3, 1,"Connect pulse generator to all antenna inputs");
xytext(3, 2,"Press ENTER when OK.    (+),(-) for y-scale     E,C for x-scale");
xytext(0,INFOLINE,"DC offset");
xytext(5,INFOLINE+1,"Min");
xytext(5,INFOLINE+2,"Max"); 
xytext(3,INFOLINE+3,"Level"); 
// Collect 0.5 seconds (or fft1_buffer full) of raw data. 
// Open and close A/D device, screen is slow and we want no overruns
rawbytes=0.5*ui.input_speed*bytes_per_fragment;
if(rawbytes > fft1_bytes)rawbytes=fft1_bytes;
rawbytes-=ad_read_bytes;
points=rawbytes/bytes_per_fragment;
if((unsigned int)(points) > timf1_bytes/(2*sizeof(float)))
  {
  points=timf1_bytes/(2*sizeof(float));
  rawbytes=points*bytes_per_fragment;
  }  
filpwr=&timf1_float[timf1_bytes/(2*sizeof(float))];
old_trig_power=1;
old_end=-1;
ygain=1;
xgain=1;
interval=0;
cal_loop:;
if(skipflag == 1)goto finish;
open_audio_input();
rdp=0;
while(rdp < rawbytes)
  {
  read(audio_in, &fft1_char[rdp], ad_read_bytes);
  rdp+=ad_read_bytes;
  vga_inkey=vga_getkey();
  if(vga_inkey != 0)
    {
    clean_vga_inkey();
    if(vga_inkey == 'X' || vga_inkey == 10)
      {
      while(rdp < rawbytes)
        {
        fft1_char[rdp]=0;
        rdp++;
        }
      skipflag=1;  
      goto rdx;
      }
    if(vga_inkey == '+')ygain*=2;
    if(vga_inkey == '-')ygain/=2;
    if(vga_inkey == 'E')xgain*=2;
    if(vga_inkey == 'C')xgain/=2;
    }
  }
rdx:;  
close_audio_input();
// Get power as a function of time.
// The A/D converters are AC coupled so we remove DC components.
imode=ui.input_mode&(DWORD_INPUT+TWO_CHANNELS+IQ_DATA);
for(i=0; i<MAX_ADCHAN; i++)
  {
  maxval[i]=0x80000000; 
  minval[i]=0x7fffffff;
  dc[i]=0;
  }
switch ( imode )
  {
// One channel 16 bit real data.  
  case 0:
  for(i=0; i<points; i++)
    {
    dc[0]+=fft1_short_int[i];
    if(minval[0]>fft1_short_int[i])minval[0]=fft1_short_int[i];
    if(maxval[0]<fft1_short_int[i])maxval[0]=fft1_short_int[i];
    }
  dc[0]/=points;
  for(i=0; i<points; i++)
    {
    timf1_float[i]=pow( 65536.0*(fft1_short_int[i]-dc[0]), 2.);
    }
  break;

// One channel 32 bit real data.  
  case 1:
  for(i=0; i<points; i++)
    {
    dc[0]+=fft1_int[i];
    if(minval[0]>fft1_int[i])minval[0]=fft1_int[i];
    if(maxval[0]<fft1_int[i])maxval[0]=fft1_int[i];
    }
  dc[0]/=points;
  for(i=0; i<points; i++)
    {
    timf1_float[i]=pow( fft1_int[i]-dc[0], 2.);
    }
  break;

// Two channels 16 bit real data  
  case 2:
  for(i=0; i<points; i++)
    {
    for(j=0; j<2; j++)
      {
      dc[j]+=fft1_short_int[2*i+j];
      if(minval[j]>fft1_short_int[2*i+j])minval[j]=fft1_short_int[2*i+j];
      if(maxval[j]<fft1_short_int[2*i+j])maxval[j]=fft1_short_int[2*i+j];
      }
    }
  for(j=0; j<2; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<2; j++)
      {
      t2=fft1_short_int[2*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1*65536.0*65536.0;
    }
  break;
  
// Two channels 32 bit real data  
  case 3:
  for(i=0; i<points; i++)
    {
    for(j=0; j<2; j++)
      {
      dc[j]+=fft1_int[2*i+j];
      if(minval[j]>fft1_int[2*i+j])minval[j]=fft1_int[2*i+j];
      if(maxval[j]<fft1_int[2*i+j])maxval[j]=fft1_int[2*i+j];
      }
    }
  for(j=0; j<2; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<2; j++)
      {
      t2=fft1_int[2*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1;
    }
  break;
  

// One channel 16 bit IQ data  
  case 4:
  for(i=0; i<points; i++)
    {
    for(j=0; j<2; j++)
      {
      dc[j]+=fft1_short_int[2*i+j];
      if(minval[j]>fft1_short_int[2*i+j])minval[j]=fft1_short_int[2*i+j];
      if(maxval[j]<fft1_short_int[2*i+j])maxval[j]=fft1_short_int[2*i+j];
      }
    }
  for(j=0; j<2; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<2; j++)
      {
      t2=fft1_short_int[2*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1*65536.0*65536.0;
    }
  break;

// One channel 32 bit IQ data  
  case 5:
  for(i=0; i<points; i++)
    {
    for(j=0; j<2; j++)
      {
      dc[j]+=fft1_int[2*i+j];
      if(minval[j]>fft1_int[2*i+j])minval[j]=fft1_int[2*i+j];
      if(maxval[j]<fft1_int[2*i+j])maxval[j]=fft1_int[2*i+j];
      }
    }
  for(j=0; j<2; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<2; j++)
      {
      t2=fft1_int[2*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1;
    }
  break;

// Two channels 16 bit IQ data  
  case 6:
  for(i=0; i<points; i++)
    {
    for(j=0; j<4; j++)
      {
      dc[j]+=fft1_short_int[4*i+j];
      if(minval[j]>fft1_short_int[4*i+j])minval[j]=fft1_short_int[4*i+j];
      if(maxval[j]<fft1_short_int[4*i+j])maxval[j]=fft1_short_int[4*i+j];
      }
    }
  for(j=0; j<4; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<4; j++)
      {
      t2=fft1_short_int[4*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1*65536.0*65536.0;
    }
  break;
  
// Two channels 32 bit IQ data  
  case 7:
  for(i=0; i<points; i++)
    {
    for(j=0; j<4; j++)
      {
      dc[j]+=fft1_int[4*i+j];
      if(minval[j]>fft1_int[4*i+j])minval[j]=fft1_int[4*i+j];
      if(maxval[j]<fft1_int[4*i+j])maxval[j]=fft1_int[4*i+j];
      }
    }
  for(j=0; j<4; j++)dc[j]/=points;
  for(i=0; i<points; i++)
    {
    t1=0;
    for(j=0; j<4; j++)
      {
      t2=fft1_int[4*i+j]-dc[j];
      t1+=t2*t2;
      }
    timf1_float[i]=t1;
    }
  break;
  }
signal_level=BIG;  
if( (ui.input_mode&(DWORD_INPUT)) == 0)lim=0x7ff0; else lim=0x7ff00000; 
for(j=0; j<ui.ad_channels; j++)
  {
  sprintf(s," %4.2f ",dc[j]);
  xytext(10+15*j,INFOLINE,s);
  sprintf(s," %d ",minval[j]);
  xytext(10+15*j,INFOLINE+1,s);
  sprintf(s," %d ",maxval[j]);
  xytext(10+15*j,INFOLINE+2,s);
  t1=-minval[j];
  if(t1<maxval[j])t1=maxval[j];
  overflow_flag=0;    
  if(t1 < signal_level)signal_level=t1; 
  if(minval[j] <= -lim || maxval[j] >= lim)
    {
    overflow_flag=1;    
    xytext(10+15*j,INFOLINE+3,"OVERFLOW");
    t1=1;
    }
  else
    {
    t1/=lim;
    sprintf(s," %1.6f%% ",100*t1);
    xytext(10+15*j,INFOLINE+3,s);
    }
  }
signal_level/=lim;  
// Find the largest power value and point wid to it.
trig_power=0;
wid=0;
for(i=0; i<points; i++)
  {
  if(trig_power < timf1_float[i])
    {
    trig_power=timf1_float[i];
    wid=i;
    }
  }
// Get an approximate width of the pulse (remember it may have a
// very ugly shape!)  
i=wid;  
while( i<points-3 && (timf1_float[i] >= trig_power/20 ||
                      timf1_float[i+1] >= trig_power/20 ||
                      timf1_float[i+2] >= trig_power/20 ||
                      timf1_float[i+3] >= trig_power/20 ))i++;
while( wid>2 && (timf1_float[wid] >= trig_power/20 ||
               timf1_float[wid-1] >= trig_power/20 ||
               timf1_float[wid-2] >= trig_power/20 ||
               timf1_float[wid-3] >= trig_power/20 ))wid--;
wid=(i-wid);
wid=wid/2+1;
// Low pass filter the power function and store in filpwr
for(i=wid; i<points; i++)
  {
  t1=0;
  for(j=i-wid; j<=i; j++)
    {
    t1+=timf1_float[j];
    }
  filpwr[i-wid/2]=t1/wid;  
  }
// Save the maximum amplitude and make trig power 31.6 times smaller
// than the maximum value (-15dB)
ampmax=sqrt(trig_power);
if(trig_power > old_trig_power)old_trig_power=trig_power; 
trig_power=old_trig_power/31.6;
// Remember trig_power for the next set of data.
old_trig_power/=2;
ia=0;
i=wid;
if(wid > points/16)goto cal_loop;
// find the position of the first pulse
interval_init:;
while( i<points-3*wid && filpwr[i] <= trig_power)i++;
ja=i;
while( i<points-wid && filpwr[i] >= trig_power/20)i++;
if(ja==wid)goto interval_init;
jb=i;
t1=0;
for(k=ja; k<jb; k++)
  {
  if(timf1_float[k] > t1)
    {
    t1=timf1_float[k];
    ia=k;
    }
  }
first_pulse=ia;  
// Find sucessive pulses and store the intervals in fft1_tmp 
k=0;
interval=0;
j=0;
find_interval:;
while( i<points-wid && filpwr[i] <= trig_power)i++;
ja=i;
while( i<points-wid && filpwr[i] >= trig_power/20)i++;
jb=i;
t1=0;
ib=i;
for(k=ja; k<jb; k++)
  {
  if(timf1_float[k] > t1)
    {
    t1=timf1_float[k];
    ib=k;
    }
  }
if(i<points-wid)
  {
  if(ib-ia<4*wid)goto find_interval;
  fft1_tmp[j]=ib-ia;
  j++;
  ia=ib;
  if(j<fft1_size)goto find_interval;
  }
// Since the pulses may vary in amplitude by more than we allow for in
// the setting of the trigger level we sort the intervals in ascending order
// and take desicions if the interval is 1,2 or 3 pulses.  
if(j==0)goto cal_loop;
for(i=0; i<j-1; i++)
  {
  for(k=i+1; k<j; k++)
    {
    if(fft1_tmp[k]<fft1_tmp[i])
      {     
      t1=fft1_tmp[k];
      fft1_tmp[k]=fft1_tmp[i];
      fft1_tmp[i]=t1;
      }
    }
  }
i=j/4+1;  
t1=0;
while(i<j && fft1_tmp[i-1]>0.7*fft1_tmp[i])
  {
  t1+=fft1_tmp[i];
  i++;
  }
i-=(j/4+1);
if(i != 0)
  {
  t1/=i;
  }
else
  {
  goto cal_loop;
  }  
k=0;
for(i=0; i<=j; i++)
  {
  if(fft1_tmp[i]>0.8*t1 && fft1_tmp[i]<1.2*t1)
    {
    interval+=fft1_tmp[i];
    k++;
    }
  else
    {
    if(fft1_tmp[i]>1.8*t1 && fft1_tmp[i]<2.2*t1)
      {
      interval+=fft1_tmp[i];
      k+=2;
      }
    else
      {
      if(fft1_tmp[i]>2.7*t1 && fft1_tmp[i]<3.3*t1)
        {
        interval+=fft1_tmp[i];
        k+=3;
        }
      }      
    }
  }
interval/=k;
// Get the noise floor.
// First locate the smallest power value.
t1=BIG;
for(i=0; i<points; i++)
  {
  if(t1>fft1_tmp[i])t1=fft1_tmp[i];
  }
// Assume the noise floor is 10dB above this level
t1*=10;
if(t1 < 0.00001*trig_power)t1=0.00001*trig_power;
if(t1 > 0.1*trig_power)t1=0.1*trig_power;
// Count the number of points above the noise floor and
// get the average value of the points below the noise floor.
j=0;
k=0;
get_floor:;
t2=0;
k=0;
j++;
if(j < 10)
  {
  for(i=wid; i<points-wid; i++)
    {
    if(filpwr[i]<t1)
      {
      k++;
      t2+=timf1_float[i];
      }
    }
  if(k < 0.4*(points-2*wid))
    {
    t1*=10;
    goto get_floor;
    }
  t2/=k;
  if(k > 0.6*(points-2*wid))
    {
    if(t1 > 2*t2) 
      {
      t1=2*t2;
      }
    else
      {
      t1/=2;
      }
    goto get_floor;
    }
  }
else
  {
  t2=0.1*trig_power;
  }          
noise_floor=3*sqrt(t2);
// Show the pulses and get an estimate of the noise floor.
while(first_pulse > interval)first_pulse -= interval; 
if(first_pulse < interval/4+1)first_pulse += interval;
ka=first_pulse-interval/2;
kb=first_pulse+interval/2;
k=0;
find_start:;
if(kb >= points)
  {
show_result:
  sprintf(s,"Scale x=%f  y=%f      S/N=%f dB      PRF %.1f Hz    ",
           xgain,ygain,20*log10(ampmax/noise_floor),ui.input_speed/interval);
  xytext(1,INFOLINE+4,s);
  goto cal_loop;
  }
k++;
if(k > 10)goto show_result;
t1=0;
ib=ka;
for(i=ka; i<=kb; i++)
  {
  if(timf1_float[i] > t1)
    {
    t1=timf1_float[i];
    ib=i;
    }
  }
ic=ib+7*WIDTH/8;
if(ic >= points)goto show_result;
ia=ib-WIDTH/8;
if(ia<0)
  {
  ka+=interval;
  kb+=interval;
  goto find_start;
  }
ka=ib+interval/2;
kb=ib+3*interval/2;  
for(j=0; j<ui.ad_channels+1; j++)
  {
  m=WIDTH*j;
  n=WIDTH*(ui.ad_channels+1);
  for(i=1; i<=old_end; i++)
    {
errno=10300;
    ck_line(cal_graph[n+i-1], cal_graph[m+i-1],
            cal_graph[n+i  ], cal_graph[m+i  ],0);
    }
  }
textend=(INFOLINE+4)*text_height;
ytop=textend+HEIGHT*0.3;
ysiz=HEIGHT-ytop;
n=-1;  
for(j=ia; j<ic; j++)
  {
  i=WIDTH/8+xgain*(j-ib);
  if(i>=0 && i<WIDTH)
    {
    n++;
    old_end=n;
    cal_graph[WIDTH*(ui.ad_channels+1)+n]=i;
    cal_graph[n]=textend+HEIGHT*(0.4-0.008*log(timf1_float[j]));
    switch ( imode )
      {
// One channel 16 bit  
      case 0:
      t2=4*6553.6*ygain*fft1_short_int[j]/ampmax;
      if(t2 <-0.45)t2=-0.45;
      if(t2 >0.45)t2=0.45;
      cal_graph[WIDTH+n]=ytop+ysiz*(0.5-t2);
      m=1;
      break;

// One channel 32 bit  
      case 1:
      t2=0.4*ygain*fft1_int[j]/ampmax;
      if(t2 <-0.45)t2=-0.45;
      if(t2 >0.45)t2=0.45;
      cal_graph[WIDTH+n]=ytop+ysiz*(0.5-t2);
      m=1;
      break;

// Two channels 16 bit real data  
      case 2:
      for(m=0; m<2; m++)
        {
        t2=2*6553.6*ygain*fft1_short_int[2*j+m]/ampmax;
        if(t2 <-0.22)t2=-0.22;
        if(t2 >0.22)t2=0.22;
        if(m == 1)t2-=0.5;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.25-t2);
        }
      m=2;  
      break;

// Two channels 32 bit real data  
      case 3:
      for(m=0; m<2; m++)
        {
        t2=0.4*ygain*fft1_int[2*j+m]/ampmax;
        if(t2 <-0.22)t2=-0.22;
        if(t2 >0.22)t2=0.22;
        if(m == 1)t2-=0.5;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.25-t2);
        }
      m=2;  
      break;
  
// One channel 16 bit IQ data  
      case 4:
      for(m=0; m<2; m++)
        {
        t2=4*6553.6*ygain*fft1_short_int[2*j+m]/ampmax;
        if(t2 <-0.45)t2=-0.45;
        if(t2 >0.45)t2=0.45;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.5-t2);
        }
      m=2;  
      break;

// One channel 32 bit IQ data  
      case 5:
      for(m=0; m<2; m++)
        {
        t2=0.4*ygain*fft1_int[2*j+m]/ampmax;
        if(t2 <-0.45)t2=-0.45;
        if(t2 >0.45)t2=0.45;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.5-t2);
        }
      m=2;  
      break;

// Two channels 16 bit IQ data  
      case 6:
      for(m=0; m<4; m++)
        {
        t2=4*6553.6*ygain*fft1_short_int[4*j+m]/ampmax;
        if(t2 <-0.22)t2=-0.22;
        if(t2 >0.22)t2=0.22;
        if(m > 1)t2-=0.5;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.25-t2);
        }
      m=4;
      break;

// Two channels 32 bit IQ data  
      case 7:
      for(m=0; m<4; m++)
        {
        t2=0.4*ygain*fft1_int[4*j+m]/ampmax;
        if(t2 <-0.22)t2=-0.22;
        if(t2 >0.22)t2=0.22;
        if(m > 1)t2-=0.5;
        cal_graph[WIDTH*(m+1)+n]=ytop+ysiz*(0.25-t2);
        }
      m=4;
      break;

      default:
      m=0;
      }
    if(n>0)
      {
errno=10301;
      ck_line(cal_graph[WIDTH*(ui.ad_channels+1)+n-1],
                    cal_graph[n-1],
                  i,cal_graph[n],15);
      for(l=0; l<m; l++)
        {
        color=10+3*(l&1);
        if( (ui.input_mode&IQ_DATA) == 0)color=14;
errno=10302;
        ck_line(cal_graph[WIDTH*(ui.ad_channels+1)+n-1],
                    cal_graph[(l+1)*WIDTH+n-1],
                  i,cal_graph[(l+1)*WIDTH+n],color);
        }
      }
    }
  }
if(skipflag == 0)vga_inkey=vga_getkey();
if(vga_inkey != 0)
  {
  clean_vga_inkey();
  if(vga_inkey == 'X')
    {
    i=0;
    goto rawcal_x;
    }
  if(vga_inkey == '+')ygain*=2;
  if(vga_inkey == '-')ygain/=2;
  if(vga_inkey == 'E')xgain*=2;
  if(vga_inkey == 'C')xgain/=2;

  if(vga_inkey == 10)
    {
finish:;    
    vga_clear();
    if( (ui.input_mode&IQ_DATA) == 0)interval/=2;
    if(interval > fft1_size/5)
      {
      xytext(5,ERRLINE,"Pulse repetition frequency too low.");
      xytext(5,ERRLINE+1,"Increase PRF or set fft1 bandwidth lower to");
      xytext(5,ERRLINE+2,"increase transform size.");
err1:;
      xytext(5,ERRLINE+5,"Press any key");
      await_keyboard();
      i=0;
      goto rawcal_x;
      }  
    if(signal_level > 0.9)
      {
      xytext(5,ERRLINE,"Pulses too strong (above 90%)");
      goto err1;
      }
    i=interval;
    goto rawcal_x;
    }
  }
goto find_start;
rawcal_x:;
return i;
}

