#include <errno.h>



#define NO_REDRAW_BLOCKS 16
// *******************************************
// Global numerical values
#define PI_L 3.1415926535897932
#define BIG 300000000000000000000000000000000000000.
#define ARROW_UP_KEY 256
#define ARROW_DOWN_KEY 257
#define ARROW_RIGHT_KEY 258
#define ARROW_LEFT_KEY 259
#define HOME_KEY 260
#define INSERT_KEY 261
#define DELETE_KEY 262
#define END_KEY 263
#define PAGE_UP_KEY 264
#define PAGE_DOWN_KEY 265
#define PAUSE_KEY 266
#define F1_KEY 267
#define F2_KEY 268
#define F3_KEY 269
#define F4_KEY 270
#define F5_KEY 271
#define F6_KEY 272
#define F7_KEY 273
#define F8_KEY 274
#define F9_KEY 275
#define F10_KEY 276
#define F10_PAD_KEY 277
#define F11_KEY 278
#define F12_KEY 279
#define SHIFT_F1_KEY 280
#define SHIFT_F2_KEY 281
#define SHIFT_F2_PAD_KEY 282
#define SHIFT_F3_KEY 283
#define SHIFT_F4_KEY 284
#define SHIFT_F5_KEY 285
#define SHIFT_F6_KEY 286
#define SHIFT_F7_KEY 287
#define SHIFT_F8_KEY 288
// values for lir_status
// Memory errors are larger than LIR_OK.
// Other abnormal conditions smaller than LIR_OK.
#define LIR_OK 0
#define LIR_TUNEERR 1
#define LIR_FFT1ERR 2
#define LIR_SPEEDERR 3
#define LIR_MEMERR 4
#define LIR_NEW_SETTINGS -1
#define LIR_SPURERR -2
#define LIR_PARERR -3
#define LIR_NEW_POL -4
// *******************************************
// Fixed array dimensions depend on these defines:

#define MODE_WCW 0
#define MODE_NCW 1
#define MODE_HSMS 2
#define MODE_SSB 3
#define MODE_FM 4
#define MODE_AM 5
#define MODE_QRSS 6
#define MODE_TXTEST 7
#define MODE_ADTEST 8
#define MODE_TUNE 9
#define MAX_RX_MODE 10

#define MAX_COLOR_SCALE 23
#define MAX_SCRO 50
#define LLSQ_MAXPAR 10
#define MAX_ADCHAN 4
#define MAX_MIX1 8           // No of different signals to process simultaneously
#define MAX_AD_CHANNELS 4
#define MAX_RX_CHANNELS 2
#define DAOUT_MAXTIME 1.
#define SPUR_N 3
#define SPUR_SIZE (1<<SPUR_N)
#define NO_OF_SPUR_SPECTRA 256
#define MAX_FFT1_ARRAYS 70
#define MAX_FFT3_ARRAYS 30
#define MAX_BASEB_ARRAYS 40
#define CG_MAXTRACE 24
#define MAX_HIRES_ARRAYS  8
#define MAX_AFC_ARRAYS 20
#define CWLONG_MAXBUF 4

// *******************************************
// Main menu line numberings.
#define SETUP_LINE 14   
#define LINE_SEP 2
// Always include stdio.h, math.h and stdlib.h
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

typedef void (*ROUTINE) (void);
#define PERMDEB fprintf(dmp,
// *******************************************
// Audio board parameter definitions
#define DWORD_INPUT 1
#define TWO_CHANNELS 2
#define IQ_DATA 4
#define BYTE_INPUT 8
#define NO_DUPLEX 16
#define MODEPARM_MAX 32
// *******************************************
// Processing parameter definitions
#define FIRST_FFT_BANDWIDTH 0
#define FIRST_FFT_SINPOW 1
#define FIRST_FFT_VERNR 2
#define FFT1_STORAGE_TIME 3 
#define FIRST_FFT_GAIN 4 
#define SECOND_FFT_ENABLE 5

#define FIRST_BCKFFT_VERNR 6
#define SELLIM_MAXLEVEL 7
#define FIRST_BCKFFT_ATT_N 8
#define SECOND_FFT_NINC 9
#define SECOND_FFT_SINPOW 10
#define SECOND_FFT_VERNR 11
#define SECOND_FFT_ATT_N 12
#define FFT2_STORAGE_TIME 13

#define AFC_ENABLE 14
#define AFC_LOCK_RANGE 15
#define AFC_MAX_DRIFT 16
#define CW_DECODE_ENABLE 17
#define MAX_NO_OF_SPURS 18
#define SPUR_TIMECONSTANT 19

#define MIX1_BANDWIDTH_REDUCTION_N 20
#define MIX1_NO_OF_CHANNELS 21
#define BASEBAND_STORAGE_TIME 22
#define OUTPUT_DELAY_MARGIN 23
#define DA_OUTPUT_SPEED 24

#define OUTPUT_MODE 25

#define AMPLITUDE_EXPAND_EXPONENT 26
#define MAX_GENPARM 27

// *****************************************
// Bit's in newdat_flag and fullbuf_flag.
// These bits in newdat_flag tell whether there is unprocessed data pending.
// In fullbuf_flag these bits indicate that the buffer is full.
#define ND_HWARE 1
#define ND_TIMF1 2
#define ND_FFT1 4
#define ND_TIMF2RAW 8
#define ND_TIMF2BLN 16
#define ND_FFT2 32
#define ND_TIMF3 128
#define ND_FFT3 256
#define ND_BASEB 512
#define ND_DASIG 1024
#define ND_DABUF 2048
#define ND_WG 4096
#define ND_BG 8192
#define ND_WGX 16384


// *******************************************************
// Defines for state machine in fft2.c
#define FFT2_NOT_ACTIVE 0
#define FFT2_B 10
#define FFT2_C 11
#define FFT2_MMXB 20
#define FFT2_MMXC 21
#define FFT2_PWRSUM 100
#define FFT2_PWRSUM_MMX 101
#define FFT2_LIMINFO 102

#define FFT2_COMPUTE_GRAPH 103
#define FFT2_SHOW_SPECTRUM 250
#define FFT2_ELIMINATE_SPURS 104
#define FFT2_ELIMINATE_SPURS_MMX 105
#define FFT2_WATERFALL_LINE 1001
#define FFT2_TEST_FINISH 9999

// *******************************************
// Structure for user interface parameters selected by operator
typedef struct {
int vga_mode;
int font_scale;
int mouse_speed;
int input_mode;
int rx_channels;
int ad_channels;
int input_speed;
int ad_device_no;
int ad_device_mode;
int ad_frag;
int da_device_no;
int min_output_speed;
int max_output_speed;
int max_output_channels;
int max_output_bytes;
int min_output_channels;
int min_output_bytes;
int da_stopstart;
int memlock;
int sleep;
int check;
}USERINT_PARM;
#define MAX_UIPARM 21
extern char *uiparm_text[MAX_UIPARM];

// *******************************************
// Screen object parameters for mouse and screen drawing.
// Each screen object is defined by it's number, scro[].no
// For each type scro[].type the number scro[].no is unique and
// defines what to do in the event of a mouse click.

// Definitions for type = GRAPH
#define GRAPH 0
#define WIDE_GRAPH 0
#define HIRES_GRAPH 1
#define AFC_GRAPH 2
#define BASEBAND_GRAPH 3
#define POL_GRAPH 4
#define COH_GRAPH 5
#define EME_GRAPH 6
#define FREQ_GRAPH 7

// Definitions for parameter input routines under mouse control
#define FIXED_INT_PARM 3
#define TEXT_PARM 4
#define FIXED_FLOAT_PARM 5
#define MAX_TEXTPAR_CHARS 32
#define EG_DX_CHARS 12
#define EG_LOC_CHARS 6 

// Definitions for type WIDE_GRAPH
#define WG_TOP 0
#define WG_BORDER 1
#define WG_BOTTOM 2
#define WG_LEFT 3
#define WG_RIGHT 4
#define WG_YSCALE_EXPAND 5
#define WG_YSCALE_CONTRACT 6
#define WG_YZERO_DECREASE 7
#define WG_YZERO_INCREASE 8
#define WG_FQMIN_DECREASE 9
#define WG_FQMIN_INCREASE 10
#define WG_FQMAX_DECREASE 11
#define WG_FQMAX_INCREASE 12
#define WG_AVG1NUM 13
#define WG_FFT1_AVGNUM 14
#define WG_WATERF_AVGNUM 15
#define WG_WATERF_ZERO 16
#define WG_WATERF_GAIN 17
#define MAX_WGBUTT 18

// Definitions for type HIRES_GRAPH
#define HG_TOP 0
#define HG_BOTTOM 1
#define HG_LEFT 2
#define HG_RIGHT 3
#define HG_BLN_STUPID 4
#define HG_BLN_CLEVER 5
#define HG_TIMF2_STATUS 6
#define HG_TIMF2_WK_INC 7
#define HG_TIMF2_WK_DEC 8
#define HG_TIMF2_ST_INC 9
#define HG_TIMF2_ST_DEC 10
#define HG_TIMF2_LINES 11
#define HG_TIMF2_HOLD 12
#define HG_FFT2_AVGNUM 13
#define MAX_HGBUTT 14

// Definitions for type BASEBAND_GRAPH
#define BG_TOP 0
#define BG_BOTTOM 1
#define BG_LEFT 2
#define BG_RIGHT 3
#define BG_YSCALE_EXPAND 4
#define BG_YSCALE_CONTRACT 5
#define BG_YZERO_DECREASE 6
#define BG_YZERO_INCREASE 7
#define BG_RESOLUTION_DECREASE 8
#define BG_RESOLUTION_INCREASE 9
#define BG_OSCILLOSCOPE 10
#define BG_OSC_INCREASE 11
#define BG_OSC_DECREASE 12
#define BG_PIX_PER_PNT_INC 13
#define BG_PIX_PER_PNT_DEC 14
#define BG_TOGGLE_EXPANDER 15
#define BG_TOGGLE_COHERENT 16
#define BG_TOGGLE_PHASING 17
#define BG_TOGGLE_CHANNELS 18
#define BG_TOGGLE_BYTES 19
#define BG_TOGGLE_TWOPOL 20
#define BG_SEL_COHFAC 21
#define BG_SEL_DELPNTS 22
#define BG_SEL_FFT3AVGNUM 23
#define BG_TOGGLE_AGC 24
#define BG_SEL_AGC_ATTACK 25
#define BG_SEL_AGC_RELEASE 26
#define BG_YBORDER 27
#define BG_WATERF_ZERO 28
#define BG_WATERF_GAIN 29
#define BG_WATERF_AVGNUM 30
#define BG_AFC2_RANGE 31
#define BG_AFC2_DELAY 32
#define MAX_BGBUTT 33

// Definitions for type AFC_GRAPH
#define AG_TOP 0
#define AG_BOTTOM 1
#define AG_LEFT 2
#define AG_RIGHT 3
#define AG_FQSCALE_EXPAND 4
#define AG_FQSCALE_CONTRACT 5
#define AG_MANAUTO 6
#define AG_WINTOGGLE 7
#define AG_SEL_AVGNUM 8
#define AG_SEL_FIT 9
#define AG_SEL_DELAY 10
#define MAX_AGBUTT 11

// Definitions for type APOL_GRAPH
#define PG_TOP 0
#define PG_BOTTOM 1
#define PG_LEFT 2
#define PG_RIGHT 3
#define PG_ANGLE 4
#define PG_CIRC 5
#define PG_AUTO 6
#define PG_AVGNUM 7
#define MAX_PGBUTT 8

// Definitions for type ACOH_GRAPH
#define CG_TOP 0
#define CG_BOTTOM 1
#define CG_LEFT 2
#define CG_RIGHT 3
#define CG_OSCILLOSCOPE 4
#define MAX_CGBUTT 5

// Definitions for type AEME_GRAPH
#define EG_TOP 0
#define EG_BOTTOM 1
#define EG_LEFT 2
#define EG_RIGHT 3
#define EG_MINIMISE 4
#define EG_LOC 5
#define EG_DX 6
#define MAX_EGBUTT 7

// Definitions for type FREQ_GRAPH
#define FG_TOP 0
#define FG_BOTTOM 1
#define FG_LEFT 2
#define FG_RIGHT 3
#define FG_INCREASE_FQ 4
#define FG_DECREASE_FQ 5
#define FG_INCREASE_GAIN 6
#define FG_DECREASE_GAIN 7
#define MAX_FGBUTT 8




// Structure for the eme database.
#define CALLSIGN_CHARS 12
typedef struct{
char call[CALLSIGN_CHARS];
float lon;
float lat;
}DXDATA;

// Structure for mouse buttons
typedef struct {
int x1;
int x2;
int y1;
int y2;
}BUTTONS;

// Structure to remember screen positions
// for processing parameter change boxes
typedef struct{
int no;
int type;
int x;
int y;
}SAVPARM;

// Structure for a screen object on which the mouse can operate.
typedef struct {
int no;
int x1;
int x2;
int y1;
int y2;
}SCREEN_OBJECT;

// Structure for the AFC graph
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int mode_control;
int avgnum;
int window;
int fit_points;
int delay;
int check;
float minston;
float search_range;
float lock_range;
float frange;
} AG_PARMS;
#define MAX_AG_INTPAR 10
extern char *ag_intpar_text[MAX_AG_INTPAR];
#define MAX_AG_FLOATPAR 4
extern char *ag_floatpar_text[MAX_AG_FLOATPAR];

// Structure for the wide graph.
// Waterfall and full dynamic range spectrum
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int yborder;
int xpoints_per_pixel;
int pixels_per_xpoint;
int first_xpoint;
int xpoints;
int fft_avg1num;
int spek_avgnum;
int waterfall_avgnum;
int check;
float yzero;
float yrange;
float waterfall_db_zero;
float waterfall_db_gain;
} WG_PARMS;
#define MAX_WG_INTPAR 13
extern char *wg_intpar_text[MAX_WG_INTPAR];
#define MAX_WG_FLOATPAR 4
extern char *wg_floatpar_text[MAX_WG_FLOATPAR];

// Structure for the high resolution and blanker control graph.
// Waterfall and full dynamic range spectrum
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int stupid_bln_mode;
int clever_bln_mode;
int timf2_display;
int timf2_display_lines;
int timf2_display_hold;
int spek_avgnum;
unsigned int stupid_bln_limit;
unsigned int clever_bln_limit;
int check;
float stupid_bln_factor;
float clever_bln_factor;
float blanker_ston_fft1;
float blanker_ston_fft2;
float timf2_display_wk_gain;
float timf2_display_st_gain;
} HG_PARMS;
#define MAX_HG_INTPAR 13
extern char *hg_intpar_text[MAX_HG_INTPAR];
#define MAX_HG_FLOATPAR 6
extern char *hg_floatpar_text[MAX_HG_FLOATPAR];


// Structure for the baseband graph
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int yborder;
int fft_avgnum;
int pixels_per_point;
int coh_factor;
int delay_points;
int agc_flag;
int agc_attack;
int agc_release;
int waterfall_avgnum;
int afc2_range;
int afc2_delay;
int check;
float filter_flat;
float filter_curv;
float yzero;
float yrange;
float db_per_pixel;
float yfac_power;
float yfac_log;
float bandwidth;
float first_frequency;
float bfo_freq;
float output_gain;
float waterfall_gain;
float waterfall_zero;
} BG_PARMS;
#define MAX_BG_INTPAR 16
extern char *bg_intpar_text[MAX_BG_INTPAR];
#define MAX_BG_FLOATPAR 13
extern char *bg_floatpar_text[MAX_BG_FLOATPAR];


typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int adapt;
int avg;
int check;
float angle;
float c1;
float c2;
float c3;
} PG_PARMS;
#define MAX_PG_INTPAR 7
extern char *pg_intpar_text[MAX_PG_INTPAR];
#define MAX_PG_FLOATPAR 4
extern char *pg_floatpar_text[MAX_PG_FLOATPAR];


typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
} CG_PARMS;
#define MAX_CG_INTPAR 4
extern char *cg_intpar_text[MAX_CG_INTPAR];
#define MAX_CG_FLOATPAR 0
extern char *cg_floatpar_text[MAX_CG_FLOATPAR];


typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int minimise;
} EG_PARMS;
#define MAX_EG_INTPAR 5
extern char *eg_intpar_text[MAX_EG_INTPAR];
#define MAX_EG_FLOATPAR 0
extern char *eg_floatpar_text[MAX_EG_FLOATPAR];


typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int passband_direction;
int gain;
int gain_increment;
double passband_increment;
double passband_center;
} FG_PARMS;
#define MAX_FG_INTPAR 7
extern char *fg_intpar_text[MAX_FG_INTPAR];
#define MAX_FG_FLOATPAR 2
extern char *fg_floatpar_text[MAX_FG_FLOATPAR];


// ***********************************************
// Structure used by float fft routines
typedef struct {
float sin;
float cos;
}COSIN_TABLE;

// Structure used by MMX fft routines
typedef struct {
short int c1p;
short int s2p;
short int c3p;
short int s4m;
}MMX_COSIN_TABLE;

/*
// Structure for spectral power density when no of rx channels = 2
typedef struct {
float x2;
float y2;
float re_xy;
float im_xy;
}TWOCHAN_POWERS;
*/

// Define setup info for each fft version.
typedef struct {
unsigned char window;
unsigned char permute;
unsigned char max_n;
unsigned char approx_iq;
unsigned char mmx;
unsigned char simd;
char *text;
} FFT_SETUP_INFO;


typedef struct {
void *pointer;
int size;
int scratch_size;
int num;
} MEM_INF;

typedef struct {
float x2;
float y2;
float im_xy;
float re_xy;
}TWOCHAN_POWER_FLOAT;

typedef struct {
int x2;
int y2;
int im_xy;
int re_xy;
}TWOCHAN_POWER_INT;


# define MAX_FFT_VERSIONS 18
# define MAX_FFT1_VERNR 8
# define MAX_FFT1_BCKVERNR 4
# define MAX_FFT2_VERNR 4
