

#include <vga.h>
#include <vgagl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/io.h>
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "hwaredef.h"
#include "seldef.h"
#include "sigdef.h"
#include "screendef.h"
#include "vernr.h"
#include "conf.h"

#define DISKSAVE_X_SIZE 48*text_width
#define DISKSAVE_Y_SIZE 6*text_height    
#define DISKSAVE_SCREEN_SIZE DISKSAVE_X_SIZE*DISKSAVE_Y_SIZE
#define DEBMEM 256

void write_wav_header(int totbytes);
fpos_t wav_wrpos;
int sel_parm;
int sel_line;

#define MAX_FILES 256
#define MAX_NAMLEN 36

FILE *open_parfile(int type, char *mode)
{
char s[80];
int i,j;
// Find out if there is a file with parameters for the main spectrum graph.
// Make the default graph if no file is found, else use old data.
i=0;
if(savefile_parname[0]!=0)
  {
  while(savefile_parname[i]!=0)
    {
    s[i]=savefile_parname[i];
    i++;
    }
  }  
else
  {  
  while(rxpar_filenames[rx_mode][i]!=0)
    {
    s[i]=rxpar_filenames[rx_mode][i];
    i++;
    }
  }
s[i  ]='_';
i++;
j=0;
while(graphtype_names[type][j] != 0)
  {
  s[i]=graphtype_names[type][j];
  i++;
  j++;
  }
s[i]=0;
return fopen(s, mode);
}


int read_modepar_file(int type)
{
int i, j, k, max_intpar, max_floatpar;
int *wgi;
float *wgf;
double *wgd;
char *parinfo;
FILE *wgfile;
wgfile = open_parfile(type,"r");
if (wgfile == NULL)return 0;
parinfo=chk_malloc(4096,1082);
i=fread(parinfo,1,4095,wgfile);
fclose(wgfile);
if(i >= 4095)goto txt_err;
wgi=(void*)(graphtype_parptr[type]);
k=0;
max_intpar = graphtype_max_intpar[type];
max_floatpar = graphtype_max_floatpar[type];
for(i=0; i<max_intpar; i++)
  {
  while( (parinfo[k]==' ' || parinfo[k]== '\n' ) && k<4095)k++;
  j=0;
  while(parinfo[k]== graphtype_partexts_int[type][i][j] && k<4095)
    {
    k++;
    j++;
    } 
  if(graphtype_partexts_int[type][i][j] != 0)
    {
txt_err:;    
    free(parinfo);
    return 0;
    }
  while(parinfo[k]!='[' && k<4095)k++;
  if(k>=4095)goto txt_err;
  sscanf(&parinfo[k],"[%d]",&wgi[i]);
  while(parinfo[k]!='\n' && k<4095)k++;
  if(k>=4095)goto txt_err;
  }
if(max_floatpar < 0)
  {
  wgd=(void*)(&wgi[max_intpar]);
  for(i=0; i<-max_floatpar; i++)
    {
    while(parinfo[k]==' ' ||
          parinfo[k]== '\n' )k++;
    j=0;
    while(parinfo[k]== graphtype_partexts_float[type][i][j]&&k<4095)
      {
      k++;
      j++;
      } 
    if(graphtype_partexts_float[type][i][j] != 0)goto txt_err;
    while(parinfo[k]!='[' && k<4095)k++;
    if(k>=4095)goto txt_err;
    sscanf(&parinfo[k],"[%lf]",&wgd[i]);
    if(k>=4095)goto txt_err;
    while(parinfo[k]!='\n' && k<4095)k++;
    }
  }
else
  {  
  wgf=(void*)(&wgi[max_intpar]);
  for(i=0; i<max_floatpar; i++)
    {
    while(parinfo[k]==' ' || parinfo[k]== '\n' )k++;
    j=0;
    while(parinfo[k]== graphtype_partexts_float[type][i][j])
      {
      k++;
      j++;
      } 
    if(graphtype_partexts_float[type][i][j] != 0)goto txt_err;
    while(parinfo[k]!='[')k++;
    sscanf(&parinfo[k],"[%f]",&wgf[i]);
    while(parinfo[k]!='\n')k++;
    }
  }
free(parinfo);
return 1;
}


void make_modepar_file(int type)
{
int i, max_intpar, max_floatpar;
int *wgi;
float *wgf;
double *wgd;
FILE *file;
file = open_parfile(type,"w");
if (file == NULL)end_program(1164);
wgi=(void*)(graphtype_parptr[type]);
max_intpar = graphtype_max_intpar[type];
max_floatpar = graphtype_max_floatpar[type];
for(i=0; i<max_intpar; i++)
  {
  fprintf(file,"%s [%d]\n",graphtype_partexts_int[type][i],wgi[i]);
  }
if(max_floatpar < 0)
  {
  wgd=(void*)(&wgi[max_intpar]);
  for(i=0; i<-max_floatpar; i++)
    {
    fprintf(file,"%s [%.30f]\n",graphtype_partexts_float[type][i],wgd[i]);
    }
  }
else
  {  
  wgf=(void*)(&wgi[max_intpar]);
  for(i=0; i<max_floatpar; i++)
    {
    fprintf(file,"%s [%.15f]\n",graphtype_partexts_float[type][i],wgf[i]);
    }
  }  
parfile_end(file);
}

int check_keyboard_actions(void)
{
int i;
vga_inkey=vga_getkey();
if(vga_inkey != 0)
  {
  clean_vga_inkey();
  if(numinput_flag != 0)
    {
    get_numinput_chars();
    }
  else
    {  
    switch (vga_inkey)
      {
      case 'A':
      ampinfo_flag^=1;
      timinfo_flag=0;
      amp_info_texts();
      break;

      if(genparm[AFC_ENABLE] != 0)
        {
        case 'E':
        init_spur_elimination();
        break;
      
        case 'C':
        cancel_spur_elimination();
        break;
        }
        
      case 'F':
      flowinfo_flag^=1;
      break;

      case 'Q':
      userdefined_q();
      break;

      case 'S':
      if(disksave_flag < 2)disksave_start_stop();
      break;

      case 'T':
      timinfo_flag^=1;
      ampinfo_flag=0;
      timing_info_texts();
      break;

      case 'U':
      userdefined_u();
      break;

      case 'W':
      wavsave_start_stop();
      break;      

      case 'X':
      return -1;
  
      case 'Z':
      clear_workload_timers();
      break;
    
      case PAUSE_KEY:
      i=wg.xright/text_width-26;
      if(i<2)i=2;
      xytext(i,wg.yborder/text_height+1,"PROVOKING OVERRUN ERRORS");
      xytext(i,wg.yborder/text_height+2,"Press any key to continue");
      await_keyboard();
   errno=40115;
      ck_fillbox(i-1,wg.yborder+text_midpoint,27*text_width,3*text_height,0);
      break;

      case F1_KEY:
      for(i=0; i<no_of_scro; i++)
        {
        if( scro[i].x1 <= mouse_x && 
            scro[i].x2 >= mouse_x &&      
            scro[i].y1 <= mouse_y && 
            scro[i].y2 >= mouse_y) 
          {
          switch (scro[i].no)
            {
            case WIDE_GRAPH:
            help_on_wide_graph();
            break;
            
            case HIRES_GRAPH:
            help_on_hires_graph();
            break;
            
            case AFC_GRAPH:
            help_on_afc_graph();
            break;
            
            case BASEBAND_GRAPH:
            help_on_baseband_graph();
            break;
            
            case POL_GRAPH:
            help_on_pol_graph();
            break;
            
            case COH_GRAPH:
            help_on_coherent_graph();
            break;
            
            case EME_GRAPH:
            help_on_eme_graph();
            break;
            
            case FREQ_GRAPH:
            help_on_freq_graph();
            break;
            }
          }
        }
      if(vga_inkey != 0)help_screen_objects();
      break;

      case F2_KEY:
      if(s_meter_avgnum <0)
        {
        s_meter_avg=0;
        s_meter_avgnum=0;
        }
      else
        {
        s_meter_avgnum=-1;
        }
      break;

      case F3_KEY:
      audio_dump_flag^=1;
      audio_dump_flag&=1;
      break;
      
      if(genparm[CW_DECODE_ENABLE] != 0)
        {
        case ARROW_UP_KEY:
        cg_osc_offset_inc*=1.5;
        if(cg_osc_offset_inc>(baseband_size<<3) )
                                          cg_osc_offset_inc=baseband_size<<3;
        shift_cg_osc_track(0);
        break;
        
        case ARROW_DOWN_KEY:
        cg_osc_offset_inc/=1.3;
        if(cg_osc_offset_inc < 4*text_width)cg_osc_offset_inc = 4*text_width;
        shift_cg_osc_track(0);
        break;

        case ARROW_RIGHT_KEY:
        shift_cg_osc_track(cg_osc_offset_inc);
        break;
        
        case ARROW_LEFT_KEY:
        shift_cg_osc_track(-cg_osc_offset_inc);
        break;
        }    
      }
    }
  }
return 0;
}

void check_mouse_actions(void)
{
int i;
dsp_mouse();
if(leftpressed != 0) rightpressed = 0;
if( rightpressed == 2)
  {
  for(i=0; i<no_of_scro; i++)
    {
    if( scro[i].x1 <= mouse_x && 
        scro[i].x2 >= mouse_x &&      
        scro[i].y1 <= mouse_y && 
        scro[i].y2 >= mouse_y) 
      {
      if(scro[i].no == WIDE_GRAPH)wide_graph_add_signal();
      }
    }
  rightpressed=0;
  }
if(numinput_flag == 0 && leftpressed != 0 )
  {
  if(mouse_active_flag != 0)
    {
    current_mouse_activity();
    }        
  else
    {
    for(i=0; i<no_of_scro; i++)
      {
      if( scro[i].x1 <= mouse_x && 
          scro[i].x2 >= mouse_x &&      
          scro[i].y1 <= mouse_y && 
          scro[i].y2 >= mouse_y) 
        {
        switch (scro[i].no)
          {
          case WIDE_GRAPH:
          mouse_on_wide_graph();
          break;
          
          case HIRES_GRAPH:
          mouse_on_hires_graph();
          break;
          
          case AFC_GRAPH:
          mouse_on_afc_graph();
          break;
          
          case BASEBAND_GRAPH:
          mouse_on_baseband_graph();
          break;
          
          case POL_GRAPH:
          mouse_on_pol_graph();
          break;
          
          case COH_GRAPH:
          mouse_on_coh_graph();
          break;
          
          case EME_GRAPH:
          mouse_on_eme_graph();
          break;
          
          case FREQ_GRAPH:
          mouse_on_freq_graph();
          break;
          } 
        }
      }
    if(leftpressed == 2)leftpressed=0;
    }
  }
}


void select_namlin(char *s,FILE *file)
{
char ch;
int i,k,num, line, col;
fpos_t fileptr[MAX_FILES];
vga_clear();
xytext(10,0,"SELECT A NUMBER FROM THE LIST:");
num=0;
line=2;
col=0;
get_name:;
if(fgetpos(file, &fileptr[num]))end_program(1118);
k=1;
sprintf(s,"%2d:",num);
i=2;
while(k==1 && s[i] != '\n' && s[i] != ' ' && i<MAX_NAMLEN)
  {
  i++;
  k=fread(&s[i],1,1,file);
  }
ch=s[i];
s[i]=0;
if(i>3)
  {
  xytext(col,line,s);
  if(ch != '\n')
  while(k==1 && ch != '\n' )
    {
    k=fread(&ch,1,1,file);
    }
  num++;
  line++;
  }
if(line>HEIGHT/text_height-1)
  {
  line=2;
  col+=MAX_NAMLEN +5;
  if( col+MAX_NAMLEN+5 >= WIDTH/text_width)goto listfull;
  }
if(k == 1 && num < MAX_FILES)goto get_name;
listfull:;
if(sel_line == -1)sel_line=vga_get_integer(42, 0, 3, 0,num-1);
if(fsetpos(file, &fileptr[sel_line]))end_program(1119);
vga_clear();
}

void get_rx_mode(void)    
{
int i;
char s[80];
settextcolor(7);
for(i=0; i<MAX_RX_MODE; i++)
  {
  sprintf(s,"%c=%s",i+'A',rxmodes[i]);
  xytext(0, i+2,s);
  }
settextcolor(15);
xytext(5, MAX_RX_MODE+3, "Select Rx mode");
rx_mode=-1;
while(rx_mode < 0 || rx_mode > MAX_RX_MODE)
  {
  await_keyboard();
  rx_mode=(toupper(vga_inkey)-'A');
  }
vga_clear();
}

void get_parfile_name(char *s)
{
int i, ptr, eq;
ptr=0;
while(s[ptr]!=0 && s[ptr]!=' ')
  {
  if(s[ptr]=='\n')
    {
    savefile_parname[0]=0;
    s[ptr]=0;
    return;
    }
  ptr++;
  }
s[ptr]=0;
ptr++;
while(s[ptr] == ' ')ptr++;
i=0;
eq=1;
while(s[ptr]!=0 && s[ptr]!=' ' && s[ptr] != '\n' && i<40)
  {
  savefile_parname[i]=s[ptr];
  if(s[i] != s[ptr])eq=0; 
  i++;
  ptr++;
  }
if(eq == 1)end_program(1120);  
if(i>= 40)end_program(1115);  
savefile_parname[i]=0;
}

int init_wavread(int sel_file)
{
char s[80];
FILE *file;
int i, k, chunk_size;
short int ishort;
int errnr;
vga_clear();
file = fopen("adwav", "rb");
if (file == NULL)
  {
  xytext(5,5,"adwav missing");
rdfile_x:;
  await_keyboard();
  return 1;
  }
if(sel_file != 0)
  {  
  select_namlin(s,file);
  }  
for(i=0; i<80; i++)s[i]=0;
fread(&s , 1, 70, file);  
fclose(file);
get_parfile_name(s);
save_file = fopen(s, "rb");
if (save_file == NULL)
  {
  xytext(5,4,"Can not open file");
  xytext(5,5,s);
  goto rdfile_x;
  }
// Read the WAV file header.
// First 4 bytes should be "RIFF"
k=fread(&i,sizeof(int),1,save_file);
errnr=0;
if(k !=1 || i != 0x46464952)
  {
  errnr=1;
headerr:;  
  xytext(5,5,s);
  sprintf(s,"Error in .wav file header [%d]",errnr);  
  xytext(5,4,s);
  goto rdfile_x;
  }
// Read file size (we do not need it)
fread(&i,sizeof(int),1,save_file);
// Now we should read "WAVE"
k=fread(&i,sizeof(int),1,save_file);
errnr=2;
if(k !=1 || i != 0x45564157)goto headerr;
// Now we should read "fmt "
k=fread(&i,sizeof(int),1,save_file);
errnr=3;
if(k !=1 || i != 0x20746d66)goto headerr;
// read the size of the format chunk.
k=fread(&chunk_size,sizeof(int),1,save_file);
errnr=4;
if(k !=1 )goto headerr; 
// read the type of data (Format Tag). We only recognize PCM data!
k=fread(&ishort,sizeof(short int),1,save_file);
errnr=5;
if(k !=1 || ishort != 1)
  {
  xytext(5,3,"Unknown WAVE format. (PCM only accepted)");
  goto headerr;
  }
// Read no of channels
k=fread(&ishort,sizeof(short int),1,save_file);
errnr=6;
if(k !=1 || ishort < 1 || ishort > 2)goto headerr;  
ui.rx_channels=ishort;
ui.ad_channels=ishort;
if(ishort == 1)
  {
  ui.input_mode=0;
  }
else  
  {
  ui.input_mode=TWO_CHANNELS;
  }
// Read the sampling speed.
k=fread(&ui.input_speed,sizeof(int),1,save_file);
errnr=7;
if(k !=1 )goto headerr; 
// Read average bytes per second (do not care what it is)
errnr=8;
k=fread(&i,sizeof(int),1,save_file);
if(k !=1 )goto headerr; 
// Read block align to get 8 or 16 bit format.
k=fread(&ishort,sizeof(short int),1,save_file);
errnr=9;
if(k !=1 )goto headerr; 
errnr=10;
ishort/=ui.rx_channels;
if(ishort == 1)
  {
  ui.input_mode+=BYTE_INPUT;
  }
else
  {  
  if(ishort != 2)goto headerr;
  }
// Skip extra bytes if present.
chunk_size-=14;
skip_chunk:;
errnr=11;
while(chunk_size != 0)
  {
  k=fread(&i,1,1,save_file);
  if(k !=1 )goto headerr; 
  chunk_size--;
  }
// Read chunks until we encounter the "data" string.
k=fread(&i,sizeof(int),1,save_file);
errnr=12;
if(k !=1 || i != 0x61746164)
  {
  k=fread(&chunk_size,sizeof(int),1,save_file);
  if(k !=1 )goto headerr; 
  goto skip_chunk;
  }
// Read how much data we have ( do not care)
fread(&i,sizeof(int),1,save_file);
rx_mode=0;
disksave_flag=2;
save_init_flag=0;
diskread_time=0;
fg.passband_center=0;
fg.passband_direction=1;
fft1_direction=fg.passband_direction;
return 0;
}  

void parfile_end(FILE *file)
{
fprintf(file,"\nChange only between brackets.");
fprintf(file,"\nIf file has errors, Linrad will ignore file and prompt");
fprintf(file,"\nfor a complete set of new parameters\n");
fprintf(file,"\n%s",PROGRAM_NAME);
fclose(file);
}


int init_diskread(int sel_file)
{
char s[80];
FILE *file;
int i,kk;
if(sel_file != -1)
  {
  sel_parm=sel_file;
  sel_line=-1;
  }
kk=sel_parm;
if( wav_read_flag != 0)
  {
  return init_wavread(kk);
  }       
vga_clear();
file = fopen("adfile", "rb");
if (file == NULL)
  {
  xytext(5,5,"adfile missing");
rdfile_x:;
  await_keyboard();
  return 1;
  }
if(kk != 0)
  {  
  select_namlin(s,file);
  }  
for(i=0; i<80; i++)s[i]=0;
fread(&s , 1, 70, file);  
fclose(file);
get_parfile_name(s);
for(i=0;i<80;i++)if(s[i]==10)s[i]=0;
save_file = fopen(s, "rb");
if (save_file == NULL)
  {
errfile:;  
  xytext(5,4,"Can not open file");
  xytext(5,5,s);
  goto rdfile_x;
  }
disksave_flag=2;
fread(&ui.input_mode,sizeof(int),1,save_file);
// In case ui.input_mode is negative the file contains more data.
if(ui.input_mode < 0)
  {
  fread(&diskread_time,sizeof(double),1,save_file);
  fread(&fg.passband_center,sizeof(double),1,save_file);
  fread(&fg.passband_direction,sizeof(int),1,save_file);
  if(abs(fg.passband_direction) != 1)goto errfile; 
  fft1_direction=fg.passband_direction;
  fread(&ui.input_mode,sizeof(int),1,save_file);
  }
else
  {
  diskread_time=0;
  fg.passband_center=0;
  fg.passband_direction=1;
  fft1_direction=fg.passband_direction;
  }
fread(&ui.rx_channels,sizeof(int),1,save_file);
if(ui.input_mode >= MODEPARM_MAX)goto errfile;
fread(&ui.ad_channels,sizeof(int),1,save_file);
if(ui.ad_channels > 4 || ui.ad_channels < 1)goto errfile;
fread(&ui.input_speed,sizeof(int),1,save_file);
save_init_flag=0;
diskread_block_counter=0;
fread(&save_init_flag,1,1,save_file);
return 0;
}  

void disksave_start_stop(void)
{
FILE *fc_file, *iq_file;
int i, rdf, wrf;
char save_filename[256];
char s[80];
char indicator[2];
char *disksave_screencopy; 
indicator[0]=' ';
if( (disksave_flag & 0xfffe) == 0)
  {
  disksave_flag^=1;
  if(disksave_flag == 1)
    {
    indicator[0]='S';
// Stop the A/D board to prevent overruns.
    rdf=audio_in;
    wrf=audio_out;
    close_audio_input();
    close_audio_output();
    disksave_screencopy=chk_malloc(DISKSAVE_SCREEN_SIZE,1047);
    gl_getbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,disksave_screencopy);
    gl_fillbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,10);
    xytext(0,1,"Enter file name for raw data.");
    xytext(0,2,"ENTER to skip.");
    xytext(0,3,"=>");
    i=vga_get_filename(2,3,save_filename);          
    if(i==0)
      {
      indicator[0]=' ';
      disksave_flag = 0;
      }
    else 
      {
      save_file = fopen( save_filename, "wb");
      if(save_file == NULL)
        {
        xytext(0,4,"Could not create file!!");
        xytext(0,5,"Press any key");
        await_keyboard();
        disksave_flag = 0;
        }
      else
        {
// Write -1 as a flag telling what version number of Linrad raw data file
// we are writing.        
        i=-1;
        fwrite(&i,sizeof(int),1,save_file); 
        update_recent_time();
        fwrite(&recent_time,sizeof(double),1,save_file);
        fwrite(&fg.passband_center,sizeof(double),1,save_file);
        fwrite(&fg.passband_direction,sizeof(int),1,save_file);
// Write mode info so we know how to process data when reading 
        fwrite(&ui.input_mode,sizeof(int),1,save_file);
        fwrite(&ui.rx_channels,sizeof(int),1,save_file);
        fwrite(&ui.ad_channels,sizeof(int),1,save_file);
        fwrite(&ui.input_speed,sizeof(int),1,save_file);
        save_init_flag=0;
        iq_file=NULL;
        fc_file=NULL;
        if( (fft1_calibrate_flag&1) == 1)
          {
          make_filfunc_filename(s);
          fc_file = fopen(s, "rb");
          if(fc_file != NULL)
            {
            save_init_flag=1;
            }
          else
            {
            end_program(1062);
            }  
          }
        if(  (ui.input_mode&IQ_DATA) != 0 && 
             (fft1_calibrate_flag&2) == 2)
          {
          make_iqcorr_filename(s);
          iq_file = fopen(s, "rb");
          if(iq_file != NULL)
            {
            save_init_flag+=2;
            }
          else  
            {
            end_program(1062);
            }  
          }
        fwrite(&save_init_flag,1,1,save_file);
        if( (save_init_flag & 1) == 1)
          {
rd1:;
          i=fread(&s,1,1,fc_file);
          if(i != 0)
            {
            fwrite(&s,1,1,save_file);
            goto rd1;
            }
          fclose(fc_file);
          fc_file=NULL;
          }
        if( (save_init_flag & 2) == 2)
          {
rd2:;
          i=fread(&s,1,1,iq_file);
          if(i != 0)
            {
            fwrite(&s,1,1,save_file);
            goto rd2;
            }
          fclose(iq_file);
          iq_file=NULL;
          }
        }
      }
    gl_putbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,disksave_screencopy);
    free(disksave_screencopy);
    fflush(NULL);
    sync();
    if(rdf != -1)open_audio_input();
    if(wrf != 1)open_audio_output();
    clear_ad_countertimer();
    if(new_baseb_flag >=0)clear_da();
    make_fft1(0);
    make_fft1(0);
    }
  else
    {
    fflush(NULL);
    fclose(save_file);
    save_file=NULL;
    sync();
    }
  }
indicator[1]=0;  
settextcolor(14);
i=wg.xright-text_width-2;
if(i<2)i=2;
errno=20122;
ck_write(i,wg.yborder+2,indicator);
settextcolor(7);
}

void wavsave_start_stop(void)
{
int i, rdf, wrf;
char wav_filename[256];
char *wav_write_screencopy; 
char indicator[2];
indicator[0]=' ';
if( wav_write_flag == 0)
  {
  wav_write_flag=1;
  indicator[0]='W';
// Stop the A/D board to prevent overruns.
  rdf=audio_in;
  wrf=audio_out;
  close_audio_input();
  close_audio_output();
  wav_write_screencopy=chk_malloc(DISKSAVE_SCREEN_SIZE,1047);
  gl_getbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,wav_write_screencopy);
  gl_fillbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,10);
  xytext(0,1,"Enter name for audio output file");
  xytext(0,2,"ENTER to skip.");
  xytext(0,3,"=>");
  i=vga_get_filename(2,3,wav_filename);          
  if(i==0)
    {
    indicator[0]=' ';
    wav_write_flag = 0;
    }
  else 
    {
    i=0;
    while(wav_filename[i] != 0)i++;
    wav_filename[i  ]='.';
    wav_filename[i+1]='w';
    wav_filename[i+2]='a';
    wav_filename[i+3]='v';
    wav_filename[i+4]=0;
    wav_file = fopen( wav_filename, "wb");
    if(wav_file == NULL)
      {
      xytext(0,4,"Could not create file!!");
      xytext(0,5,"Press any key");
      await_keyboard();
      wav_write_flag = 0;
      }
    else
      {
// Write the .wav header, but make the file size gigantic
// We will write the correct size when closing, but if
// system crashes we can fix the header and recover data.
// Speed, channels and bits will be ok. Size will indicate loss of data. 
      write_wav_header(0x7fffffff);
      wavfile_bytes=0;
      }
    }
  gl_putbox(0,0,DISKSAVE_X_SIZE,DISKSAVE_Y_SIZE,wav_write_screencopy);
  free(wav_write_screencopy);
  fflush(NULL);
  sync();
  if(rdf != -1)open_audio_input();
  if(wrf != 1)open_audio_output();
  clear_ad_countertimer();
  if(new_baseb_flag >=0)clear_da();
  make_fft1(0);
  make_fft1(0);
  }
else
  {
  fflush(NULL);
  write_wav_header(wavfile_bytes);
  fclose(wav_file);
  wav_write_flag = 0;
  fflush(NULL);
  sync();
  }
indicator[1]=0;  
settextcolor(14);
i=wg.xright-2*text_width-2;
if(i<2)i=2;
errno=20123;
ck_write(i,wg.yborder+2,indicator);
settextcolor(7);
}

void write_wav_header(int totbytes)
{
int i;
// Write the header for a .wav file using the current output
// settings. 
// First point to start of file.
if(totbytes == 0x7fffffff)
  {
  if(fgetpos(wav_file, &wav_wrpos))end_program(1114);
  }
else
  {
  if(fsetpos(wav_file, &wav_wrpos) != 0)end_program(1111);
  }
// First 4 bytes should be "RIFF"
if(fwrite("RIFF",sizeof(int),1,wav_file)!= 1)
  {
headerr:;
  end_program(1112);
  }
// Now file size in bytes -8
// The format chunk uses 16 bytes. 
i=totbytes+16;
if(fwrite(&i,sizeof(int),1,wav_file)!= 1)goto headerr;
// The next code word pair is "WAVEfmt "
if(fwrite("WAVEfmt ",sizeof(int),2,wav_file)!= 2)goto headerr;
// Write the size of the format chunk = 16
i=16;
if(fwrite(&i,sizeof(int),1,wav_file)!= 1)goto headerr;
// Write the type of data (Format Tag = 1 for PCM data)
i=1;
if(fwrite(&i,sizeof(short int),1,wav_file)!=1)goto headerr;
// Write no of channels
if(fwrite(&da_output_channels,sizeof(short int),1,wav_file)!=1)goto headerr;
// Write the output speed twice. 
for(i=0; i<2; i++)
  {
  if(fwrite(&genparm[DA_OUTPUT_SPEED],sizeof(int),1,wav_file)!=1)goto headerr;
  }
i=da_output_channels*da_output_bytes;
// Write block align.
if(fwrite(&i,sizeof(short int),1,wav_file)!=1)goto headerr;
// Write bits per sample
i=8*da_output_bytes;
if(fwrite(&i,sizeof(short int),1,wav_file)!=1)goto headerr;
// Now write the code word "data"
if(fwrite("data",sizeof(int),1,wav_file)!=1)goto headerr;
// And finally the size of the data block
if(fwrite(&totbytes,sizeof(int),1,wav_file)!=1)goto headerr;
}

void init_memalloc(MEM_INF *mm, int max)
{
memalloc_no=0;
memalloc_max=max;
memalloc_mem=mm;
}

void mem(int num, void *pointer, int size, int scratch_size)
{
// Skip if outside array. Error code will come on return from memalloc.
if(memalloc_no<0 || memalloc_no>=memalloc_max)goto skip;
memalloc_mem[memalloc_no].pointer=pointer;
if(size<0)end_program(1125);
memalloc_mem[memalloc_no].size=(size+15)&0xfffffff0;
memalloc_mem[memalloc_no].scratch_size=(scratch_size+15)&0xfffffff0;
memalloc_mem[memalloc_no].num=num;
skip:;
memalloc_no++;
}

int memalloc( int *handle, char *s)
{
int i,j,k,m,totbytes;
char *testbuf;
char *x;
int *ptr;
if(memalloc_no<0 || memalloc_no >= memalloc_max)end_program(1136);
memalloc_mem[memalloc_no].pointer=s;
memalloc_mem[memalloc_no].size=-1;
totbytes=16+DEBMEM;
for(i=0; i<memalloc_no; i++)
  {
  totbytes+=memalloc_mem[i].size+memalloc_mem[i].scratch_size+DEBMEM;
  }
handle[0]=(int)(malloc(totbytes));
if(handle[0] == 0)return 0;
if(ui.memlock == 0)
  {
  testbuf=(char*)handle[0];
  m=getpagesize();
  i=m/2;
  j=0;
  while( i<totbytes)
    {
    testbuf[i]=0;
    i+=m;
    j++;
    if(j > 1000)
      {
      if(mlockall(MCL_CURRENT) != 0)
        {
        handle[0]=(int)(chk_free(testbuf));
        munlockall();
        return 0;
        }
      j=0;
      }  
    }  
  }
k=((int)(handle[0])+15)&0xfffffff0;
for(i=0; i<memalloc_no; i++)
  {
  x=(void*)(k);
  k+=DEBMEM;
  for (j=0; j<DEBMEM; j++)x[j]=j&0xff;
  k+=memalloc_mem[i].scratch_size;
  ptr=memalloc_mem[i].pointer;
  ptr[0]=k;
  k+=memalloc_mem[i].size;
  }
x=(void*)(k);
k+=DEBMEM;
for (j=0; j<DEBMEM; j++)x[j]=j&0xff;
if(dmp != NULL)PERMDEB"%s: %3.1f Megabytes(%d arrays)\n",s,
                                totbytes*0.000001,memalloc_no);
return totbytes;
}

void memcheck(int callno, MEM_INF *mm, int *handle)
{
int i, j, k, errflag;
unsigned char *x;
k=(handle[0]+15)&0xfffffff0;
i=0;
errflag=0;
while(mm[i].size != -1)
  {
  x=(void*)(k);
  k+=DEBMEM;
  for (j=0; j<DEBMEM; j++)
    {
    if( x[j] != (j&0xff) )
      {
      errflag=1;
      if(dmp != NULL)
        {
        PERMDEB"\nMEMORY ERROR mm=%d[%d]  data is%d(%d)  Call no %d",
                                            mm[i].num,j,x[j],j&0xff,callno);
        }
      printf("\nMEMORY ERROR mm=%d[%d]  data is%d (%d)  Call no %d",
                                            mm[i].num,j,x[j],j&0xff,callno);
      } 
    }
  k+=mm[i].size;
  k+=mm[i].scratch_size;
  i++;
  }
if(errflag != 0)
  {
  PERMDEB"\nError in %s\n",(char*)mm[i].pointer);
  printf("\nError in %s\n",(char*)mm[i].pointer);
  fflush(NULL);
  exit(0);
  }
}

void update_recent_time(void)
{
struct timeval t;
gettimeofday(&t,NULL);
recent_time=0.000001*t.tv_usec+t.tv_sec;
}

void clear_workload_timers(void)
{
int i;
update_recent_time();
workload_max_latency=0;
workload_start_time=recent_time;
workload_total=0.00000001;
workload_idle=0;
info_time=recent_time;
da_maxempty=0;
for(i=0; i<ui.ad_channels; i++)
  {
  ad_maxamp[i]=0;
  }
for(i=0; i<ui.rx_channels; i++)
  {
  timf1_maxamp[2*i  ]=0;
  timf1_maxamp[2*i+1]=0;
  fft1_maxamp[2*i  ]=0;
  fft1_maxamp[2*i+1]=0;
  fft2_maxamp[i]=0;
  }
for(i=0; i<MAX_FLOWCNT; i++)flowcnt[i]=0;
flowtime=0;
idle_flag=0;
if(s_meter_avgnum >= 0)
  {
  s_meter_avgnum=0;
  s_meter_avg=0;
  }
}

void clear_sound_timers(void)
{
update_recent_time();
sound_start_time=recent_time;
bg_maxamp_time=recent_time;
}

void show_name_and_size(void)
{
int i,imax;
char s[120];
float fftx_,base_,afc_,hires_,fft3_;
if(wg.yborder-3.5*text_height<wg.ytop)return;
fftx_=0.000001*fftx_totmem;
base_=0.000001*baseband_totmem;
afc_= 0.000001*afc_totmem;
fft3_=0.000001*fft3_totmem;
hires_=0.000001*hires_totmem;
sprintf(s,"%s %.1f Mbytes. fft1,fft2=%.1f  hires %.1f \
 fft3=%.1f  afc=%.1f  bas=%.1f)",
           PROGRAM_NAME,(fftx_+base_+afc_+hires_+fft3_),
           fftx_,hires_,fft3_,afc_,base_);
imax=(wg.xright-wg.xleft)/text_width-6;
if(imax > 120)imax=120;
i=0;
while(s[i]!=0  && i<imax)i++;
s[i]=0; 
settextcolor(15);
errno=20124;
ck_write(wg.xleft+4*text_width,wg.yborder-2*text_height,s);
settextcolor(7);
}

void audio_output_routines(void)
{
int i;
float t1;
if(new_baseb_flag == 0)
  {  
  make_audio_output();
  }
else
  {
  make_timing_info();
  if(new_baseb_flag >= 2)
    {
    if(new_baseb_flag == 3)
      {
      daout_px=daout_pa;
      }
    else
      {  
      if(total_wttim > da_wait_time)
        {
        new_baseb_flag=1;
        }
      else
        {
        if(disksave_flag == 2)
          {
          newdat_flag &= -1-ND_BASEB-ND_DASIG;
          } 
        }
      }
    }  
  else   //(new_baseb_flag == 1)
    {
// We want data in buffers corresponding to da_wait_time.
// We arrive here when there is more than that, move pointers
// to discard data corresponding to total_wttim-da_wait_time
// Start with the daout pointers.

    t1=total_wttim-da_wait_time;
    i=t1*genparm[DA_OUTPUT_SPEED];
    if(i > db_wts)i=db_wts;
    t1-=(float)(i)/genparm[DA_OUTPUT_SPEED];
    i=i*da_output_channels*da_output_bytes;
    daout_px=(daout_px+i+daout_bufmask+1)&daout_bufmask;
    if( ((daout_pa-daout_px+daout_bufmask+1)&daout_bufmask) 
                                    <= daout_block) newdat_flag&=-1-ND_DASIG;
// If there is still too much data in memory, move the baseb pointer
// to discard more data.
    if(t1*genparm[DA_OUTPUT_SPEED] > 4)
      {
      i=t1*baseband_sampling_speed; 
      if(i > baseb_wts-4)
        {
        baseb_fx=(baseb_pa-4+baseband_size)&baseband_mask;
        baseb_wts=4;
        newdat_flag&=-1-ND_BASEB;
        new_baseb_flag=2;
        return;
        }
      baseb_fx=baseb_fx+i;
      if(baseb_fx>baseband_size)baseb_fx-=baseband_size;
      }    
    dasync_counter=0;
    dasync_sum=0;
    dasync_avg1=-1.5;
    dasync_avg2=-1.5;
    dasync_avg3=-1.5;
    new_baseb_flag=0;
    make_audio_signal();
    make_audio_output();
    da_block_counter=0;
    if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1080);
    if(da_info.bytes > daout_block) fullbuf_flag &=-1-ND_DABUF;
    update_recent_time();
    dasync_time=recent_time;
    dasync_avgtime=recent_time;
    daspeed_time=recent_time; 
    da_start_time=recent_time; 
    da_start_bytes=da_info.bytes;
    da_empty_flag = 0;
    }  
  }
}

void ensure_valid_input(void)
{
audio_buf_info ad_info;
int i;
double start_time;
update_recent_time();
start_time=recent_time;
// Force a read from audio input to start A/D converter.
timf1p_pa=0;  
timf1p_px=0;
ctimf1p_pa=0;
ctimf1p_px=0;
fft1_pa=0;
fft1_px=0;
fft1_na=0;
fft1_nm=0;
fft1_nb=0;
fft1_nc=0;
fft1_nx=0;
if(network_flag != 2)make_fft1(0);
// Make transforms until we know data is valid.
if(genparm[FIRST_FFT_SINPOW]==0)
  {
  i=1;
  }
else  
  {
  if(genparm[FIRST_FFT_SINPOW]<=2)
    {
    i=2;
    }
  else
    {
    i=3;
    }
  }
while( ((fft1_pa-fft1_px+fft1_mask+1)&fft1_mask) < i*fft1_block)
  {
  if(disksave_flag < 2)
    { 
    if(network_flag == 2)
      { 
      if(ioctl(network_fd, FIONREAD, &ad_info.bytes)==-1)end_program(1154);
      ad_info.bytes-=16;if(ad_info.bytes<0)ad_info.bytes=0;
      } 
    else 
      { 
      if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1150);
      }
    if(ad_info.bytes>ad_read_bytes)make_fft1(0);
    }
  if(disksave_flag == 2)make_fft1(0);
  }
fft1_px=(fft1_pa-(i-1)*fft1_block+fft1_mask+1)&fft1_mask;
// Make sure we do not have data in the hardware buffer.
if(disksave_flag < 2)
  {
  i=0;
clear_inbuf:;
  i++;
  if(network_flag == 2)
    {
    if(ioctl(network_fd, FIONREAD, &ad_info.bytes)==-1)end_program(1155);
    ad_info.bytes-=16;
    if(ad_info.bytes<0)ad_info.bytes=0;
    } 
  else 
    { 
    if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1151);
    }
  if(ad_info.bytes>ad_read_bytes)
    {
    make_fft1(0);
    update_recent_time();
    if(recent_time - start_time < 4)goto clear_inbuf;
    }
  }
clear_ad_countertimer();
clear_sound_timers();
}

int check_timing(int idle)
{
char s[80];
float t1,t2;
struct timeval tim;
// Update recent time.
// Issue a warning in case it was long ago we reached this far
update_recent_time();  
if( (recent_time-da_start_time > 4 && new_baseb_flag == 0) || 
                                new_baseb_flag < 0)
  {                                
  t1=recent_time - latest_idle_time; 
  if(t1 > 1)
    {
    if(t1 > 5)
      {
      wg_error("SPEED ERROR",4);
      }
    else
      {
      wg_error("SPEED WARNING",3);
      }
    latest_idle_time=recent_time;
    }
  if(idle != 0)latest_idle_time=recent_time;
  }
else
  {
  latest_idle_time=recent_time;  
  }
// Find out how much data we have in the different buffers and how much
// time it corresponds to.
if( recent_time - info_time > 0.5)
  {
  if(timinfo_flag!=0)show_timing_info();
  if(ampinfo_flag!=0)show_amp_info();
  info_time=recent_time;
  }
if(disksave_flag == 1 && disksync_time-recent_time > 1.4)return -1;
// Calculate the A/D sampling speed (if A/D is in use).
if(disksave_flag < 2)
  {
  if(ad_block_counter > ad_block_check && 
     overrun_flag == 0 && 
     da_empty_flag == 0 &&
     network_flag != 2)
    {
// Calculate the ad sampling speed.
    ad_block_check+=interrupt_rate;
    gettimeofday(&tim,NULL);
    t1=0.000001*(tim.tv_usec-ad_tim.tv_usec)+(tim.tv_sec-ad_tim.tv_sec);
    measured_ad_speed=(ad_block_counter+0.25)*ad_read_fragments/t1;
    if(timinfo_flag != 0)
      {
      sprintf(s,"A/D:%dHz ",(int)(measured_ad_speed));
      xytext(0,HEIGHT/text_height-8,s);
      }
// The buffer is empty now so the uncertainty is about one ad_block.
// Generously allow 1% deviation from nominal sampling speed
// and check if the speed is within limits.
    t2=1.01+3/ad_block_check;
    if(measured_ad_speed > ui.input_speed*t2 || 
                                     measured_ad_speed*t2 < ui.input_speed)
      {
      sprintf(s,"A/D SPEED ERROR: %dHz (nominal %dHz)",
                                   (int)(measured_ad_speed),ui.input_speed);
      wg_error(s,2);
      }
    }
  }  
if(new_baseb_flag == 0)
  {
// In case the output has been running for 3 seconds or more, find the
// output sampling speed.
  if(recent_time-daspeed_time > 3 && da_empty_flag == 0)
    {
    daspeed_time+=1;
    update_recent_time();
    if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1109);
    if(da_info.bytes > daout_block) fullbuf_flag &=-1-ND_DABUF;
    measured_da_speed=(da_block_counter*daout_block+da_info.bytes
                        -da_start_bytes)/(da_output_channels*da_output_bytes*
                                       (recent_time-da_start_time));
    if(timinfo_flag != 0)
      {
      sprintf(s,"D/A:%dHz ",(int)(measured_da_speed));
      xytext(14,HEIGHT/text_height-8,s);
      }
    }
  }
// Keep track of the error in the frequency ratio between input and output.
// Average the total processing delay and adjust resampling rate to 
// keep it constant. 
#define DASYNC_TIMINC 0.25 //Interval in seconds for summing up averages.
#define DASYNC_MAXCOUNT 15 //Number used to form average
if(  disksave_flag != 2 && 
     new_baseb_flag == 0 &&
     recent_time-dasync_avgtime > DASYNC_TIMINC && 
     da_empty_flag == 0)
  {
  dasync_avgtime=recent_time;
  make_timing_info();
  if(timinfo_flag != 0)
    {
    sprintf(s,"sync:%f ",total_wttim-dasync_avg3);
    xytext(26,HEIGHT/text_height-8,s);
    }
  if(dasync_avg3 >= 0 && fabs(total_wttim-dasync_avg3)>0.5)
    {
    dasync_errors++;
    sprintf(s,"DA SYNC ERRORS %d",dasync_errors);
    wg_error(s,1);
    if(new_baseb_flag >= 0)clear_da();
    return -1;
    }
  dasync_counter++;
  dasync_sum+=total_wttim;
  if(dasync_counter >= DASYNC_MAXCOUNT)
    {
    dasync_avg1=dasync_sum/dasync_counter;
    dasync_counter=0;
    dasync_sum=0;
    update_recent_time();
    t2=recent_time-dasync_time;
    dasync_time=recent_time;
    if(dasync_avg2 > 0)
      {
// The drift between input and output after t2 seconds is
// the difference between the two averages.
// Convert to frequency ratio, but use half the error only 
// to avoid oscillations.
// Correct for the deviation from the first average, but only 20%
// so we drift slowy towards it if we are off.
      t1= (t2+0.5*(dasync_avg2-dasync_avg1)+
                                   0.2*(dasync_avg3-dasync_avg1))/t2;
      new_da_resample_ratio=da_resample_ratio*t1;
      dasync_avg2=dasync_avg1;
      }
    else
      {
      if(dasync_avg3 <0) 
        {
        dasync_avg3=dasync_avg1;
        }
      else
        {
        dasync_avg2=dasync_avg1;
        }
      }
    }
  }
// Show baseband max amplitude in the baseband graph
if(recent_time-bg_maxamp_time > 0.05)
  {
  bg_maxamp_time=recent_time;
  if( bg_maxamp != 0)show_bg_maxamp();
  }
if(fg.passband_center != old_passband_center  && network_flag==2)
  {
  show_center_frequency();
  old_passband_center=fg.passband_center;
  }
return 0;
}

void show_workload(void)
{
char s[80];
workload_idle+=recent_time-idle_start;
flowtime+=recent_time-workload_start_time-workload_total;
workload_total=recent_time-workload_start_time;
workload_percentage=100*(1-workload_idle/(recent_time-workload_start_time));
sprintf(s,"%5.2f%% %6.4f ",workload_percentage, workload_max_latency);
if(dmp != 0)settextcolor(12);
xytext(0,HEIGHT/text_height-1,s);
if(dmp != 0)settextcolor(7);
}


void modesub_idle(int idle)
{
struct timeval timtest;
audio_buf_info ad_info;
float t1;
double latency;
float idle_time;
int i;
int work_flag;
int idle_ad_bytes;
int idle_da_bytes;
char s[80];
if(disksave_flag < 2)
  { 
  if(network_flag == 2)
    {
    if(ioctl(network_fd, FIONREAD, &ad_info.bytes)==-1)end_program(1156);
    ad_info.bytes-=16;
    if(ad_info.bytes<0)ad_info.bytes=0;
    }
  else
    {
    if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1152);
    }
  if(ad_info.bytes>ad_read_bytes)
    {
    show_workload();
    return;
    }
  }
idle_ad_bytes=ad_info.bytes;
if(new_baseb_flag == 0)
  {
  if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1091);
  if(da_info.bytes > daout_block) fullbuf_flag &=-1-ND_DABUF;
  idle_da_bytes=da_info.bytes;
  }
else
  {
  idle_da_bytes=0;
  }
update_recent_time();
idle_start=recent_time;
if(ui.memlock == 0)
  {
  if(mlockall(MCL_CURRENT) != 0)lir_status=LIR_MEMERR;
  }
work_flag=0;


// ******************************************************************
//                            IDLE   
// ******************************************************************
// Below this line we just check whether it is time to do something
// or if we can stay idle a little longer.
if(disksave_flag == 1)
  {
  if(fabs(disksync_time-recent_time) > 0.2)
    {
    disksync_time = recent_time;
    sync();
    }
  }
while(work_flag == 0)
  {
  make_timing_info();
  latency=recent_time;
  update_recent_time();
  latency=recent_time-latency;
  if(workload_max_latency < latency)workload_max_latency=latency;
  idle_time=recent_time-idle_start;
// Call the hardware control routine if flag says so.
  if(hware_flag != 0 && network_flag != 2)
    {
    if(recent_time-hware_time > 0.0001)
      {
      control_hware();
      hware_time=recent_time;
      }
    }
// Save the max time for one turn in this loop.
// It will be an indication to the user that the kernel is doing
// it's own things in the background.
  if(disksave_flag < 2)
    {
    if(network_flag == 2)
      {
      if(ioctl(network_fd, FIONREAD, &ad_info.bytes)==-1)end_program(1157);
      ad_info.bytes-=16;
      if(ad_info.bytes<0)ad_info.bytes=0;
      }
    else
      {
      if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1153);
      }
    if(ad_info.bytes>ad_read_bytes)
      {
      work_flag=2;
      newdat_flag|=ND_HWARE;
      }
    if(idle_time > 2)
      {
      if( idle_ad_bytes == ad_info.bytes)
        {
        end_program(1104);
        }
      work_flag=1;
      }  
    }  
  if(new_baseb_flag >= 0 && (ui.input_mode&NO_DUPLEX) == 0 && audio_out != -1)
    {
    if(recent_time-da_start_time < 1) da_maxempty=0;
    if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1092);
    if(da_info.bytes > daout_block)
      {
      fullbuf_flag &=-1-ND_DABUF;
      if( (newdat_flag&(ND_DASIG+ND_BASEB)) != 0 && 
          (fullbuf_flag & ND_DABUF) == 0 )work_flag=3;
      }
    }  
  if(new_baseb_flag >= 0 && idle_time > 0.5 && idle_da_bytes == da_info.bytes)
    {
    overrun_flag++;
    work_flag=1;
    }
  if(disksave_flag == 2 || disksave_flag == 4)
    {
    newdat_flag&=-1-ND_HWARE;  
    switch (new_baseb_flag)
      {
      case -1:
// There is no output at the moment.
// Use diskread_timer to decide if we waited long enough already.
      t1=(ad_block_counter+1)*ad_read_fragments/
                                    (0.0000001+(recent_time-diskread_timer));
      if(t1 < ui.input_speed)
        {
        newdat_flag|=ND_HWARE;
        work_flag=1;
        }
      break;

      case 0:
      make_timing_info();
      if(total_wttim < da_wait_time)
        {
        newdat_flag|=ND_HWARE;
        work_flag=1;
        }
      if(recent_time-idle_start > .2)
        {
        if(ioctl(audio_out,SNDCTL_DSP_RESET,0)==-1)end_program(62245);
        close_audio_output();
        open_audio_output();
        newdat_flag|=ND_HWARE;
        work_flag=1;
        }
// Clear ad_block_counter and diskread_timer but allow count above15
// so we get the timing and amplitude margin printouts
      if(ad_block_counter > 16)
        {
        diskread_timer=recent_time;
        ad_block_counter=0;
        }
      break;

      case 1:
      newdat_flag|=ND_HWARE;
      work_flag=1;
      break;

      default:
      newdat_flag|=ND_HWARE;
      work_flag=1;
      break;
      }
    }
  if(idle == 0)work_flag=5; 
  if(work_flag == 0)
    {  
    idle_flag++;
    switch (idle_flag)
      {

      case 1:
      if(eme_active_flag != 0)
        {
        if(recent_time - eme_time > 3)
          {
          calculate_moon_data();
          eme_time=recent_time;
          }
        }
      break;
     
      case 2:  
      if(fft1_handle != NULL)memcheck(198,fft1mem,(int*)&fft1_handle);
      if(baseband_handle != NULL)memcheck(298,basebmem,(int*)&baseband_handle);
      if(afc_handle != NULL)memcheck(398,afcmem,(int*)&afc_handle);
      if(fft3_handle != NULL)memcheck(498,fft3mem,(int*)&fft3_handle);
      if(hires_handle != NULL)memcheck(598,hiresmem,(int*)&hires_handle);
      break;

      case 3:
      if(ui.sleep == 0)
        { 
        if(ad_info.bytes < (ad_read_bytes >>2))
          {
// The usleep function has to be preceeded by a call to 
// gettimeofday on some systems. For some reason this makes
// usleep() less likely to last longer than 50 milliseconds.
// Without this statement, overrun errors occur now and then
// because of excessive delay.
          gettimeofday(&timtest,NULL);
          usleep(5000);
          }
        }  
      if( allow_parport != 0)
        {
// Read now and then to make apm aware Linrad is running.
// See z_APM.txt
        i=inb(0x379);
        }
      break;

      default:
      idle_flag=0;
      break;
      }
    }
  }

if(new_baseb_flag >= 0)
  {
  if(bg_oscill_on != 0)
    {
    if( ((timf3_pa-timf3_ps+timf3_mask) & timf3_mask) > 
                                             timf3_oscilloscope_limit)
      {
      timf3_oscilloscope();
      }
    }
  } 
latency=recent_time;
update_recent_time();
latency=recent_time-latency;
if(workload_max_latency < latency)workload_max_latency=latency;
show_workload();
if(flowinfo_flag != 0)
  {
  if(flowtime > 0.3)
    {
    flowtime*=.2;
    for(i=0; i<MAX_FLOWCNT; i++)
      {
      flowcnt[i]*=.2;
      }
    }  
  for(i=0; i<MAX_FLOWCNT; i++)
    {
    sprintf(s,"%2d %4d ",i,(int)(10*flowcnt[i]/flowtime));
    xytext(WIDTH/text_width-9,i,s); 
    }
  }
}
