
#include <vgagl.h>
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "sigdef.h"
#include "screendef.h"

#define ZZ 0.0000001

void remove_dash(void)
{
float t1;
int ia, ib;
ia=cw[cw_ptr].midpoint-0.5*dash_pts+baseband_size;
ia&=baseband_mask;
ib=(ia+dash_pts+1)&baseband_mask;
while(ia != ib)
  {
  baseb_fit[2*ia]=0;
  baseb_fit[2*ia+1]=0;
  baseb_detect_flag[ia]=0;
  ia=(ia+1)&baseband_mask;
  }
ia=cw_ptr+1;
while(ia < no_of_cwdat)
  {
  cw[ia-1]=cw[ia];
  ia++;
  }
if(cw_ptr !=0)
  {  
  t1=cw[cw_ptr].midpoint-cw[cw_ptr-1].midpoint;
  if(t1 < 0)t1+=baseband_size;
  cw[cw_ptr].sep=t1;
  }
no_of_cwdat--;
}  


void set_long_region_envelope(void)
{
int i, ia, ib, len;
float t1,t2,r1,r2;
// We have two dashes pointed to by cw_ptr and cw_ptr-1.
// We know the envelope at these points but the separation is large.
// Set envelope at both ends by extrapolation.
ia=cw[cw_ptr-1].midpoint+(dash_pts>>2);
ia&=baseband_mask;
ib=baseband_size+cw[cw_ptr].midpoint-(dash_pts>>2);
ib&=baseband_mask;
t1=baseb_envelope[2*ia  ];
r1=baseb_envelope[2*ia+1];
t2=baseb_envelope[2*ib  ];
r2=baseb_envelope[2*ib+1];
len=2*dash_pts;
i=0;
while( i < len )
  {
  baseb_envelope[2*ia]=t1;
  baseb_envelope[2*ia+1]=r1;
  baseb_envelope[2*ib]=t2;
  baseb_envelope[2*ib+1]=r2;
  ia=(ia+1)&baseband_mask;
  if(ia == ib)goto avgenv;
  ib=(ib+baseband_mask)&baseband_mask;
  if(ia == ib)goto avgenv;
  i++;
  }
return;
avgenv:;
t1=0.5*(t1+t2);
r1=0.5*(r1+r2);
while( i < len )
  {
  baseb_envelope[2*ia]=t1;
  baseb_envelope[2*ia+1]=r1;
  baseb_envelope[2*ib]=t1;
  baseb_envelope[2*ib+1]=r1;
  ia=(ia+1)&baseband_mask;
  if(ia == ib)goto avgenv;
  ib=(ib+baseband_mask)&baseband_mask;
  if(ia == ib)goto avgenv;
  i++;
  }
}  


void set_region_envelope(void)
{
int ia, ib, len;
float t1,t2,r1,r2;
// We have two dashes pointed to by cw_ptr and cw_ptr-1.
// We know the envelope at these points.
// Set a reasonable envelope across the interval by linear interpolation.
ia=cw[cw_ptr-1].midpoint+(dash_pts>>2);
ia&=baseband_mask;
ib=baseband_size+cw[cw_ptr].midpoint-(dash_pts>>2);
ib&=baseband_mask;
t1=baseb_envelope[2*ia  ];
r1=baseb_envelope[2*ia+1];
len=(ib-ia+baseband_size)&baseband_mask;
t2=(baseb_envelope[2*ib  ]-t1)/len;
r2=(baseb_envelope[2*ib+1]-r1)/len;
while(ia != ib)
  {
  t1+=t2;
  r1+=r2;
  baseb_envelope[2*ia]=t1;
  baseb_envelope[2*ia+1]=r1;
  ia=(ia+1)&baseband_mask;
  }
}  

void clear_region(void)
{
int ia, ib;
// We have two dashes pointed to by cw_ptr and cw_ptr-1.
// We know the envelope at these points.
// Clear baseb_fit across the interval.
ia=cw[cw_ptr-1].midpoint+0.5*dash_pts+baseband_size;
ib=cw[cw_ptr].midpoint-0.5*dash_pts+1+baseband_size;
ia&=baseband_mask;
ib&=baseband_mask;
while(ia != ib)
  {
  baseb_fit[2*ia]=0;
  baseb_fit[2*ia+1]=0;
  ia=(ia+1)&baseband_mask;
  }
}  

void accurate_store_dash(void)
{
float sum1,sum2;
float re, im, c0, t1, t2, f1, f2;
int ia, ib, ja, skipflag;
sum1=0;
sum2=0;
re=cg_wave_re;
im=cg_wave_im;
ia=cg_wave_midpoint+0.5;
if(ia < cg_wave_midpoint) ia++;
ia &= baseband_mask;
ib=(ia+baseband_mask)&baseband_mask;
if( (dash_pts&1) == 0)
  {
  c0=0.5*(dash_pts-1);
  }
else
  {
  c0=(dash_pts>>1)+1;
  }  
// cg_wave_midpoint points to the center of the dash. 
// c0 points to the center of dash_waveform.
store:;
skipflag=0;
t1=ia-cg_wave_midpoint;
if(t1<0)t1+=baseband_size;
t1=c0+t1;
if(t1 >= dash_pts-1)
  {
  baseb_fit[2*ia  ]=0;
  baseb_fit[2*ia+1]=0;
  baseb_detect_flag[ia]=DET_DASH;
  baseb_envelope[2*ia  ]=re;
  baseb_envelope[2*ia+1]=im;
  skipflag=1;
  }
else
  {
  ja=t1;
  t1-=ja;
  t2=1-t1;
  f1=t2*dash_waveform[2*ja  ]+t1*dash_waveform[2*ja+2];
  f2=t2*dash_waveform[2*ja+1]+t1*dash_waveform[2*ja+3];
  baseb_fit[2*ia  ]=re*f1+im*f2;
  baseb_fit[2*ia+1]=re*f2-im*f1;
  baseb_detect_flag[ia]=DET_DASH;
  baseb_envelope[2*ia  ]=re;
  baseb_envelope[2*ia+1]=im;
  sum1+=baseb_totpwr[ia];
  t1=baseb[2*ia  ]-baseb_fit[2*ia  ];
  t1=baseb[2*ia+1]-baseb_fit[2*ia+1];
  sum2+=t1*t1+t2*t2;
  }
t1=cg_wave_midpoint-ib;
if(t1<0)t1+=baseband_size;
t1=c0-t1/1;
if(t1 < 1)
  {
  baseb_fit[2*ib  ]=0;
  baseb_fit[2*ib+1]=0;
  skipflag++;
  }
else
  {
  ja=t1;
  t1-=ja;
  t2=1-t1;
  f1=t2*dash_waveform[2*ja  ]+t1*dash_waveform[2*ja+2];
  f2=t2*dash_waveform[2*ja+1]+t1*dash_waveform[2*ja+3];
  baseb_fit[2*ib  ]=re*f1+im*f2;
  baseb_fit[2*ib+1]=re*f2-im*f1;
  baseb_detect_flag[ib]=DET_DASH;
  baseb_envelope[2*ib  ]=re;
  baseb_envelope[2*ib+1]=im;
  sum1+=baseb_totpwr[ia];
  t1=baseb[2*ia  ]-baseb_fit[2*ia  ];
  t1=baseb[2*ia+1]-baseb_fit[2*ia+1];
  sum2+=t1*t1+t2*t2;
  }
ia=(ia+1)&baseband_mask;
ib=(ib+baseband_mask)&baseband_mask;
if(skipflag != 2)goto store;
cg_wave_dat=sum1/dash_pts;
cg_wave_err=sum2/dash_pts;
}

void store_space(void)
{
float t1;
int ia,ib;
// The separation between two dots should be 2* cwbit_pts
// When two dots surround a space, data is stored in dot_pts points.
// The two end points are probably zero.
// fill zeroes in the remaining points.
t1=2*cwbit_pts-dot_pts+2;
ia=cg_wave_midpoint-0.5*t1+baseband_size;
ia&=baseband_mask;
ib=(ia+(int)(t1+0.5))&baseband_mask;
space_midpoint[no_of_spaces]=cg_wave_midpoint;
no_of_spaces++;
if(no_of_spaces >= max_no_of_spaces)end_program(1166);
// Store zero. The phase is assumed to be already in place.
while(ia != ib)
  {
  baseb_fit[2*ia  ]=0;
  baseb_fit[2*ia+1]=0;
  baseb_detect_flag[ia]=DET_KEY_UP;
  ia=(ia+1)&baseband_mask;
  }
}

void store_tripple_space(void)
{
int ia,ib;
ia=cg_wave_midpoint-0.5*dash_pts+baseband_size;
ia&=baseband_mask;
ib=(ia+dash_pts+1)&baseband_mask;
trispace_midpoint[no_of_trispaces]=cg_wave_midpoint;
no_of_trispaces++;
if(no_of_trispaces >= max_no_of_trispaces)end_program(1167);
// Store zero. The phase is assumed to be already in place.
while(ia != ib)
  {
  baseb_fit[2*ia  ]=0;
  baseb_fit[2*ia+1]=0;
  baseb_detect_flag[ia]=DET_KEY_UP;
  ia=(ia+1)&baseband_mask;
  }
}



float check_dash(float pos)
{
int ia, ib, ja, sw;
float c0, t1, t2, r1, r2, f1, f2;
float p,err1,err2;
// Place a dash at pos using the phase/amplitude information 
// in baseb_envelope.
// Calculate the transformation required for optimum similarity
// between the observed waveform in baseb and the dash we have stored.
err1=0;
err2=0;
p=0;
ib=pos;
if( pos-ib < 0.001)pos+=0.001; 
ia=(ib+1)&baseband_mask;
if( (dash_pts&1) == 0)
  {
  c0=0.5*(dash_pts-1);
  }
else
  {
  c0=(dash_pts>>1)+1;
  }  
// pos is the center of where we want to place a dash.
sw=0;
while( sw!=2)
  {
  sw=0;
  t1=ia-pos;
  if(t1<0)t1+=baseband_size;
  t1=c0+t1;
  if(t1 >= dash_pts-1)
    {
    baseb_fit[2*ia  ]=0;
    baseb_fit[2*ia+1]=0;
    sw++;
    }
  else
    {
    ja=t1;
    t1-=ja;
    t2=1-t1;
    f1=t2*dash_waveform[2*ja  ]+t1*dash_waveform[2*ja+2];
    f2=t2*dash_waveform[2*ja+1]+t1*dash_waveform[2*ja+3];
    r1=baseb_envelope[2*ia  ];
    r2=baseb_envelope[2*ia+1];
    baseb_fit[2*ia  ]=r1*f1+r2*f2;
    baseb_fit[2*ia+1]=r1*f2-r2*f1;
    }
  err1+=baseb_totpwr[ia];
  p+=baseb_fit[2*ia  ]*baseb_fit[2*ia  ]+baseb_fit[2*ia+1]*baseb_fit[2*ia+1];
  err2+=(baseb_fit[2*ia  ]-baseb[2*ia  ])*(baseb_fit[2*ia  ]-baseb[2*ia  ])+
        (baseb_fit[2*ia+1]-baseb[2*ia+1])*(baseb_fit[2*ia+1]-baseb[2*ia+1]);
  t1=pos-ib;
  if(t1<0)t1+=baseband_size;
  t1=c0-t1;
  if(t1 < 1)
    {
    baseb_fit[2*ib  ]=0;
    baseb_fit[2*ib+1]=0;
    sw++;
    }
  else
    {
    ja=t1;
    t1-=ja;
    t2=1-t1;
    f1=t2*dash_waveform[2*ja  ]+t1*dash_waveform[2*ja+2];
    f2=t2*dash_waveform[2*ja+1]+t1*dash_waveform[2*ja+3];
    r1=baseb_envelope[2*ib  ];
    r2=baseb_envelope[2*ib+1];
    baseb_fit[2*ib  ]=r1*f1+r2*f2;
    baseb_fit[2*ib+1]=r1*f2-r2*f1;
    }
  err1+=baseb_totpwr[ib];
  p+=baseb_fit[2*ib  ]*baseb_fit[2*ib  ]+baseb_fit[2*ib+1]*baseb_fit[2*ib+1];
  err2+=(baseb_fit[2*ib  ]-baseb[2*ib  ])*(baseb_fit[2*ib  ]-baseb[2*ib  ])+
        (baseb_fit[2*ib+1]-baseb[2*ib+1])*(baseb_fit[2*ib+1]-baseb[2*ib+1]);
  ia=(ia+1)&baseband_mask;
  ib=(ib+baseband_mask)&baseband_mask;
  }
// In a perfect fit (noise free) we expect this result:
//
//       Case     err1     err2
//       dash      p        0
//       space     0        p
//
// Due to the presence of noise we expect this instead:
//
//       Case     err1     err2       err1-err2
//       dash     p+n        n           p
//       space     n        p+n         -p 
return (err1-err2)/p;
}    


float check_dot(float pos)
{
int ia, ib, ja, sw;
float c0, t1, t2, r1, r2, f1, f2;
float p,err1,err2;
// Place a dot at pos using the phase/amplitude information 
// in baseb_envelope.
// Calculate the transformation required for optimum similarity
// between the observed waveform in baseb and the dot we have stored.
err1=0;
err2=0;
p=0;
ib=pos;
if( pos-ib < 0.001)pos+=0.001; 
ia=(ib+1)&baseband_mask;
if( (dot_pts&1) == 0)
  {
  c0=0.5*(dot_pts-1);
  }
else
  {
  c0=(dot_pts>>1)+1;
  }  
// pos is the center of where we want to place a dot.
sw=0;
while( sw!=2)
  {
  sw=0;
  t1=ia-pos;
  if(t1<0)t1+=baseband_size;
  t1=c0+t1;
  if(t1 >= dot_pts-1)
    {
    baseb_fit[2*ia  ]=0;
    baseb_fit[2*ia+1]=0;
    sw++;
    }
  else
    {
    ja=t1;
    t1-=ja;
    t2=1-t1;
    f1=t2*dot_waveform[2*ja  ]+t1*dot_waveform[2*ja+2];
    f2=t2*dot_waveform[2*ja+1]+t1*dot_waveform[2*ja+3];
    r1=baseb_envelope[2*ia  ];
    r2=baseb_envelope[2*ia+1];
    baseb_fit[2*ia  ]=r1*f1+r2*f2;
    baseb_fit[2*ia+1]=r1*f2-r2*f1;
    }
  err1+=baseb_totpwr[ia];
  p+=baseb_fit[2*ia  ]*baseb_fit[2*ia  ]+baseb_fit[2*ia+1]*baseb_fit[2*ia+1];
  err2+=(baseb_fit[2*ia  ]-baseb[2*ia  ])*(baseb_fit[2*ia  ]-baseb[2*ia  ])+
        (baseb_fit[2*ia+1]-baseb[2*ia+1])*(baseb_fit[2*ia+1]-baseb[2*ia+1]);
  t1=pos-ib;
  if(t1<0)t1+=baseband_size;
  t1=c0-t1;
  if(t1 < 1)
    {
    baseb_fit[2*ib  ]=0;
    baseb_fit[2*ib+1]=0;
    sw++;
    }
  else
    {
    ja=t1;
    t1-=ja;
    t2=1-t1;
    f1=t2*dot_waveform[2*ja  ]+t1*dot_waveform[2*ja+2];
    f2=t2*dot_waveform[2*ja+1]+t1*dot_waveform[2*ja+3];
    r1=baseb_envelope[2*ib  ];
    r2=baseb_envelope[2*ib+1];
    baseb_fit[2*ib  ]=r1*f1+r2*f2;
    baseb_fit[2*ib+1]=r1*f2-r2*f1;
    }
  err1+=baseb_totpwr[ib];
  p+=baseb_fit[2*ib  ]*baseb_fit[2*ib  ]+baseb_fit[2*ib+1]*baseb_fit[2*ib+1];
  err2+=(baseb_fit[2*ib  ]-baseb[2*ib  ])*(baseb_fit[2*ib  ]-baseb[2*ib  ])+
        (baseb_fit[2*ib+1]-baseb[2*ib+1])*(baseb_fit[2*ib+1]-baseb[2*ib+1]);
  ia=(ia+1)&baseband_mask;
  ib=(ib+baseband_mask)&baseband_mask;
  }
// In a perfect fit (noise free) we expect this result:
//
//       Case     err1     err2
//       dot       p        0
//       space     0        p
//
// Due to the presence of noise we expect this instead:
//
//       Case     err1     err2       err1-err2
//       dot      p+n        n           p
//       space     n        p+n         -p 
return (err1-err2)/p;
}    




float short_region_relpower(int ia, int ib, int ss, int pp)
{
int nn,kk;
float t1,t2,t3,t4,limpwr;
// ia is the last point of a key down region.
// ib is the first point of the next key down region.
// Get the cosine of the carrier phase twist over the current region.
t3=baseb_carrier[2*ss  ]*baseb_carrier[2*pp  ]+
   baseb_carrier[2*ss+1]*baseb_carrier[2*pp+1];
// In case there is a large phase twist, possibly 180 degrees, the
// carrier had a null here. A null means there may be a dot here that
// we will be unable to detect by use of the average amplitude
// from both sides.
if(t3 < 0.5)return -0.3;
// Get the amplitude at both sides of the current region.
t1=baseb_envelope[2*pp  ]*baseb_envelope[2*pp  ]+
   baseb_envelope[2*pp+1]*baseb_envelope[2*pp+1];
t2=baseb_envelope[2*ss  ]*baseb_envelope[2*ss  ]+
   baseb_envelope[2*ss+1]*baseb_envelope[2*ss+1];
// Return if amplitude of signal differs by more than 6 dB across 
// the current region.
if(t1 < t2)
  {
  if(t1 < 0.25*t2 )return -0.1;
  limpwr=t2;
  }
else
  {
  if(t2 < 0.25*t1 )return -0.1;
  limpwr=t1;
  }
// Locate the largest and the second largest data points.
t3=0;
t4=0;
nn=(ia+1)&baseband_mask;
kk=nn;
while(nn != ib)
  {
  if(t3<baseb_totpwr[nn])
    {
    t4=t3;
    t3=baseb_totpwr[nn];
    kk=nn;
    }
  nn=(nn+1)&baseband_mask;
  }
// If the largest one is the first or the last point, use the
// second largest point. It may be adjacent to a key down region
// too, but just neglecting it also seems a bit risky.    
if(kk==((ia+1)&baseband_mask) || kk==((ib+baseband_mask)&baseband_mask))
  {
  t3=t4;
  }
t3/=dot_pwrmax;
refpwr=0.5*(t1+t2);
return t3/refpwr;
}

void store_dot(void)
{
int ir,ia,ib;
float t2,r2;
// We calculated the dot product between the current waveform and 
// the average waveform for a dash when cg_wave_start was computed.
// Normalise and use to set up fitted waveforms.
ia=cg_wave_midpoint-0.5*dot_pts+baseband_size;
ia&=baseband_mask;
ib=(ia+dot_pts)&baseband_mask;
t2=cg_wave_re;
r2=cg_wave_im;
dot_midpoint[no_of_dots]=cg_wave_midpoint;
no_of_dots++;
if(no_of_dots >= max_no_of_dots)end_program(1162);
// Store the average waveform at the optimum phase and amplitude giving
// the smallest residue when subtracted from baseb.
ir=0;
while(ia != ib)
  {
  baseb_fit[2*ia  ]=t2*dot_waveform[2*ir  ]+r2*dot_waveform[2*ir+1];
  baseb_fit[2*ia+1]=t2*dot_waveform[2*ir+1]-r2*dot_waveform[2*ir  ];
  baseb_envelope[2*ia  ]=t2;
  baseb_envelope[2*ia+1]=r2;
  baseb_detect_flag[ia]=DET_DOT;
  ir++;
  ia=(ia+1)&baseband_mask;
  }
}

void store_dash(void)
{
short int ir;
int ia,ib;
float t2,r2;
// We calculated the dot product between the current waveform and 
// the average waveform for a dash when cg_wave_start was computed.
// Normalise and use to set up fitted waveforms.
ia=cg_wave_midpoint-0.5*dash_pts+baseband_size;
ia&=baseband_mask;
ib=(ia+dash_pts)&baseband_mask;
t2=cg_wave_re;
r2=cg_wave_im;
// Store the average waveform at the optimum phase and amplitude giving
// the smallest residue when subtracted from baseb.
ir=0;
while(ia != ib)
  {
  baseb_fit[2*ia  ]=t2*dash_waveform[2*ir  ]+r2*dash_waveform[2*ir+1];
  baseb_fit[2*ia+1]=t2*dash_waveform[2*ir+1]-r2*dash_waveform[2*ir  ];
  baseb_envelope[2*ia  ]=t2;
  baseb_envelope[2*ia+1]=r2;
  ir++;
  ia=(ia+1)&baseband_mask;
  }
}

