

#include <vga.h>
#include <vgagl.h>
#include <string.h>
#include "globdef.h"
#include "uidef.h"
#include "screendef.h"
#include "vernr.h"

#define PGW 7*text_width
#define PGH 7*text_height  


void show_pol(void)
{
char s[80];
int angstep,i,x1,x2,y9,y2;
float major,minor,z1,z2,a1,a2,c1,c2,c3;
float pol_a,sina,pol_b;
// Make an ellipse to describe circular polarisation
i=pgbutt[PG_ANGLE].x2-pgbutt[PG_ANGLE].x1;
errno=40109;
ck_fillbox(pgbutt[PG_ANGLE].x1,pgbutt[PG_ANGLE].y1,i,i,130);
major=0.45*i;
// The complex amplitudes in our two channels are:
// re_x=cos(z)*cos(a)
// im_x=sin(z)*cos(a)
// re_y=sin(a)*(cos(z)*cos(w)-sin(z)*sin(w))
// im_y=sin(a)*(sin(z)*cos(w)+cos(z)*sin(w))
// z is the time function and a is the angle between main axis of
// the incoming wave and the x channel.
// If the antenna was twisted by the angle a, the signals would become:
// re_a=cos(a)*re_x+sin(a)*re_y
// im_a=cos(a)*im_x+sin(a)*im_y
// re_b=cos(a)*re_y-sin(a)*re_x
// im_b=cos(a)*im_y-sin(a)*im_x
// This would maximise the a signal and minimise the b signal.
// The power in the b channel is re_*re_b+im_b*im_b, and that is the
// function we want the minimum for over one cycle of z.
// re_b=cos(a)*sin(a)*(cos(z)*cos(w)-sin(z)*sin(w))-sin(a)*cos(z)*cos(a)
// im_b=cos(a)*sin(a)*(sin(z)*cos(w)+cos(z)*sin(w))-sin(a)*sin(z)*cos(a)
// re_b=sin(2*a)*(cos(z)*cos(w)-sin(z)*sin(w)-cos(z))/2
// im_b=sin(2*a)*(sin(z)*cos(w)+cos(z)*sin(w)-sin(z))/2
// The power in the b channel at minimum is:
// B=sin(2*a)^2*( [cos(z)*cos(w)-sin(z)*sin(w)-cos(z)]^2 +
//                [sin(z)*cos(w)+cos(z)*sin(w)-sin(z)]^2   )/4
// B=0.5*sin(2*a)^2*(1-cos(2*w) )
// For linear polarisation B=0 and for circular B=0.5
// Illustrate the polarisation with an ellipse.
// the axis power lengths are B and (1-B)
pol_a=acos(pg.c1);
if(pg.c3 == 0)
  {
  pg_b=0;
  }
else
  {
  pg_b=atan2(pg.c3,pg.c2); 
  }
if(fabs(1-pg.c1*pg.c1-pg.c2*pg.c2-pg.c3*pg.c3) >0.001)end_program(887999);
// Rotate the coordinate system by pg.angle
pol_b=pol_a-pg.angle*PI_L/180;
if(pol_b > PI_L)pol_b-=PI_L;
if(pol_b <0)pol_b+=PI_L;
c1=cos(pol_b);
c2=sin(pol_b)*cos(pg_b);
c3=sin(pol_b)*sin(pg_b);
z1=pow(sin(2*pol_a),2.)*0.5*(1-cos(2*pg_b));
minor=major*z1;
angstep=2*major/PI_L;
angstep*=sqrt(fabs(minor/major));
angstep++;
angstep&=0xfffffffe;
if(angstep < 2)angstep=2;
z2=2*PI_L/angstep;
z1=pol_a;
angstep++;
sina=sin(pol_b);
x1=pg_x0+major*c1;
y9=pg_y0-major*sina;
for(i=0; i<angstep; i++)
  {
  a1=major*cos(z1-pol_a);
  a2=minor*sin(z1-pol_a);
  x2=pg_x0+(a1*c1+a2*sina);
  y2=pg_y0+(a2*c1-a1*sina);
errno=10544;
  ck_line(x1,y9,x2,y2,7);
errno=-10544;

  x1=x2;
  y9=y2;
  z1+=z2;
  }
pg_pol_angle=180*pol_b/PI_L;
sprintf(s,"%3d",(int)(pg_pol_angle));
errno=20105;
ck_write(pgbutt[PG_CIRC].x1+text_width/2,pgbutt[PG_CIRC].y2+1,s);

errno=40109;
ck_fillbox(pgbutt[PG_CIRC].x1,pgbutt[PG_CIRC].y1,
           pgbutt[PG_CIRC].x2-pgbutt[PG_CIRC].x1,
           pgbutt[PG_CIRC].y2-pgbutt[PG_CIRC].y1,135);
if(pol_a < 0.5*PI_L)
  {
  a1=1;
  }
else
  {
  a1=-1;
  }
if(a1*pg_b > 0)
  {
  s[0]='R';
  }
else
  {
  s[0]='L';
  }
s[1]=0;
errno=20106;
ck_write(pg_x0-text_width/4,pg_y0-text_height/3,s); 
i=0.88*pg_b*(pgbutt[PG_CIRC].x2-pgbutt[PG_CIRC].x1)/PI_L;
errno=10545;
ck_line(pg_x0+i,pgbutt[PG_CIRC].y1,pg_x0+i,pgbutt[PG_CIRC].y2-1,15);



}

  
void check_pg_borders(void)
{
current_graph_minh=PGH;
current_graph_minw=PGW;
check_graph_placement((void*)(&pg));
set_graph_minwidth((void*)(&pg));
}




void help_on_pol_graph(void)
{
int msg_no;
int event_no;
// Set msg invalid in case we are not in any select area.
msg_no=-1;
for(event_no=0; event_no<MAX_PGBUTT; event_no++)
  {
  if( pgbutt[event_no].x1 <= mouse_x && 
      pgbutt[event_no].x2 >= mouse_x &&      
      pgbutt[event_no].y1 <= mouse_y && 
      pgbutt[event_no].y2 >= mouse_y) 
    {
    switch (event_no)
      {
      case PG_TOP:
      case PG_BOTTOM:
      case PG_LEFT:
      case PG_RIGHT:
      msg_no=101;
      break;

      case PG_AUTO:
      msg_no=54;
      break;
  
      case PG_ANGLE:
      msg_no=55;
      break;

      case PG_CIRC:
      msg_no=56;
      break;

      case PG_AVGNUM:
      msg_no=57;
      break;
      }
    }  
  }
help_message(msg_no);
}  




void mouse_continue_pol_graph(void)
{
int i, j;
float t1,t2;
hide_mouse();
switch (mouse_active_flag-1)
  {
  case PG_TOP:
  case PG_BOTTOM:
  case PG_LEFT:
  case PG_RIGHT:
  pg_borders(0);
  if(pg_oldx==-10000)
    {
    pg_oldx=mouse_x;
    pg_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-pg_oldx;
    j=mouse_y-pg_oldy;  
    pg_oldx=mouse_x;
    pg_oldy=mouse_y;
    pg.ytop+=j;
    pg.ybottom+=j;      
    pg.xleft+=i;
    pg.xright+=i;
    check_pg_borders();
    pg_borders(15);
    }
  break;
    

  default:
  goto await_release;
  }
if(leftpressed == 2)goto finish;
show_mouse();
return;
await_release:;
if(leftpressed == 1)
  {
  show_mouse();
  return;
  }
switch (mouse_active_flag-1)
  {
  case PG_AUTO:
  pg.adapt^=1;
  break;
  
  case PG_ANGLE:
  t1=mouse_x-pg_x0;
  t2=mouse_y-pg_y0;
  if(fabs(t1)+fabs(t2)>0)
    {
    t1=atan2(-t2,t1)+PI_L*pg.angle/180;
    }
  else
    {
    t1=0;
    }
// Polarisation plane is periodic in 180 degrees.
// Place angle in the interval 0 to 180 degrees so sin(angle) is
// always positive.     
  while(t1 < 0)t1+=PI_L;
  if(t1>=PI_L)t1-=PI_L;
  pg.c1=cos(t1);
  t2=sin(t1);
  pg.c2=sin(t1)*cos(pg_b);    
  pg.c3=sin(t1)*sin(pg_b);    
  break;

  case PG_CIRC:
  pg_b=1.2*PI_L*(mouse_x-pg_x0)/(pgbutt[PG_CIRC].x2-pgbutt[PG_CIRC].x1);
  if(pg_b > 0.499*PI_L)pg_b= 0.499*PI_L;
  if(pg_b < -0.499*PI_L)pg_b=-0.499*PI_L;
  if(pg.c1 == 0)pg.c1=0.0001;
  t2=sqrt(1-pg.c1*pg.c1);
  if(pg.c2 < 0)t2*=-1;
  if(t2 == 0)t2=0.0001;
  pg.c2=t2*cos(pg_b);  
  pg.c3=t2*sin(pg_b);  
  break;

  case PG_AVGNUM:
  pg.avg=mouse_x-pg_x0;
  break;
  }      
finish:;
leftpressed=0;  
mouse_active_flag=0;
for(i=0; i<no_of_scro; i++)
  {
  if(scro[i].no == POL_GRAPH)
    {
    pg_borders(0);
errno=40110;
    ck_fillbox(pg_old_x1,pg_old_y1,pg_old_x2-pg_old_x1,pg_old_y2-pg_old_y1,0);
    make_pol_graph(i);
    if(genparm[SECOND_FFT_ENABLE] != 0)new_hg_pol();
    pg_oldx=-10000;
    show_mouse();
    return;
    }
  }
end_program(854);  
}

void mouse_on_pol_graph(void)
{
int event_no;
// First find out is we are on a button or border line.
for(event_no=0; event_no<MAX_PGBUTT; event_no++)
  {
  if( pgbutt[event_no].x1 <= mouse_x && 
      pgbutt[event_no].x2 >= mouse_x &&      
      pgbutt[event_no].y1 <= mouse_y && 
      pgbutt[event_no].y2 >= mouse_y) 
    {
    pg_old_y1=pg.ytop;
    pg_old_y2=pg.ybottom;
    pg_old_x1=pg.xleft;
    pg_old_x2=pg.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_pol_graph;
    }
  }
// Not button or border.
// Do nothing.
}

void pg_borders(int color)
{
errno=10546;
ck_hline(pg.xleft,pg.ytop,pg.xright,color);
errno=10547;
ck_hline(pg.xleft,pg.ybottom,pg.xright,color);
errno=10548;
ck_line(pg.xleft,pg.ytop,pg.xleft,pg.ybottom,color);
errno=10549;
ck_line(pg.xright,pg.ytop,pg.xright,pg.ybottom,color);
}

void make_pol_graph(int m)
{
int i, j;
char *adafix[]={"Adapt","Fixed"};
check_pg_borders();
scro[m].no=POL_GRAPH;
scro[m].x1=pg.xleft;
scro[m].x2=pg.xright;
scro[m].y1=pg.ytop;
scro[m].y2=pg.ybottom;
pgbutt[PG_LEFT].x1=pg.xleft;
pgbutt[PG_LEFT].x2=pg.xleft;
pgbutt[PG_LEFT].y1=pg.ytop;
pgbutt[PG_LEFT].y2=pg.ybottom;
pgbutt[PG_RIGHT].x1=pg.xright;
pgbutt[PG_RIGHT].x2=pg.xright;
pgbutt[PG_RIGHT].y1=pg.ytop;
pgbutt[PG_RIGHT].y2=pg.ybottom;
pgbutt[PG_TOP].x1=pg.xleft;
pgbutt[PG_TOP].x2=pg.xright;
pgbutt[PG_TOP].y1=pg.ytop;
pgbutt[PG_TOP].y2=pg.ytop;
pgbutt[PG_BOTTOM].x1=pg.xleft;
pgbutt[PG_BOTTOM].x2=pg.xright;
pgbutt[PG_BOTTOM].y1=pg.ybottom;
pgbutt[PG_BOTTOM].y2=pg.ybottom;
// Draw the border lines
pg_borders(7);
settextcolor(7);
pg_oldx=-10000;
// Place the Adapt/Fixed button in lower left corner
i=1+text_width/2;
pgbutt[PG_AUTO].x1=pg.xleft+i;
pgbutt[PG_AUTO].x2=pgbutt[PG_AUTO].x1+5*text_width+i;
pgbutt[PG_AUTO].y2=pg.ybottom-2;
pgbutt[PG_AUTO].y1=pgbutt[PG_AUTO].y2-text_width-i;
errno=20107;
ck_write(pgbutt[PG_AUTO].x1+i,pgbutt[PG_AUTO].y1+i/2,adafix[pg.adapt]);
errno=10550;
ck_hline(pgbutt[PG_AUTO].x1,pgbutt[PG_AUTO].y1,pgbutt[PG_AUTO].x2,7);
errno=10551;
ck_hline(pgbutt[PG_AUTO].x1,pgbutt[PG_AUTO].y2,pgbutt[PG_AUTO].x2,7);
errno=10552;
ck_line(pgbutt[PG_AUTO].x1,pgbutt[PG_AUTO].y1,
        pgbutt[PG_AUTO].x1,pgbutt[PG_AUTO].y2,7);
errno=10553;
ck_line(pgbutt[PG_AUTO].x2,pgbutt[PG_AUTO].y1,
        pgbutt[PG_AUTO].x2,pgbutt[PG_AUTO].y2,7);
// Control for major axis angle
pg_x0=(pg.xleft+pg.xright)/2;
i=pg.xright-pg.xleft-2;
j=pg.ybottom-pg.ytop-3*text_height;
if(i>j)i=j;
i/=2;
pg_y0=pg.ytop+i+2;
pgbutt[PG_ANGLE].x1=pg_x0-i;
pgbutt[PG_ANGLE].x2=pg_x0+i;
pgbutt[PG_ANGLE].y1=pg_y0-i;
pgbutt[PG_ANGLE].y2=pg_y0+i;
// Control for interchannel phase (circular / linear)
pgbutt[PG_CIRC].x1=pgbutt[PG_ANGLE].x1;
pgbutt[PG_CIRC].x2=pgbutt[PG_ANGLE].x2;
pgbutt[PG_CIRC].y1=pgbutt[PG_ANGLE].y2+text_height/2;
pgbutt[PG_CIRC].y2=pgbutt[PG_CIRC].y1+text_height/2;
// Control for averaging parameter.
pgbutt[PG_AVGNUM].x1=pg_x0+1;
pgbutt[PG_AVGNUM].x2=pgbutt[PG_ANGLE].x2;
pgbutt[PG_AVGNUM].y1=pgbutt[PG_CIRC].y2+1;
pgbutt[PG_AVGNUM].y2=pgbutt[PG_CIRC].y2+text_height-2;
if(pg.adapt == 0)
  {
  i=pgbutt[PG_AVGNUM].x2-pgbutt[PG_AVGNUM].x1-1;
  if(pg.avg>i)pg.avg=i;
  i=pgbutt[PG_AVGNUM].x1+pg.avg;
errno=40111;
  ck_fillbox(pgbutt[PG_AVGNUM].x1,pgbutt[PG_AVGNUM].y1,
             pgbutt[PG_AVGNUM].x2-pgbutt[PG_AVGNUM].x1,
             pgbutt[PG_AVGNUM].y2-pgbutt[PG_AVGNUM].y1,135);
errno=10554;
  ck_line(i,pgbutt[PG_AVGNUM].y1,i,pgbutt[PG_AVGNUM].y2,15);
  }
show_pol();
dpg.xleft=pg.xleft;        
dpg.xright=pg.xright;        
dpg.ytop=pg.ytop;        
dpg.ybottom=pg.ybottom;        
make_modepar_file(GRAPHTYPE_PG);
}

void select_pol_default(void)
{
char s[80];
int iang,isign,iaut,istrt;
float t1;
char sgn[]={'+','-'};
char *amode[]={"Auto","Man"};
char sta[]={'H','V'};
iang=0;
isign=0;
iaut=0;
istrt=0;
dpg.avg=8;
pol_menu:;
vga_clear();
settextcolor(14);
xytext(5,5,"INIT POLARISATION PARAMETERS FOR TWO CHANNEL RADIO");
settextcolor(7);
sprintf(s,"A => Angle for channel 1 (0=Hor, 90=vert)  %d",iang);
xytext(5,8,s); 
sprintf(s,"B => Toggle sign for channel 2             %c",sgn[isign]);
xytext(5,9,s); 
sprintf(s,"C => Toggle mode                           %s",amode[iaut]);
xytext(5,10,s); 
sprintf(s,"D => Toggle start polarisation             %c",sta[istrt]);
xytext(5,11,s); 
xytext(5,13,"S => Save to disk and exit");
await_keyboard();
clean_vga_inkey();
switch (vga_inkey)
  {
  case 'A':
  xytext(10,16,"Angle:");
  iang=vga_get_integer(16, 16, 4, 0, 90);
  break;

  case 'B':
  isign^=1;
  break;

  case 'C':
  iaut^=1;
  break;

  case 'D':
  istrt^=1;
  break;

  case 'S':
  dpg.angle=iang;
  if(isign != 0)dpg.angle+=180;
  dpg.adapt=iaut;
// Get polarisation coefficients for start polarisation    
  t1=dpg.angle;
  if(istrt != 0)t1+=90;
  t1*=PI_L/180;
  dpg.c1=cos(t1);
  dpg.c2=sin(t1);
  dpg.c3=0;
  dpg.check=PG_VERNR;
  make_modepar_file(GRAPHTYPE_PG);
  return;
  }
goto pol_menu;    
}


void init_pol_graph(void)
{
float t1;
if (read_modepar_file(GRAPHTYPE_PG) == 0)
  {
pg_default:;
  dpg.xleft=5;
  dpg.xright=5+PGW;
  dpg.ybottom=HEIGHT-4*text_height;
  dpg.ytop=dpg.ybottom-PGH;
  select_pol_default();
  lir_status=LIR_NEW_POL;
  }
if(dpg.check != PG_VERNR)goto pg_default;
if(dpg.avg < 1 || dpg.avg > 1000)goto pg_default;
t1=dpg.c1*dpg.c1+dpg.c2*dpg.c2+dpg.c3*dpg.c3;
if(t1 < 0.98 || t1 > 1.02)goto pg_default;
pg=dpg;
pg_oldx=-10000;
pg_flag=1;
make_pol_graph(no_of_scro);
no_of_scro++;
if(no_of_scro >= MAX_SCRO)end_program(89);
}
