
#include <vga.h>
#include <vgagl.h>
#include <unistd.h>
#include <fcntl.h>

#include "globdef.h"
#include "uidef.h"
#include "sigdef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "conf.h"

#define MAX_DEVNAMES 65

#define ABOVE_MAX_SPEED 768000
#if SND_DEV == 1
#define DEVNAME_SIZE 12
char devname[DEVNAME_SIZE]="/dev/dsp   ";
#define DEVNAMPOS 8
#else 
#define DEVNAME_SIZE 18
char devname[DEVNAME_SIZE]="/dev/sound/dsp   ";
#define DEVNAMPOS 14
#endif


char *audiomode_text[4]={"One rx channel (normal audio)",
                         "One rx channel I/Q (direct conversion rx)",
                         "Two rx channels (adaptive polarisation/phasing)",
                         "Two rx channels I/Q"};

void flush_sync(void)
{
fflush(NULL);
sync();
}

void make_devname(int n)
{
int k,num,m;
if(n < 0)end_program(9235);
num=n;
k=DEVNAMPOS;
num--;  
if(num>MAX_DEVNAMES)end_program(9236);
m=num/10;
if(m > 0)
  {
  devname[k]=m+'0';
  k++;
  }
devname[k]=num-m*10+'0';
if(devname[k]<'0')devname[k]=0;
k++;
devname[k]=0;  
}

void set_da_parms(void)
{
int i;
char s[128];
ui.min_output_channels=1;
ui.max_output_channels=2;
if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &ui.max_output_channels) == -1)
                                                     ui.max_output_channels=-1;
if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &ui.min_output_channels) == -1)
                                ui.min_output_channels=ui.max_output_channels;
if(ui.min_output_channels == -1)end_program(1069);
if(ioctl(audio_out, SNDCTL_DSP_GETFMTS, &i) == -1)end_program(1071);
ui.max_output_bytes=1;
ui.min_output_bytes=1;
if(i&AFMT_S16_LE)ui.max_output_bytes=2;
ui.max_output_speed=ABOVE_MAX_SPEED;
ui.min_output_speed=1;
if(ioctl(audio_out, SNDCTL_DSP_SPEED, &ui.max_output_speed) == -1)
                                                             end_program(1072);
if(ioctl(audio_out, SNDCTL_DSP_SPEED, &ui.min_output_speed) == -1)
                                                             end_program(1073);
if(ui.max_output_speed==ABOVE_MAX_SPEED || 
   ui.min_output_speed <= 1 ||
   ui.max_output_speed < ui.min_output_speed)
  {
  xytext(0,7,"Unable to verify output speed. Limits unknown");
  genparm[DA_OUTPUT_SPEED]=ui.input_speed;
  ui.max_output_speed=48000;
  ui.min_output_speed=5000;
  }
else
  {
  sprintf(s,"\nOutput speed:%.3f to %.3f kHz",0.001*(float)(ui.min_output_speed),
                                           0.001*(float)(ui.max_output_speed));
  xytext(0,7,s);
  }
genparm[DA_OUTPUT_SPEED]=ui.input_speed;
}

void close_audio_input(void)
{
if ( audio_in != -1)
  {
  if(audio_out == audio_in)close_audio_output();
  if(close(audio_in) == -1)end_program(1005);
  if(ui.ad_device_no > 255)if(close(audio_in2) == -1)end_program(1173);
  }
audio_in=-1;
}

int make_dafrag(void)
{
int i, frag;
frag=5+da_output_channels+da_output_bytes;
i=genparm[DA_OUTPUT_SPEED];
while(i>5500)
  {
  frag++;
  i/=2;
  } 
frag|=0x7fff0000;
return frag;
}

void close_audio_output(void)
{
if ( audio_out != -1)
  {
  if( da_wrbuf != NULL)
    {
    free(da_wrbuf);
    da_wrbuf=NULL;
    }
  if ( audio_out != audio_in)
    {
    if(ioctl(audio_out,SNDCTL_DSP_RESET,0)==-1)end_program(1026);
    if(close(audio_out)==-1)end_program(1006);
    }
  }  
audio_out=-1;
}

void open_audio_input(void)
{
float t1;
int i, j;
audio_buf_info ad_info;
// Clear scratch area used by some fft routines.
clear_real2iq();
if(disksave_flag == 2 || network_flag == 2)return;
if( (ui.input_mode&NO_DUPLEX) != 0 && audio_out != -1)return;
if(  (audio_in) != -1)end_program(1000);
make_devname(ui.ad_device_no&255);
audio_in=open( devname ,ui.ad_device_mode , 0);
if(audio_in == -1)end_program(1007);
if(ui.ad_device_no > 255)
  {
  i=ui.ad_device_no/255-1;
  make_devname(i);
  audio_in2=open( devname ,ui.ad_device_mode , 0);
  if(audio_in2 == -1)end_program(1174);
  }
if(ui.ad_frag != 0)
  {
  i=ui.ad_frag;
  i=i|0x7fff0000;
  if(ioctl(audio_in, SNDCTL_DSP_SETFRAGMENT, &i) == -1)end_program(1008);  
  if(ui.ad_device_no > 255)
    {
    if(ioctl(audio_in2, SNDCTL_DSP_SETFRAGMENT, &i) == -1)end_program(1175);  
    }
  }
i=AFMT_S16_LE;
#ifdef AFMT_S32_LE
if( (ui.input_mode&DWORD_INPUT) != 0)i=AFMT_S32_LE;
#endif
j=i;
if(ioctl(audio_in, SNDCTL_DSP_SETFMT, &i) == -1)end_program(1013);
if(i!=j)end_program(1014);
if(ui.ad_device_no > 255)
  {
  j=i;
  if(ioctl(audio_in2, SNDCTL_DSP_SETFMT, &i) == -1)end_program(1176);
  if(i!=j)end_program(1177);
  }
// Set number of ad_channels
j=ui.ad_channels;
if(ui.ad_device_no > 255)j=2;
i=j;
if(ioctl(audio_in, SNDCTL_DSP_CHANNELS, &i) == -1)end_program(1009);  
if(i != j)end_program(1010);
if(ui.ad_device_no > 255)
  {
  i=j;
  if(ioctl(audio_in2, SNDCTL_DSP_CHANNELS, &i) == -1)end_program(1178);  
  if(i != j)end_program(1179);
  }
i=ui.input_speed;
if(ioctl(audio_in, SNDCTL_DSP_SPEED, &i)==-1)end_program(1011);
t1=(float)(i)/ui.input_speed;
if(t1 > 1.05 || t1 < 0.95)end_program(1012);
ui.input_speed=i;
if(ui.ad_device_no > 255)
  {
  if(ioctl(audio_in2, SNDCTL_DSP_SPEED, &i)==-1)end_program(1180);
  if(i != ui.input_speed)end_program(1181);
// Now that both devices are configured to be equal, synchronize them.

  i=0;
  if(ioctl(audio_in,SNDCTL_DSP_SETTRIGGER,&i)==-1)end_program(1183);
  i=0;
  if(ioctl(audio_in2,SNDCTL_DSP_SETTRIGGER,&i)==-1)end_program(1184);
  i=PCM_ENABLE_INPUT;
  if(ioctl(audio_in,SNDCTL_DSP_SETTRIGGER,&i)==-1)end_program(1183);
  i=PCM_ENABLE_INPUT;
  if(ioctl(audio_in2,SNDCTL_DSP_SETTRIGGER,&i)==-1)end_program(1184); 

  }
if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1109);
overrun_limit=ad_info.fragstotal-1;
if(ui.ad_device_no > 255)
  {
  if(ioctl(audio_in2, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1109);
  if(overrun_limit != ad_info.fragstotal-1)end_program(1109);
  }
}  



void open_audio_output(void)
{
int i,j;
int old_disksave_flag, old_frag, old_mode, old_chan, old_speed;
float t1,t2;
int frag;
if(  (audio_out) != -1)end_program(1015);
if( (ui.input_mode&NO_DUPLEX) != 0 && audio_in != -1) return;
if(ui.da_device_no == -1)
  {
  if(disksave_flag >=2)
    {
    if(audio_in != -1)end_program(9865);
    make_devname(ui.ad_device_no&255);
    audio_out=open( devname ,O_WRONLY , 0);
    if(audio_out == -1)end_program(1086);
    goto set_da;
    }
  else
    {
    return;
    }
  }
// *******************************************
if(ui.ad_device_no == ui.da_device_no && 
   ui.ad_device_mode == O_RDWR &&
   network_flag != 2)
  {
  old_disksave_flag=disksave_flag;
  disksave_flag=0;
  old_frag=ui.ad_frag;
  ui.ad_frag=make_dafrag();
  old_mode=ui.input_mode;
  ui.input_mode=0;
  old_chan=ui.ad_channels;
  ui.ad_channels=da_output_channels;
  old_speed=ui.input_speed;
  ui.input_speed=genparm[DA_OUTPUT_SPEED];
  if(audio_in == -1)open_audio_input();
  audio_out=audio_in;
  if(ioctl(audio_out, SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1147);
  daout_block=da_info.fragsize;
  disksave_flag=old_disksave_flag;
  ui.ad_frag=old_frag;
  ui.input_mode=old_mode;
  ui.ad_channels=old_chan;
  ui.input_speed=old_speed;
  }
else
  {
  make_devname(ui.da_device_no);
  audio_out=open( devname ,O_WRONLY|O_NONBLOCK , 0);
  if(audio_out == -1)end_program(1017);
set_da:;
// Set a small fragment size so we can stop playing quickly.
// One channel, one byte gives frag=7 (128 bytes = 39Hz interrupt freq at 5kHz) 
// Too large fragments cause delay when stopping audio output because the
// DSP_RESET command waits for the current fragment to complete before
// it clears the buffer. Waiting would cause overrun errors.
  frag=make_dafrag();
  if(ioctl(audio_out,SNDCTL_DSP_SETFRAGMENT,&frag)==-1)end_program(1018);
  i=da_output_channels;
  if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &i) == -1)end_program(1019);  
  if(i != da_output_channels)end_program(1099);
  if(da_output_bytes == 1)
    {
    i=AFMT_U8;
    }
  else
    {
    i=AFMT_S16_LE;
    }
  j=i;    
  if(ioctl(audio_out, SNDCTL_DSP_SETFMT, &i) == -1)end_program(1020);
  if(i != j)end_program(1021);
  i=genparm[DA_OUTPUT_SPEED];
  if(i > 0)
    {
    if(ioctl(audio_out, SNDCTL_DSP_SPEED, &i)==-1)end_program(1022);
    t1=(float)(i)/genparm[DA_OUTPUT_SPEED];
    if(t1 > 1.05 || t1 < 0.95)end_program(1023);
    genparm[DA_OUTPUT_SPEED]=i;
    }
// The fragment size we asked for (and probably get back here) should
// give an interrupt rate in the order of 50Hz.
// Only when really looking for fast response there is a reason to write
// small blocks.
// Set daout_block for a delay of 100 to 200 milliseconds if transform
// delay is above 500 milliseconds.
// Otherwise make sure delay is at least 50 ms.
  if(ioctl(audio_out, SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1029);
  daout_block=da_info.fragsize;
  t2=da_info.fragsize;
  t2/=da_output_channels*da_output_bytes*genparm[DA_OUTPUT_SPEED];
  while( t2 < 0.05)
    {
    t2*=2;
    daout_block*=2;
    }
  if(genparm[SECOND_FFT_ENABLE] == 0)
    {
    t1=fft1_size;
    }
  else
    {  
    t1=fft2_size;
    }
  t1/=timf1_sampling_speed;
  if(t1 > 0.5)  
    {
    while( t2 < 0.1)
      {
      t2*=2;
      daout_block*=2;
      } 
    }
  }
da_wrbuf=chk_malloc(daout_block,1025);
}

int devmodes[3]={O_RDONLY,O_WRONLY,O_RDWR};
char *devmode_txt[3]={"RDONLY","WRONLY","RDWR"};

void set_analog_io(void)
{
int dev_flag[MAX_DEVNAMES];
int dev_rd_channels[MAX_DEVNAMES];
int dev_rd_speed[MAX_DEVNAMES];
int dev_rd_bits[MAX_DEVNAMES];
int dev_wr_channels[MAX_DEVNAMES];
int dev_wr_speed[MAX_DEVNAMES];
int dev_wr_bits[MAX_DEVNAMES];
int dev_rdwr_channels[MAX_DEVNAMES];
int dev_rdwr_speed[MAX_DEVNAMES];
int dev_rdwr_bits[MAX_DEVNAMES];
int speed_warning;
int i, j, k, n, nn, cn, m, min, max, olbytes;
int rdwr_channels, caps;  
int trySpeed, deltaSpeed;
int rdwr_speed;
int rdwr_fmt;
int line, maxline, column, maxcolumn, bottom_line;
int mode, device_no;
audio_buf_info ad_info;
char *testbuff;
char ss[1024];
char s[256];
char color;
FILE *sndlog;
struct timeval tim1,tim2;
float t1;
char *logfile_name="soundboard_init.log";
color=7;
speed_warning=0;
vga_clear();
sndlog = fopen(logfile_name, "w");
if(sndlog == NULL)end_program(1016);
// First step through all devices and check which ones
// we can open in read, write and read-write mode.
maxline=HEIGHT/text_height-1;
maxcolumn=WIDTH/text_width-1;
bottom_line=0;
settextcolor(14);
xytext(10,0,"Checking device drivers");
settextcolor(7);
testbuff = chk_malloc(2048,1044);
n=0;
column=0;
line=1;
for(device_no=0; device_no<MAX_DEVNAMES; device_no++)
  {
  dev_flag[device_no]=0;
  make_devname(device_no);
  fprintf(sndlog,"Checking %s\n",devname);
  flush_sync();
  for(mode=0; mode<3; mode++)
    {
    k=1<<mode;
    sprintf(s,
    "Trying to open %s %s. Device defective if system hangs here. Read %s   ",
    devname, devmode_txt[mode],logfile_name);
    flush_sync();
    settextcolor(12);
    xytext(0,line+1,s);
    settextcolor(7);
    audio_out=open( devname, devmodes[mode]|O_NONBLOCK, 0);
    if(audio_out != -1)
      {
      for(nn=0; nn<80; nn++)fprintf(sndlog,"-");
      fprintf(sndlog,"\n%s opened as %s ",devname, devmode_txt[mode]);
      flush_sync();
      rdwr_fmt=16;
      rdwr_channels=0;
// Check for 16 bits
      if(ioctl(audio_out, SNDCTL_DSP_GETFMTS, &m) == -1)
        {
        fprintf(sndlog,"\nioctl(GETFMTS)  failed");
baddev:;
        flush_sync();
        k=0;
        goto cls;
        }
      if( (m&AFMT_S16_LE) == 0)
        {
        fprintf(sndlog,"\n16bit format NOT supported");
        k=0;
        }
      else
        {  
        fprintf(sndlog,"\n16bit format supported");
        }
#ifdef AFMT_S32_LE
// Check for 32 bits
      if( (m&AFMT_S32_LE) != 0)
        {
        rdwr_fmt=32;
        k=1<<mode;
        fprintf(sndlog,"\n32bit format supported");
        }
      else
        {
        fprintf(sndlog,"\n32bit format NOT supported");
        }          
#endif   
      flush_sync();
      if(k == 0)goto cls;
// Find max no of channels.
// We never want more than 4 input channels!!
      rdwr_channels=4;
      find_channels:;
      m=rdwr_channels;
      if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &m) == -1 || rdwr_channels != m)
        {
        rdwr_channels/=2;
        if(rdwr_channels == 0)
          {
          fprintf(sndlog,"\nMax no of channels = 0");
          goto baddev;
          }
        goto find_channels;
        }
      else
        {
        fprintf(sndlog,"\nMax no of channels = %d",m);
        }  
      flush_sync();
// Now find maximum input speed
      rdwr_speed=ABOVE_MAX_SPEED;
      if(ioctl(audio_out, SNDCTL_DSP_SPEED, &rdwr_speed)==-1)
        {
        fprintf(sndlog,"\nioctl(SPEED)  failed (max)");
        goto baddev;
        }
      fprintf(sndlog,"\nMax input speed %d Hz",rdwr_speed);
      flush_sync();
      if(mode != 1)
        {
        if(ioctl(audio_out, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
          {
          fprintf(sndlog,"\nioctl(GETISPACE)  failed");
          goto baddev;
          }
// If it is opened RDWR, check we have a chance to write while reading
// Some devices allow write or read in rdwr mode - but not simultaneously.
        if(mode == 2)
          {
          read(audio_out, testbuff, 128);
          if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
            {
            fprintf(sndlog,"\nioctl(GETOSPACE)  failed while reading");
            goto baddev;
            }
          }
        }
cls:;
      if(close(audio_out)==-1)end_program(1037);  
      if(k != 0)
        {
        if(k != 2)
          {
          ui.ad_device_no=device_no;
          n++;
          }
        dev_flag[device_no]|=k;
        switch (mode)
          {
          case 0:
          dev_rd_channels[device_no]=rdwr_channels;
          dev_rd_speed[device_no]=rdwr_speed;
          dev_rd_bits[device_no]=rdwr_fmt;
          break;
          
          case 1:
          dev_wr_channels[device_no]=rdwr_channels;
          dev_wr_speed[device_no]=rdwr_speed;
          dev_wr_bits[device_no]=rdwr_fmt;
          break;
          
          case 2:
          dev_rdwr_channels[device_no]=rdwr_channels;
          dev_rdwr_speed[device_no]=rdwr_speed;
          dev_rdwr_bits[device_no]=rdwr_fmt;
          break;
          }  
        }
      fprintf(sndlog,"\n");
      }          
    }      
  if(dev_flag[device_no] != 0)
    {
  fprintf(sndlog,"\nFLAG= %d",dev_flag[device_no]);


// One device may be different hardwares for input and output.
// Put them on different lines if characteristics differ.    
    k=0;
    j=0;
// Store equal pairs in k
// bit 0=1 => READ==WRITE
// bit 1=1 => READ==RDWR
// bit 2=1 => WRITE==RDWR
    if( (dev_flag[device_no]&3) == 3)
      {
      j|=1;
      if(dev_rd_channels[device_no]==dev_wr_channels[device_no] &&
         dev_rd_speed[device_no]==dev_wr_speed[device_no] &&
         dev_rd_bits[device_no]==dev_wr_bits[device_no]) k|=1;
      } 
    if( (dev_flag[device_no]&5) == 5)
      {
      j|=2;
      if(dev_rd_channels[device_no]==dev_rdwr_channels[device_no] &&
         dev_rd_speed[device_no]==dev_rdwr_speed[device_no] &&
         dev_rd_bits[device_no]==dev_rdwr_bits[device_no])k|=2;
      }
    if( (dev_flag[device_no]&6) == 6)
      {
      j|=4;
      if(dev_wr_channels[device_no]==dev_rdwr_channels[device_no] &&
         dev_wr_speed[device_no]==dev_rdwr_speed[device_no] &&
         dev_wr_bits[device_no]==dev_rdwr_bits[device_no])k|=4;
      }
  fprintf(sndlog,"\nj=%d   k=%d",j,k);
    if(j == k)
      {
// There are no differences between devices. Put everything on one line.
      m=1;
      if( (dev_flag[device_no]&1) == 1)
        {
        sprintf(ss,"\n%02d:%12s  %7d Hz   %d Chan.   %d bit",
                    device_no,devname,dev_rd_speed[device_no], 
                    dev_rd_channels[device_no], 
                    dev_rd_bits[device_no]);
        }
      else
        {  
        if( (dev_flag[device_no]&2) == 2)
          {
          sprintf(ss,"\n%02d:%12s  %7d Hz   %d Chan.   %d bit",
                     device_no,devname,dev_wr_speed[device_no], 
                     dev_wr_channels[device_no], 
                     dev_wr_bits[device_no]);
          }
        }
      for(j=0; j<3; j++)
        {
        k=1<<j;  
        if( (dev_flag[device_no]&k) != 0)
          {
          column=0;
          while(ss[column]!=0)column++;
          sprintf(&ss[column],"  %s",devmode_txt[j]);
          }
        }  
      if( (dev_flag[device_no]&5) == 0)
        {
        color=0x17;
        }
      else
        {
        color=0x1c;
        }
      }
    else
      {
// There are differences.
// We can not arrive here unless there are at least two modes
// and consequently j can not be zero.
      if( (j&1) == 0)
        {
// We do not have both read and write so it is one of them plus rdwr.
// This does not seem right, but this code although untested will hopefully
// give meaningful information in case we ever arrive here.
end_program(86765);

        }
      else
        {
// There is both read and write.
        if( (dev_flag[device_no]&4) == 0)
          {
// but there is no rdwr
          m=2;
          color=0x17;
          sprintf(&ss[maxcolumn],
                     "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                      device_no,devname,dev_wr_speed[device_no], 
                      dev_wr_channels[device_no], 
                      dev_wr_bits[device_no],devmode_txt[1]);
          sprintf(ss,"\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                      device_no,devname,dev_rd_speed[device_no], 
                      dev_rd_channels[device_no], 
                      dev_rd_bits[device_no],devmode_txt[0]);
          }
        else
          {  
// There is rdwr, read and write.
          if(k == 0)
            {
// They are all different
            m=3;
            color=0x17;
            sprintf(&ss[2*maxcolumn],
                  "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                      device_no,devname,dev_wr_speed[device_no], 
                      dev_wr_channels[device_no], 
                      dev_wr_bits[device_no],devmode_txt[1]);
            sprintf(&ss[maxcolumn],
                  "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                      device_no,devname,dev_rd_speed[device_no], 
                      dev_rd_channels[device_no], 
                      dev_rd_bits[device_no],devmode_txt[0]);
            sprintf(ss,
                  "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                      device_no,devname,dev_rdwr_speed[device_no], 
                      dev_rdwr_channels[device_no], 
                      dev_rdwr_bits[device_no],devmode_txt[2]);
            }
          else
            {
// rdwr equals read or write.
            m=2;
            if( (k&4) == 0)
              {
              color=0x17;
              sprintf(&ss[maxcolumn],
                        "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                        device_no,devname,dev_wr_speed[device_no], 
                        dev_wr_channels[device_no], 
                        dev_wr_bits[device_no],devmode_txt[1]);
              sprintf(ss,
                  "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s  %s",
                          device_no,devname,dev_rd_speed[device_no], 
                        dev_rd_channels[device_no], dev_rd_bits[device_no],
                            devmode_txt[0],devmode_txt[2]);
              }
            else
              {
              color=0x1d;
              sprintf(&ss[maxcolumn],
                        "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s  %s",
                        device_no,devname,dev_wr_speed[device_no], 
                        dev_wr_channels[device_no], dev_wr_bits[device_no],
                        devmode_txt[1],devmode_txt[2]);
              sprintf(ss,
                  "\n%02d:%12s  %7d Hz   %d Chan.   %d bit  %s",
                          device_no,devname,dev_rd_speed[device_no], 
                        dev_rd_channels[device_no], dev_rd_bits[device_no],
                            devmode_txt[0]);
              }
            }  
          }
        }
      }
    while (m >0)   
      {
      m--;
      line++;
      if(line >= maxline)
        {
        for(column=0; column<maxcolumn; column++)s[column]=' ';
        s[maxcolumn]=0;
        xytext(0,line,s);
        settextcolor(14);
        xytext(10,line,"(press ENTER to continue)");
        settextcolor(7);
        await_keyboard();
        xytext(0,line,s);
        bottom_line=maxline;
        line=1;
        }
      settextcolor(color);  
      column=0;
      while(ss[m*maxcolumn+column]!=0)column++;
      while(column < maxcolumn)
        {
        ss[m*maxcolumn+column]=' ';
        column++; 
        }
      ss[(m+1)*maxcolumn-1]=0;  
      xytext(0,line,&ss[m*maxcolumn]);
fprintf(sndlog,"\n%s    color=%d m=%d",&ss[m*maxcolumn],color,m);

      color=0x1c;
      }
fprintf(sndlog,"\n");
    }
  }
settextcolor(7);  

for(column=0; column<maxcolumn; column++)ss[column]=' ';
ss[maxcolumn]=0;
line++;
xytext(0,line,ss);
if(bottom_line > line)line=bottom_line;
if(n == 0)
  {
  if(line > maxline-12)line=maxline-12;
  xytext(0,line,"No sound device found (for 16/32 bit input).");
  xytext(10,line+2,"If you are using OSS:");
  xytext(0,line+3,"Run soundconf command to make sure OSS is correctly");
  xytext(0,line+4,"initialised for your sound board(s).");
  xytext(0,line+5,"Then start sound with soundon command.\n");
  xytext(10,line+7,"Without OSS:");
  xytext(0,line+8,"Find out how to configure your sound device drivers(s)");
  xytext(0,line+9,"Some distributions have the sndconfig command.");
  xytext(0,line+10,
        "Sound board drivers are usually configured during Linux install");
  xytext(20,line+12,"press any key");
  await_keyboard();  
  fprintf(sndlog,"\nNo input device found");
  goto setad_errexit;
  }
if(n > 1)
  {
  settextcolor(14);
  xytext(0,line,"Select device for input by first number on line");
get_addev:;
  settextcolor(7);
  ui.ad_device_no=vga_get_integer(48, line, 3, 0,MAX_DEVNAMES-1);
  if( (5 & dev_flag[ui.ad_device_no]) == 0)
    {
    settextcolor(15);
    xytext(40,line,"???");
    xytext(35,line+1,"ERROR (press any key)");
    await_keyboard();
    clear_lines(line+1,line+1);
    settextcolor(7);
    goto get_addev;
    }
  }
vga_clear();
make_devname(ui.ad_device_no);
fprintf(sndlog,"\n\n %s selected by user for input.",devname);
fprintf(sndlog,"   dev_flag=%d",dev_flag[ui.ad_device_no]);
flush_sync();
// High end users might need two devices to read four A/D channels.
// This is an option introduced April 2004 to allow two receive
// channels with the Lynx Two card (model A).
// Do not confuse other users by asking for a second input device
// unless the selected device has 32 bits but only allows 2 channels.
if( dev_rd_bits[ui.ad_device_no] == 32 && 
    dev_rd_channels[ui.ad_device_no]==2 &&
    (dev_flag[ui.ad_device_no]&1) != 0)
  {
  fprintf(sndlog,"\n\nUser selected 32 bit device with max two channels.");
  j=0;
  line=1;
  for(i=0; i<MAX_DEVNAMES; i++)
    {
    if(i != ui.ad_device_no && 
       dev_rd_speed[i] == dev_rd_speed[ui.ad_device_no] &&
       dev_rd_bits[i] == 32 &&
       (dev_flag[i]&1) != 0)
      {
// This device can also read 32 bits at the same speed.
// Check if it can be opened simultaneously in read mode.
      if(j==0)
        {
        xytext(5,10,"This device allows two channels only.");
        xytext(5,11,"Do you want to use a second device to run two");
        xytext(5,12,"receive channels with four audio channels ?");
        xytext(5,13,"(Y/N)?");
gt_dual:;
        vga_inkey=vga_getkey();
        clean_vga_inkey();
        if(vga_inkey == 'N')
          {
          fprintf(sndlog,"\nUser skipped dual devices"); 
          goto skip_dual_devices;
          }
        if(vga_inkey != 'Y')goto gt_dual;  
        vga_clear();
        fprintf(sndlog,"\nUser wants to try to open two read devices"); 
        make_devname(ui.ad_device_no);
        audio_in=open( devname, O_RDONLY, 0);
        if(audio_in == -1)end_program(1102);
// Check for DSP_CAP_TRIGGER capability
        if(ioctl(audio_in, SNDCTL_DSP_GETCAPS,&caps) == -1)end_program(1182);
        if( (caps&DSP_CAP_TRIGGER) == 0)
          {
          fprintf(sndlog,"\nDevice does not have DSP_CAP_TRIGGER"); 
          close(audio_in);
          goto skip_dual_devices;
          }
        j=2;  
        if(ioctl(audio_in, SNDCTL_DSP_CHANNELS, &j) == -1)end_program(1185);  
        if(j != 2)end_program(1186);
        }
      make_devname(i);
      audio_in2=open( devname, O_RDONLY|O_NONBLOCK, 0);
      if(audio_in2 != -1)
        {
        line++;
        fprintf(sndlog,"\n%s opened as second read device",devname);
        sprintf(s,"\n%02d:%12s  %7d Hz   %d Chan.   %d bit",
                i,devname,dev_rd_speed[i], dev_rd_channels[i], 
                dev_rd_bits[i]);
        xytext(0,line,s);
        close(audio_in2);
        dev_flag[i]|=256;
        }
      }
    }
  if(line == 1)goto skip_dual_devices;
  settextcolor(14);
  xytext(0,line+1,"Select second input device by first number on line");
get_addev2:;
  settextcolor(7);
  i=vga_get_integer(51, line+1, 3, 0,MAX_DEVNAMES-1);
  if( (256 & dev_flag[i]) == 0)
    {
    settextcolor(15);
    xytext(40,line,"???");
    xytext(35,line+1,"ERROR (press any key)");
    await_keyboard();
    clear_lines(line+1,line+1);
    settextcolor(7);
    goto get_addev2;
    }
// We store the second A/D device in the second byte of ad_device_no
  make_devname(i);
  ui.ad_device_no+=256*(i+1);
  for(i=0; i<MAX_DEVNAMES; i++)
    {
    dev_flag[i]&=255;
    }    
  audio_in2=open( devname, O_RDONLY|O_NONBLOCK, 0);
  if(audio_in2 == -1)end_program(1172);
// Check for DSP_CAP_TRIGGER capability
  if(ioctl(audio_in2, SNDCTL_DSP_GETCAPS,&caps) == -1)end_program(1182);
  if( (caps&DSP_CAP_TRIGGER) == 0)
    {
    fprintf(sndlog,"\nSecond input device does not have DSP_CAP_TRIGGER"); 
    close(audio_in);
    close(audio_in2);
    goto skip_dual_devices;
    }
  fprintf(sndlog,"\n %s is second input device.",devname);
  ui.input_mode=TWO_CHANNELS+IQ_DATA;
  ui.rx_channels=2;
  ui.ad_channels=4;
  vga_clear();
  goto continue_dual;
skip_dual_devices:;
  vga_clear();
  }
// Now that we decided what device to use for input, open it and let
// the user decide how to use it.
make_devname(ui.ad_device_no);
flush_sync();
ui.ad_device_mode=O_RDONLY;
mode=0;
if( (4 & dev_flag[ui.ad_device_no]) != 0)
  {
  if( (1 & dev_flag[ui.ad_device_no]) != 0)
    {
gt_rdo:;    
    sprintf(s,"Open %s as RDONLY or RDWR? (O/W)\n=>",devname);  
    xytext(10,10,s);
    vga_inkey=vga_getkey();
    clean_vga_inkey();
    if(vga_inkey == 'W')goto rdwr;
    if(vga_inkey != 'O')goto gt_rdo;
    fprintf(sndlog,"\nUser selected RDONLY");
    }
  else
    {
rdwr:;    
    ui.ad_device_mode=O_RDWR;
    mode=2;
    fprintf(sndlog,"\nUser selected RDWR");
    }
  }
else
  {
  fprintf(sndlog,"\nOnly RDONLY available");
  }
flush_sync();
vga_clear();  
sprintf(s,"For analog input: %s opened in %s mode",devname,devmode_txt[mode]);
xytext(0,0,s);
audio_in=open( devname, ui.ad_device_mode, 0);
if(audio_in == -1)end_program(1038);
j=dev_rd_channels[ui.ad_device_no];
xytext(0,10,"Select radio interface:");
if(j == 2)j=3;
for(i=0; i<j; i++)
  {
  sprintf(s,"%d: %s",i+1,audiomode_text[i]);
  xytext(5,12+i,s);
  }
chsel:;
await_keyboard();
clean_vga_inkey();
if(vga_inkey-'0'>j)goto chsel;
switch (vga_inkey)
  {
  case '1':
  ui.rx_channels=1;
  ui.input_mode=0;
  ui.ad_channels=1;
  break;

  case '2':
  ui.input_mode=IQ_DATA;
  ui.rx_channels=1;
  ui.ad_channels=2;
  break;
    
  case '3':
  ui.input_mode=TWO_CHANNELS;
  ui.rx_channels=2;
  ui.ad_channels=2;
  break;

  case '4':
  ui.input_mode=TWO_CHANNELS+IQ_DATA;
  ui.rx_channels=2;
  ui.ad_channels=4;
  break;

  default:
  goto chsel;
  }
i=ui.ad_channels;
if(ioctl(audio_in, SNDCTL_DSP_CHANNELS, &i) == -1)end_program(1039);  
if(i!=ui.ad_channels)end_program(1027);
continue_dual:;
min=1;
if(ioctl(audio_in, SNDCTL_DSP_SPEED, &min)==-1)end_program(1040);
fprintf(sndlog,"\nMin sampling speed = %d Hz",min);
for(i=10; i<12+j; i++)xytext(0,i,ss);
#ifdef SOUND_OSS
sprintf(s,"Use ossmix");
#else
sprintf(s,"Use system mixer program");
#endif
column=0;
while(s[column] != 0)column++;
sprintf(&s[column]," to select input and");
xytext(0,2,s);
xytext(0,3,"to disable direct connection from input to output etc.");
xytext(0,5,"In case selection of sampling frequencies is not what you");
xytext(0,6,"expect, check README files for your board - there may be ways");
xytext(0,7,"to reconfigure the hardware. (Under OSS, run ossmix)");
max=dev_rd_speed[ui.ad_device_no&255];
if(max == ABOVE_MAX_SPEED || max < 1000)
  {
  max = ABOVE_MAX_SPEED;
  }
sprintf(s,"Select sampling speed(%d to %d)", min,max);
xytext(0,10,s);
if(speed_warning != 0)
  {
  settextcolor(12);
  xytext(0,12,
    "WARNING!!!! device does not respond correctly to speed setting command");
   xytext(0,13,"Incorrect (not supported) values may cause system to hang");
  settextcolor(7);
  }
column=0;
while(s[column] != 0)column++;
ui.input_speed=vga_get_integer(column+2, 10, 8, min,max);
rdwr_speed=ui.input_speed;
fprintf(sndlog,"\n%d Hz selected by user",ui.input_speed);
for(i=5; i<11; i++)xytext(0,i,ss);
if(ioctl(audio_in, SNDCTL_DSP_SPEED, &ui.input_speed)==-1)end_program(1041);
if( (float)ui.input_speed > (float)rdwr_speed*1.05 ||
    (float)ui.input_speed < (float)rdwr_speed*0.95 )
  {
// The user selected something that was unacceptable to the device driver.
// Rather than just accepting whatever value returned,
// could be -22 (an unspecified error code) we step the desired speed
// upwards, looking for a valid speed.
// In case we fail, an attempt is made downwards as well.
// ****************************************************************
// This code is written by Ulf, SM0LCB
  fprintf(sndlog,"\nReqested speed didn't work");
// This code will search up/down from requested speed
// until the driver will respond with valid code
// Search use delta stepping of 3% from current speed
// This to speed up the search for a valid speed
// First search up in fq
  fprintf(sndlog,"\nSearch up : ");
  for( trySpeed=rdwr_speed, deltaSpeed=(float)(rdwr_speed)*0.03;
                 trySpeed<192000;
                 trySpeed+=deltaSpeed )
    {
    ui.input_speed=trySpeed;
    if(ioctl(audio_in, SNDCTL_DSP_SPEED, &ui.input_speed)==-1)
                                                            end_program(1024);
// test if valid speed
// must be float converted because int will overflow
    if( (float)ui.input_speed < (float)trySpeed*1.05 &&
        (float)ui.input_speed > (float)trySpeed*0.95 )
      {
      fprintf(sndlog,"try speed %d Hz ", trySpeed);
      goto spok;
      }
    deltaSpeed=(float)(trySpeed)*0.03;
    }
  fprintf(sndlog,"no speed found");
// Then search down in fq
  fprintf(sndlog,"\nSearch down : ");
  for( trySpeed=rdwr_speed, deltaSpeed=(float)(rdwr_speed)*0.03;
                 trySpeed>5000;
                 trySpeed-=deltaSpeed )
    {
    ui.input_speed=trySpeed;
    if(ioctl(audio_in, SNDCTL_DSP_SPEED, &ui.input_speed)==-1)
                                                            end_program(1135);
// test if valid speed
    if( (float)ui.input_speed < (float)trySpeed*1.05 &&
        (float)ui.input_speed > (float)trySpeed*0.95 )
      {
      fprintf(sndlog,"try speed %d Hz ", trySpeed);
      goto spok;
      }
    deltaSpeed=(float)(trySpeed)*0.03;
    }
  fprintf(sndlog,"no speed found");
// If we didn't find any vaid speed
  fprintf(sndlog,"\nNo valid speed found !");
// Exit program
  end_program(1134);
  }
spok:;
fprintf(sndlog,"\n%d Hz reported by device",ui.input_speed);
flush_sync();
sprintf(s,"Input speed: %.3fkHz",0.001*(float)(ui.input_speed));
xytext(0,5,s);
if( dev_rd_bits[ui.ad_device_no&255] == 32)
  {
  xytext(0,10,"Select A/D data format:");
  xytext(0,12,"1: 16 bit");
  xytext(0,13,"2: 32 bit");
selfmt:;
  await_keyboard();
  clean_vga_inkey();
  if(vga_inkey == '2')
    {
    ui.input_mode|=DWORD_INPUT;
    }
  else
    {
    if(vga_inkey != '1')goto selfmt;
    }
  }
for(i=10; i<14;i++)xytext(0,i,ss);
i=AFMT_S16_LE;
j=16;
#ifdef AFMT_S32_LE
if( (ui.input_mode&DWORD_INPUT) != 0)
  {
  i=AFMT_S32_LE;
  j=32;
  }
#endif
sprintf(s,"Input format: %d bit",j);
fprintf(sndlog,"\nInput format: %d bit",j);
xytext(30,5,s);
j=i;
if(ioctl(audio_in, SNDCTL_DSP_SETFMT, &i) == -1)end_program(1042);
if(i!=j)end_program(1043);



/*
// Close the input device(s) and reopen
close_audio_input();
open_audio_input();
*/




settextcolor(12);
sprintf(s,"Trying to read %s",devname);
xytext(10,10,s);
xytext(0,12,"If you see this message for more than 2 seconds READ failed");
xytext(0,13,"Then try another device");
if( ui.ad_device_mode == O_RDWR)xytext(37,13,"or open as RDONLY");
settextcolor(7);
gettimeofday(&tim2,NULL);
read(audio_in, testbuff, 2048);
if(ui.ad_device_no > 255)read(audio_in2, testbuff, 2048);
if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1045);
t1=0;
j=ui.input_speed*ui.ad_channels;
k=j;
if( (ui.input_mode&IQ_DATA) != 0)j*=2;
while(t1<1.5 && j>0 && k>0)
  {
  gettimeofday(&tim1,NULL);
  t1=0.000001*(tim1.tv_usec-tim2.tv_usec)+(tim1.tv_sec-tim2.tv_sec);
  if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1028);
  if(ad_info.bytes > 1024)
    {
    read(audio_in, testbuff, 1024);
    j-=1024;
    }
  if(ui.ad_device_no > 255)
    {
    if(ioctl(audio_in2, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
                                                          end_program(1028);
    if(ad_info.bytes > 1024)
      {
      read(audio_in2, testbuff, 1024);
      k-=1024;
      }
    else
      {
      k=j;
      }
    }
  }
if(j > 0)
  {
  settextcolor(12);
  xytext(10,14,"Read failed!!!  (press any key)");
  close_audio_input();
  free(testbuff);
  await_keyboard();
  clean_vga_inkey();
  ui.input_mode=-1;
  fprintf(sndlog,"\nRead test failed");
  goto setad_errexit;
  }
fprintf(sndlog,"\nRead test OK");
for(i=10; i<14; i++)xytext(0,i,ss);
sprintf(s,"Now trying to open a second device while %s is open",devname);
xytext(0,6,s);
xytext(0,7,"This may cause your system to hang. Try ctrlC to exit (or reboot)");
xytext(0,8,"Try another input device or install another sound system."); 
// Find devices that we may use for output.
n=0;
cn=0;
line=10;
for(i=0; i<MAX_DEVNAMES; i++)
  {
  if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1088);
  while(ad_info.bytes > 1024)
     {
     read(audio_in, testbuff, 1024);
     if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info) == -1)end_program(1093);
     }
  make_devname(i);
  cn=0;
  sprintf(s,"%02d:%12s",i,devname);
  xytext(0,line,s);
  column=0;
  while(s[column] != 0)column++;
  if( i == (ui.ad_device_no&255) )
    {
    if( ui.ad_device_mode == O_RDWR)
      {
      cn=1;
      if( (ui.input_mode&DWORD_INPUT) != 0 || ui.ad_channels > 2)
        {
        sprintf(&s[column],
             " RDWR  useless (too many bits or channels) %d: %s",i,devname);
        dev_flag[i]=0;
        }
      else
        {  
        sprintf(&s[column],"  %d Hz   %d Chan.   16 bit RDWR",
                                          ui.input_speed, ui.ad_channels);
        n++;
        ui.da_device_no=i;  
        }
      }
    }
  else
    { 
// Remove duplex flags on all devices except the one we use for input
    dev_flag[i]&=0xffff-4;
    if( (dev_flag[i]&2) != 0 )
      {
      audio_out=open( devname, O_WRONLY|O_NONBLOCK, 0);
      if(audio_out != -1)
        { 
        fprintf(sndlog,"\n%s opened as WRONLY",devname);
        flush_sync();
        if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, 
                                          &ad_info) == -1)end_program(1094);
        olbytes=ad_info.bytes;
        j=AFMT_U8;
        if(ioctl(audio_out, SNDCTL_DSP_SETFMT, &j) == -1)
          {
          fprintf(sndlog,"\nCould not set 8-bit output");
          dev_wr_bits[i]=0;
          }
        else
          {  
          fprintf(sndlog,"\n8-bit output supported");
          dev_wr_bits[i]=8;
          }
       flush_sync();
       j=AFMT_S16_LE;
        if(ioctl(audio_out, SNDCTL_DSP_SETFMT, &j) == -1  || j != AFMT_S16_LE) 
          {
          fprintf(sndlog,"\nCould not set 16-bit output");
          }
        else
          {
          dev_wr_bits[i]=16;
          }
        fprintf(sndlog,"\n16-bit output supported");
        flush_sync();
        if(dev_wr_bits[i]==0) goto skip_wr_dev;
        flush_sync();  
        j=2;
        if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &j) == -1)
          {
          j=1;
          if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &j) == -1)
            {
            fprintf(sndlog,"\nioctl(CHANNELS) failed");
            goto skip_wr_dev;
            }
          }
        if( j != 1 && j != 2)
          {
          fprintf(sndlog,"\nNo of channels not 1 or 2");
          goto skip_wr_dev;
          }
        fprintf(sndlog,"\nNo of output channels= %d",j);
        flush_sync();
        dev_wr_channels[i]=j;
        j=ABOVE_MAX_SPEED;
        if(ioctl(audio_out, SNDCTL_DSP_SPEED, &j) == -1 || j==ABOVE_MAX_SPEED)
          {
          fprintf(sndlog,"\nIncorrect speed response in %d  out %d",
                                                       ABOVE_MAX_SPEED, j);
          if(ioctl(audio_out, SNDCTL_DSP_SPEED, &j)==-1)
            {
            fprintf(sndlog,"\nioctl(SPEED) failed");
            goto skip_wr_dev;
            }
          }    
        dev_wr_speed[i]=j;
        fprintf(sndlog,"\nMax output speed %d", dev_wr_speed[i]);
        flush_sync();
        for(j=0; j<1024; j++)testbuff[j]=0;
        if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
          {
          fprintf(sndlog,"\nWrite failed");
          goto skip_wr_dev;
          }
        update_recent_time();
        sound_start_time=recent_time;  
        while(ad_info.bytes > 1024)
          {
          write(audio_out, testbuff, 1024);
          if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
            {
            fprintf(sndlog,"\nSNDCTL_DSP_GETOSPACE failed");
            goto skip_wr_dev;
            }
          update_recent_time();
          if( recent_time- sound_start_time > 0.7)  
            {
            fprintf(sndlog,"\nTimeout");
            goto skip_wr_dev;
            }
          }
        if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, 
                                            &ad_info) == -1)end_program(1095);
        while(ad_info.bytes > 1024)
          {
          read(audio_in, testbuff, 1024);
          if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,
                                            &ad_info) == -1)end_program(1096);
          }
        olbytes=ad_info.bytes;       
        gettimeofday(&tim2,NULL);
        while(ad_info.bytes-olbytes < 512)
          {
          if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,
                                            &ad_info) == -1)end_program(1097);
          gettimeofday(&tim1,NULL);
          t1=0.000001*(tim1.tv_usec-tim2.tv_usec)+(tim1.tv_sec-tim2.tv_sec);
          if(t1 > 0.25)goto skip_wr_dev;
          }
        if(close(audio_out)==-1)goto skip_wr_dev;  
        sprintf(&s[column]," %d Hz   %d Chan.   %d bit WRONLY",
                dev_wr_speed[i], dev_wr_channels[i], dev_wr_bits[i]);
        cn=1;
        n++;
        ui.da_device_no=i;  
        }
      else 
        {
        goto clr_wrdev;  
skip_wr_dev:;
        flush_sync();
        if(close(audio_out)==-1)end_program(1067);  
clr_wrdev:;
        dev_flag[i]&=0xffff-2;
        }
      }         
    }
  if(cn != 0)
    {
    xytext(0,line,s);
    if(line < maxline+1)line++;
    }
  }
if(cn == 0)
  {
  xytext(0,line,ss);
  }
else
  {
  line++;
  }
if(n == 0)
  {
  if(line > maxline-5)line=maxline-5;
  xytext(0,line,"No device found for 8/16 bit output while input is open.");
  xytext(0,line+1,"You may use Linrad to watch spectra and record raw data.");
  xytext(0,line+2,"Linrad will now look for devices that can be used to");
  xytext(0,line+3,"output sound while processing recorded data.");
  xytext(0,line+4,"(press any key)");
  fprintf(sndlog,"\nNo output device found while input is open.");
  close_audio_input();
  await_keyboard();
  clean_vga_inkey();
  vga_clear();
  n=0;
  cn=0;
  line=0;
  for(i=0; i<MAX_DEVNAMES; i++)
    {
    make_devname(i);
    fprintf(sndlog,"\n%s ",devname);
    cn=0;
    sprintf(s,"%02d:%12s",i,devname);
    xytext(0,line,s);
    column=0;
    while(s[column] != 0)column++;
    audio_out=open( devname, O_WRONLY|O_NONBLOCK, 0);
    if(audio_out != -1)
      { 
      fprintf(sndlog,"\n%s opened as WRONLY with no input open.",devname);
      flush_sync();
      j=AFMT_U8;
      if(ioctl(audio_out, SNDCTL_DSP_SETFMT, &j) == -1)
        {
        fprintf(sndlog,"\nCould not set 8-bit output");
        dev_wr_bits[i]=0;
        }
      else
        {  
        dev_wr_bits[i]=8;
        }
      j=AFMT_S16_LE;
      if(ioctl(audio_out, SNDCTL_DSP_SETFMT, &j) == -1  || j != AFMT_S16_LE) 
        {
        fprintf(sndlog,"\nCould not set 16-bit output");
        }
      else
        {
        dev_wr_bits[i]=16;
        }
      if(dev_wr_bits[i]==0) goto skip_wr_dev_nodup;
      flush_sync();  
      j=2;
      if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &j) == -1)
        {
        j=1;
        if(ioctl(audio_out, SNDCTL_DSP_CHANNELS, &j) == -1)
          {
          fprintf(sndlog,"\nioctl(CHANNELS) failed");
          goto skip_wr_dev_nodup;
          }
        }
      if( j != 1 && j != 2)
        {
        fprintf(sndlog,"\nNo of channels not 1 or 2");
        goto skip_wr_dev_nodup;
        }
      dev_wr_channels[i]=j;
      j=ABOVE_MAX_SPEED;
      if(ioctl(audio_out, SNDCTL_DSP_SPEED, &j) == -1 || j==ABOVE_MAX_SPEED)
        {
        fprintf(sndlog,"\nIncorrect speed response in %d  out %d",
                                                      ABOVE_MAX_SPEED, j);
        j=ui.input_speed;
        if(ioctl(audio_out, SNDCTL_DSP_SPEED, &j)==-1)
          {
          fprintf(sndlog,"\nioctl(SPEED) failed");
          goto skip_wr_dev_nodup;
          }
        }    
      dev_wr_speed[i]=j;
      fprintf(sndlog,"\nMax output speed %d", dev_wr_speed[i]);
      flush_sync();
      for(j=0; j<1024; j++)testbuff[j]=0;
      if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
        {
        fprintf(sndlog,"\nWrite failed");
        goto skip_wr_dev_nodup;
        }
      while(ad_info.bytes > 1024)
        {
        write(audio_out, testbuff, 1024);
        if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
                                               goto skip_wr_dev_nodup;
        }
      if(close(audio_out)==-1)goto skip_wr_dev_nodup;  
      sprintf(&s[column]," %d Hz   %d Chan.   %d bit WRONLY",
              dev_wr_speed[i], dev_wr_channels[i], dev_wr_bits[i]);
      cn=1;
      n++;
      ui.da_device_no=i;  
      dev_flag[i] |= 2;
      }
    else 
      {
      goto clr_wrdev_nodup;  
skip_wr_dev_nodup:;
      flush_sync();
      if(close(audio_out)==-1)end_program(1138);  
clr_wrdev_nodup:;
      dev_flag[i]&=0xffff-2;
      }
    if(cn != 0)
      {
      xytext(0,line,s);
      if(line < maxline+1)line++;
      }
    }
  if(cn == 0)
    {
    xytext(0,line,ss);
    }
  else
    {
    line++;
    }
  if(n == 0)
    {
    fprintf(sndlog,"\n");
    fclose(sndlog);
    end_program(1139);
    }
  ui.input_mode|=NO_DUPLEX;  
  ad_info.bytes=0;
  }
if(n > 1)
  {
get_dadev:;
  i=0;
  j=0;
  settextcolor(14);
  xytext(0,line,"Select device for output by line number");
  settextcolor(7);
  if( (ui.input_mode&NO_DUPLEX) == 0)
    {
    if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1089);
    }
  while(i<2)
    {
    if(ad_info.bytes > 1024)
      {
      read(audio_in, testbuff, 1024);
      }
    vga_inkey=vga_getkey();
    clean_vga_inkey();
    if(vga_inkey >= '0' && vga_inkey <= '9')
      {
      sprintf(s,"%c",vga_inkey);
      xytext(40+i,line,s);
      i++;
      j=10*j+vga_inkey-'0';
      }
    if(vga_inkey == 10 && i>0)i=2;
    }
  ui.da_device_no=j;
  if( (6 & dev_flag[ui.da_device_no]) == 0) goto get_dadev;
  }
make_devname(ui.da_device_no);
fprintf(sndlog,"\nUser selected %s for output.",devname);  
for(i=6; i<=line; i++)xytext(0,i,ss);  
xytext(0,8,"Use close and reopen for D/A when");
xytext(0,9,"a new signal is selected? (Y/N)");
get_da_stopstart:;
vga_inkey=vga_getkey();
clean_vga_inkey();
if(vga_inkey == 'Y')
  {
  ui.da_stopstart=1;
  }
else  
  {
  if(vga_inkey != 'N')goto get_da_stopstart;
  ui.da_stopstart=0;
  }
for(i=8; i<=9; i++)xytext(0,i,ss);  
// If we selected the same device for input and output,
// check if it was opened ad RDWR or if we shall open it
// for output now
if( ui.da_device_no == ui.ad_device_no && 
    ui.ad_device_mode == O_RDWR &&
    (ui.input_mode&NO_DUPLEX) == 0 )
  {
  audio_out=audio_in;
  ui.max_output_channels=ui.ad_channels;
  ui.min_output_channels=ui.ad_channels;
  ui.max_output_bytes=2;
  ui.min_output_bytes=2;
  ui.max_output_speed=ui.input_speed;
  ui.min_output_speed=ui.input_speed;
  }
else
  {
  audio_out=open( devname ,O_WRONLY , 0);
  sprintf(s,"\nUsing %s for output",devname);
  xytext(0,6,s);
  if(audio_out == -1)end_program(1068);
  da_wrbuf = NULL;
  set_da_parms();
  close_audio_output();
  }
if( (ui.input_mode&NO_DUPLEX) == 0) close_audio_input();
sprintf(s,"Min output channels: %d",ui.min_output_channels);  
xytext(0,8,s);
sprintf(s,"Max output channels: %d",ui.max_output_channels);  
xytext(0,9,s);
da_output_channels=ui.min_output_channels;
da_output_bytes=ui.max_output_bytes;
sprintf(s,"Output no of bits=%d",8*ui.max_output_bytes);  
xytext(35,7,s);
if(ui.max_output_bytes != ui.min_output_bytes)
  {
  sprintf(s,"or %d",8*ui.min_output_bytes);  
  xytext(56,7,s);
  }
// Reopen with the current parameters. 
// We will not return here if something goes wrong.
ui.ad_frag=8;
if(  (ui.input_mode&NO_DUPLEX) == 0)
  {
  open_audio_input();
  genparm[DA_OUTPUT_SPEED]=0;
  open_audio_output();
  read(audio_in, testbuff, 2048);
  if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)end_program(1001);
  close_audio_input();
  }
close_audio_output();
xytext(0,10,"Audio board configured without errors.");
free(testbuff);
xytext(0,13,"Use ""W = Save"" in main menu to keep");
xytext(10,14,"(press any key)");
ui.ad_frag=0;
await_keyboard();
clean_vga_inkey();
fprintf(sndlog,"\nNormal end");
setad_errexit:;
fprintf(sndlog,"\n");
fclose(sndlog);
}

