




// Structure for Morse decoding data.
typedef struct {
unsigned char type;    // Part type (CW_DOT, CW_SPACE,....A,B,C,...)
unsigned char unkn;    // Number of unknown states before this part
unsigned char down;    // Number of states to lower character boundary
unsigned char len;     // Length of character in states 
float midpoint;
float sep;
float re;
float im;
}MORSE_DECODE_DATA;

// Defines for MORSE_DECODE_DATA.type (cw.type)
// Note that CW_DASH must have the lowest value.
// The data in cw_item_len[255-CW_DASH]; must agree with these definitions!


#define CW_WORDSEP 255
#define CW_SPACE 254
#define CW_DOT 253
#define CW_DASH 252






// Defines for state variable in main receive loops.
#define RX_LOW_PRIO -1
#define RX_GET_RAW_DATA 10
#define RX_FFT1_FIRST 12
#define RX_FFT1_SECOND 14
#define RX_FFT1_THIRD 16
#define RX_FFT1_FOURTH 19
#define RX_FFT1_TO_TIMF2 25
#define RX_RUN_BLANKER 30
#define RX_CHK_FFT2 40
#define RX_MIX1_BLOCK_FFT1 50
#define RX_MIX1_BLOCK_FFT2 55
#define RX_FFT3_BLOCK 60
#define RX_SHOW_BASEB_SPECTRUM 65
#define RX_MIX2_BLOCK 70
#define RX_AUDIO_BLOCK 80
#define RX_WATMO 90
#define RX_USER_CONTROL 100
#define RX_MIX1_BLOCK_TXTEST 300
#define RX_SHOW_FFT1 310
#define RX_SHOW_WIDE 320
#define RX_PREPARE_WIDE 330



// Defines for baseb_detect_flag
#define DET_KEY_UP 4
#define DET_DOT 5
#define DET_DASH 6
#define DET_NEW_DATA 0 

// Time span over which we try to determine keying speed
// The time is expressed in data bits that would fit the selected
// filter bandwidth.
#define CWSPEED_UNITS 200

// The number of dots and dashes we can use in the detect routine
// is computed from the maximum number of dots the buffer can hold
// multiplied by this factor
#define MAX_CW_PARTS 0.2

// Parameters that define detect accuracy.
#define DASH_DETECT_LIMIT 0.82
#define DASH_MISFIT_LIMIT 0.7
#define DOT_DETECT_LIMIT 0.88
#define DOT_MISFIT_LIMIT 0.75
// The error in the detect waveform is given by the S/N ratio.
// This is the limit at which we decide that the keying speed is 
// found accurately enough so we can switch from looking for
// the keying speed to actually start to look for code parts.
#define CWSPEED_STON 0.05


// Defines for the Morse detect oscilloscope.
#define CG_A 80
#define CG_AMAX 40
#define COH_GRAPH_ACTIVE -1000


#define POLEVAL_SIZE 32
typedef struct {
float a2;
float b2;
float re;
float im;
} POLCOFF;


extern int cw_item_len[256-CW_DASH];
extern int cg_no_of_traces;
extern int cg_lsep;
extern char cg_color[];
extern int cg_osc_ptr;
extern int cg_osc_offset;
extern int cg_osc_offset_inc;
extern int cg_osc_shift_flag;
extern int cg_max_trlevel;
extern float cg_code_unit;
extern float cg_decay_factor;
extern int cg_collect_points;
extern int cg_shape_bufno;
extern int cwlong_store_points;
extern int cwspeed_store_points;
extern int cg_statistics_sum;
extern int cg_wave_start;
extern float cg_wave_midpoint;
extern float cg_wave_fit;
extern float cg_wave_re;
extern float cg_wave_im;
extern float cg_wave_dat;
extern float cg_wave_err;


extern int cg_update_interval;
extern int cg_update_count;

extern float cw_stoninv;
extern int no_of_cwdat;

extern double da_block_counter;
extern float baseband_sampling_speed;
extern int *baseband_handle;
extern int baseband_size;
extern int baseband_mask;
extern int baseband_neg;
extern float *baseb_out;
extern float *baseb;
extern float *baseb_raw;
extern float *baseb_raw_orthog;
extern float *baseb_carrier;
extern float *baseb_carrier_ampl;
extern float *baseb_totpwr;
extern float *baseb_envelope;
extern float *baseb_upthreshold;
extern float *baseb_threshold;
extern float *baseb_fit;
extern float *baseb_errpwr;
extern float *baseb_agc_level;
extern unsigned char *baseb_detect_flag;
extern float *basblock_maxpower;
extern float *basblock_avgpower;
extern short int *baseb_ramp;
extern float *baseb_tmp;
extern float *cg_tmpbuf;
extern float *cg_map;
extern float *cg_statistics;
extern float *cwlong_buffer;
extern short int *cg_traces;
extern float *dash_waveform;
extern float *dot_waveform;
extern float *dot_midpoint;
extern float *space_midpoint;
extern float *trispace_midpoint;
extern int cg_yzer[CG_MAXTRACE];
extern int no_of_dots;
extern int max_no_of_dots;
extern int max_no_of_dashes;
extern int no_of_spaces;
extern int max_no_of_spaces;
extern int no_of_trispaces;
extern int max_no_of_trispaces;
extern int cw_ptr;
extern int dash_waveform_established;
extern int cw_detect_flag;
extern int baseb_pa;
extern int baseb_pb;
extern int baseb_pc;
extern int baseb_pd;
extern int baseb_pe;
extern int baseb_pf;
extern int baseb_px;
extern float baseb_fx;
extern int mix2_size;
extern int mix2_n;
extern short int *mix2_permute;
extern COSIN_TABLE *mix2_table;
extern char *daout;
extern MORSE_DECODE_DATA *cw;
extern int daout_mask;
extern int daout_size;
extern int new_baseb_flag;
extern float da_resample_ratio;
extern float new_da_resample_ratio;
extern int daout_pa;
extern int daout_px;
extern int daout_block;
extern int daout_bytes;
extern int daout_bufmask;
extern int bfo_xpixel;
extern int bfo10_xpixel;
extern int bfo100_xpixel;
extern int bfo_flag;
extern int baseb_queue;
extern float daout_phstep;
extern float daout_phase;
extern float daout_gain;
extern int daout_gain_y;
extern char *bg_volbuf;
extern int da_wts;
extern int db_wts;
extern int da_maxempty;
extern int baseb_channels;
extern int old_baseb_channels;
extern float baseb_wts;
extern float da_wttim;  
extern float baseb_wttim;
extern float db_wttim;
extern float da_wait_time;
extern int poleval_pointer;
extern POLCOFF poleval_data[POLEVAL_SIZE];
extern int da_empty_flag;
extern int bg_expand;
extern int bg_coherent;
extern int bg_delay;
extern int bg_da_channels;
extern int bg_da_bytes;
extern int bg_twopol;
extern float bg_expand_a;
extern float bg_expand_b;
extern float bg_agc_amplimit;
extern float bg_amplimit;
extern float bg_maxamp;
extern int bg_amp_indicator_x;
extern int bg_amp_indicator_y;
extern int da_ch2_sign;
extern int bg_carrier_x;
extern char *bg_behind_carrcur;
extern int cg_size;
extern int dash_pts;
extern int dot_pts;
extern float dash_sumsq;
extern float dot_sumsq;
extern float cwbit_pts;
extern float dot_pwrmax;
extern float refpwr;
extern int basblock_size;
extern int basblock_pa;
extern int basblock_mask;
extern int basblock_hold_points;
extern float cg_maxmax;
extern float cg_maxavg;
extern float baseband_pwrfac;
extern float agc_factor;
extern float agc_sumpow1;
extern float agc_sumpow2;
extern float agc_attack_factor1;
extern float agc_attack_factor2;
extern float agc_release_factor;
extern float am_dclevel1;
extern float am_dclevel2;
extern float am_dclevel_factor1;
extern float am_dclevel_factor2;
extern int s_meter_avgnum;
extern float s_meter_avg;
extern int max_cwdat;


void show_bg_maxamp(void);
void store_dot(void);
void store_dash(void);
void get_average_shapes(void);
void fit_dash(void);
void fit_dot(void);
void add_cwlong(int mid);
void shift_cg_osc_track(int shift);
void store_space(void);
void store_tripple_space(void);
float check_dot(float pos);
float check_dash(float pos);
void set_region_envelope(void);
void set_long_region_envelope(void);
void clear_region(void);
void store_symmetry_adapted_dashdot(int itot);
void update_cg_traces(void);
void accurate_store_dash(void);  
void detect_cw_speed(void);
void skip_old_dashes(void);
void first_refine_carrier(void);
void seldash_cwspeed(int da, int db);
void remove_dash(void);
void insert_item(int type);
void show_bg_buttons(void);
