#include <vgagl.h>
#include "conf.h"
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "seldef.h"
#include "sigdef.h"

void clear_ad_countertimer(void)
{
// The number of blocks read and the total time elapsed are
// collected to make it possible to measure ad sampling speed.
// Speed may be seriously wrong for various reasons, particularly
// when a half duplex sound board is set up for different speeds
// for input and output. 
gettimeofday(&ad_tim,NULL);
ad_block_counter=0;
// Check after 5 seconds the first time.
ad_block_check=5*interrupt_rate;
}


void make_ad_wttim(void)
{
audio_buf_info ad_info;
// ***********************************************
// ad_wts, ad_wttim. Data waiting before first fft.
if(network_flag == 2)
  {
  if(ioctl(network_fd, FIONREAD, &ad_wts) == -1)end_program(1145);
  }
else
  {  
  if(disksave_flag < 2)
    {
    if(ioctl(audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)end_program(1083);
    ad_wts=ad_info.bytes;
    }
  else
    {
    ad_wts=0;
    }  
  }
ad_wts+=(timf1p_pa-timf1p_px+timf1_bytes)&timf1_bytemask;
ad_wts/=bytes_per_fragment;
if(fft_cntrl[FFT1_CURMODE].approx_iq != 0)
  {
  ad_wts+=((ctimf1p_pa-ctimf1p_px+ctimf1_bytes)&ctimf1_bytemask)
                                           /(ui.rx_channels*sizeof(int));
  }
ad_wttim=(float)(ad_wts)/ui.input_speed;
}

void make_fft2_wttim(void)
{
// ***********************************************
// In case second fft is enabled we have:
// timf2_wtb back transformed fft1 blocks waiting in timf2 and
// fft2_wtb completed fft2 transforms waiting to be back transformed
// fft2_wtb should be small or zero unless AFC is operating on delayed signal.
// Back transformed fft1 transforms waiting in timf2.
timf2_wtb=(float)((timf2_pa-timf2_px+timf2_mask+1)&timf2_mask)
                                                        /timf2_input_block;
timf2_wttim=timf2_wtb*fft1_blocktime;   
// fft2 transforms waiting for back transformation.
fft2_wtb=(fft2_na-fft2_nx+max_fft2n)&fft2n_mask;  
fft2_wttim=fft2_wtb*fft2_blocktime;    
}

void make_da_wttim(void)
{
// Data written to device driver but not yet sent to the speaker.
if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)end_program(1149);
if(da_info.bytes > da_maxempty) da_maxempty=da_info.bytes;
if(da_info.bytes > daout_block) fullbuf_flag &=-1-ND_DABUF;
da_wts=da_info.fragstotal*da_info.fragsize-da_info.bytes;
da_wts/=(da_output_channels*da_output_bytes);
da_wttim=(float)(da_wts)/genparm[DA_OUTPUT_SPEED];
total_wttim+=da_wttim;
}



void make_timing_info(void)
{
make_ad_wttim();
total_wttim=ad_wttim;
if( genparm[SECOND_FFT_ENABLE] != 0 )
  {
// ***********************************************
// fft1_wtb completed fft1 transforms waiting to be back transformed
// to the time domain.
// This number should be small or zero unless AFC is operating without
// second fft on delayed signal.
  fft1_wtb=(float)((fft1_pa-fft1_px+fft1_mask+1)&fft1_mask)/fft1_block;
  fft1_wttim=fft1_wtb*fft1_blocktime;
  total_wttim+=fft1_wttim;
  if(mix1_selfreq[0] >=0 )
    {
    make_fft2_wttim();
    total_wttim+=timf2_wttim;
    total_wttim+=fft2_wttim;
    }
  else
    {
    timf2_wttim=0;
    fft2_wttim=0;
    }  
  }
else
  {
  if(mix1_selfreq[0] >=0 )
    {
// ***********************************************
// fft1_wtb completed fft1 transforms waiting to be 
// transformed to the baseband by mix1.
    if(genparm[AFC_ENABLE] == 0)
      {
      fft1_wtb=(float)((fft1_pa-fft1_px+fft1_mask+1)&fft1_mask)/fft1_block;
      }
    else
      {
      fft1_wtb=(fft1_na-fft1_nx+max_fft1n)&fft1n_mask;
      }
    fft1_wttim=fft1_wtb*fft1_blocktime;
    total_wttim+=fft1_wttim;
    }
  else
    {
    fft1_wttim=0;
    }  
  }  
// ************************************************
// If a frequency is selected for audio output: 
if(mix1_selfreq[0] >=0 )
  {
// Samples waiting in timf3.
  timf3_wts=((timf3_pa-timf3_px+timf3_mask+1)&timf3_mask)/twice_rxchan;
  timf3_wttim=timf3_wts/timf3_sampling_speed;
  total_wttim+=timf3_wttim;
// Blocks waiting as finished transforms. 
  fft3_wtb=(fft3_pa-fft3_px+fft3_totsiz)&fft3_mask;
  fft3_wtb/=fft3_block;
// We get fft3_size/2 points from each block
  fft3_wttim=0.5*fft3_wtb*fft3_size/timf3_sampling_speed;
  total_wttim+=fft3_wttim;
// Samples waiting in the baseband 
  baseb_wttim=baseb_wts/baseband_sampling_speed;
  total_wttim+=baseb_wttim;
// Data waiting for output 
  db_wts=(daout_pa-daout_px+daout_bufmask+1)&daout_bufmask;
  db_wts/=(da_output_channels*da_output_bytes);
  db_wttim=(float)(db_wts)/genparm[DA_OUTPUT_SPEED];
  total_wttim+=db_wttim;
  if(audio_out != -1)
    {
    make_da_wttim();
    }
  else
    {
    da_wts=0;
    da_wttim=0;
    da_maxempty=0;
    }
  }
else
  {
  db_wts=0;
  db_wttim=0;
  da_wts=0;
  da_wttim=0;
  }
}


void show_amp_info(void)
{
char s[80];
int line, i, k;
float t1;
line=HEIGHT/text_height-1;
sprintf(s,"%6.2f  ",sqrt(fft1_noise_floor));
xytext(20,line,s);
line--;
k=5;
for(i=0; i<ui.ad_channels; i++)
  {
  t1=ad_maxamp[i];
  if(t1<1)t1=1;
  t1/=0x8000;
  t1=-20*log10(t1);
  sprintf(s,"%2.2f ",t1);
  xytext(k,line,s);
  k+=6;
  }
if(rx_mode == MODE_TXTEST)return;
if( genparm[SECOND_FFT_ENABLE] != 0 )
  {
  if(swmmx_fft2)
    {
    line--;  
    k=10;
    for(i=0; i<ui.rx_channels; i++)
      {
      t1=fft2_maxamp[i];
      if(t1<1)t1=1;
      t1/=0x8000;
      t1=-20*log10(t1);
      sprintf(s,"%2.2f ",t1);
      xytext(k,line,s);
      k+=6;
      }
    }  
  line--;  
  k=10;
  for(i=0; i<ui.rx_channels; i++)
    {
    t1=timf1_maxamp[i];
    if(t1<1)t1=1;
    t1/=0x8000;
    t1=-20*log10(t1);
    sprintf(s,"%2.2f ",t1);
    xytext(k,line,s);
    k+=6;
    }
  line--;  
  k=10;
  for(i=0; i<ui.rx_channels; i++)
    {
    t1=timf1_maxamp[MAX_RX_CHANNELS+i];
    if(t1<1)t1=1;
    t1/=0x8000;
    t1=-20*log10(t1);
    sprintf(s,"%2.2f ",t1);
    xytext(k,line,s);
    k+=6;
    }
  if(swmmx_fft1)
    {
    line-=2;  
    k=10;
    for(i=0; i<ui.rx_channels; i++)
      {
      t1=fft1_maxamp[2*i];
      if(t1<1)t1=1;
      t1/=0x8000;
      t1=-20*log10(t1);
      sprintf(s,"%2.2f ",t1);
      xytext(k,line+1,s);
      t1=fft1_maxamp[2*i+1];
      if(t1<1)t1=1;
      t1/=0x8000;
      t1=-20*log10(t1);
      sprintf(s,"%2.2f ",t1);
      xytext(k,line,s);
      k+=6;
      }
    }
  }
}

void amp_info_texts(void)
{
int line;
   errno=40116;
ck_fillbox(0,HEIGHT-8*text_height,32*text_width,8*text_height,0);
line=HEIGHT/text_height-1;
xytext(14,line,"Floor");
line--;
xytext(0,line,"A/D");
if( genparm[SECOND_FFT_ENABLE] != 0 )
  {
  if(swmmx_fft2)
    {
    line--;
    xytext(4,line,"fft2");
    }
  line--;
  xytext(0,line,"timf2 Wk");
  line--;
  xytext(0,line,"timf2 St");
  if(swmmx_fft1)
    {
    line--;
    xytext(2,line,"fft1 Wk");
    line--;
    xytext(2,line,"fft1 St");
    }
  } 
line--;
xytext(0,line,"Amplitude margins (dB)");
}


void show_timing_info(void)
{
char s[80];
int line;
float t1;
make_timing_info();
line=HEIGHT/text_height-1;
sprintf(s,"%1.2f ",total_wttim);
xytext(18,line,s);
line--;
sprintf(s,"%1.2f ",ad_wttim);
xytext(6,line,s);
sprintf(s,"%1.2f",fft1_wttim);
xytext(18,line,s);
line--;
if( genparm[SECOND_FFT_ENABLE] != 0 )
  {
  sprintf(s,"%1.2f ",timf2_wttim);
  xytext(6,line,s);
  sprintf(s,"%1.2f ",fft2_wttim);
  xytext(18,line,s);
  line--;
  }
sprintf(s,"%1.2f ",timf3_wttim);
xytext(6,line,s);
sprintf(s,"%1.2f ",fft3_wttim);
xytext(18,line,s);
line--;
sprintf(s,"%1.2f ",baseb_wttim);
xytext(6,line,s);
// Write buf, how much time the data in daout[] corresponds to
sprintf(s,"%1.2f ",db_wttim);
xytext(18,line,s);
line--;
// Write D/A, time of data in output buffer (already written to device)
sprintf(s,"%1.2f ",da_wttim);
xytext(6,line,s);
// Write MIN, the minimum time margin in the output buffer.
t1=(da_info.fragstotal-1)*da_info.fragsize-da_maxempty;
if(t1<0)t1=0;	
t1/=da_output_channels*da_output_bytes*genparm[DA_OUTPUT_SPEED];
sprintf(s,"%1.2f ",t1);
xytext(18,line,s);
}





void timing_info_texts(void)
{
int line;
errno=40117;
ck_fillbox(0,HEIGHT-7*text_height,32*text_width,7*text_height,0);
line=HEIGHT/text_height-1;
xytext(14,line,"Tot");
line--;
xytext(2,line,"Raw");
xytext(13,line,"fft1");
line--;
if( genparm[SECOND_FFT_ENABLE] != 0 )
  {
  xytext(0,line,"timf2");
  xytext(13,line,"fft2");
  line--;
  }
xytext(0,line,"timf3");
xytext(13,line,"fft3");
line--;
xytext(0,line,"baseb");
xytext(14,line,"buf");
line--;
xytext(0,line,"D/A");
xytext(14,line,"MIN");
line--;
xytext(6,line,"Delay times");
}


