

#include <vga.h>
#include <vgagl.h>
#include <unistd.h>
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft3def.h"
#include "conf.h"
#include "sigdef.h"
#include "screendef.h"
#include "seldef.h"

#define MAX_USER 8  //Interrupt rate is around 30Hz so this is 0.25 seconds

// *********************  newdat_flag *************************
// newdat_flag holds info about the amount of unprocessed data.
// A bit that is set indicates that the amount of data is large
// enough for the next processing stage.
// Bits are cleared at the point when the amount of data is smaller
// than required to process one more block.
// ND_HWARE =>    Raw data buffer in device driver. 
//                Input is automatically generated by interrupt driven 
//                routines which may cause overrun errors if data is 
//                not fetched in time. 
//                This flag is set and reset here depending on the result
//                of an ioctl() call to the A/D converter or network.
//                In case data is read from the hard disk this flag is
//                set by a timer in the idle routine until end of file 
//                is encountered. When reading from disk, this flag is
//                reset by read_rawdat.
// ND_TIMF1 =>    This flag is set by read_rawdat if there is enough
//                data for a new fft1 transform. It is reset by fft1_b
// ND_FFT1 =>     This flag is set by fft1_c when at least one new
//                fft1 transform is present. It is reset at different
//                places depending on what parameters the operator has
//                selected. 
// ND_TIMF2RAW => Set by make_timf2, reset by first_noise_blanker.
// ND_TIMF2BLN => Set by first_noise_blanker, reset by make_fft2. 
// ND_FFT2 =>     Set by make_fft2, reset by fft2_mix1_fixed or fft2_mix1_afc
// ND_TIMF3 =>    Set by mix1 routines, reset by make_fft3_all  
// ND_FFT3 =>     set by make_fft3_all, 


//                reset in mix2_loop 
// ND_BASEB =>    a new block of data present in the baseband
//                set by fft3_mix2  
//                reset by make_audio_signal
// ND_DASIG =>    a new block of data present for D/A
//                set by make_audio_signal
//                reset by make_audio_output
//
//
// *********************  fullbuf_flag *************************
// fullbuf_flag holds info about full buffers.
// A bit that is set indicates that there is not enough space for the 
// next processing stage.
// ND_TIMF1 =>    set by read_rawdat, reset by fft1_b
// ND_FFT1 =>     set by fft1_b, reset by fft1_c
// ND_TIMF2RAW => set by make_timf2, reset by make_fft2
// ND_FFT2 =>     Set by make_fft2, reset by fft2_mix1_fixed or fft2_mix1_afc
// ND_TIMF3 =>    Set by mix1 routines, reset by make_fft3_all  
// ND_FFT3 =>     set by make_fft3_all, 

#define TXTEST_PEAK_POWER_COLOR 0x30
#define TXTEST_PEAK_DECAY_COLOR 14
#define TXTEST_WIDE_AVERAGE_COLOR 0x28
#define TXTEST_NARROW_COLOR 15
#define PEAK_REDRAW_COUNT 10


float txtest_decayfac2;
float txtest_decayfac1;
int icnt, jcnt;
int max_ijcnt;
int txtest_xpix1;
int txtest_last_xpix;
float txtest_fxpix1;
int txtest_ws_ia;
int txtest_peak_redraw;
int txtest_ka;
float txtest_saturated;

void txtest_init(void)
{
int i, j, k;
float t1;
int pnts_per_seg;
txtest_decayfac1=fft1_blocktime;
txtest_decayfac2=1-txtest_decayfac1;
txtest_no_of_segs=(fft1_last_point-fft1_first_point)*fft1_hz_per_point/600;
if(txtest_no_of_segs >= WIDTH)end_program(1159);
for(i=0; i<WIDTH; i++)
  {
  txtest_ypeak[i]=wg.ybottom-1;
  txtest_ypeak_decay[i]=wg.ybottom-1;
  txtest_yavg[i]=wg.ybottom-1;
  fft1_spectrum[i]=wg.ybottom-1;
  }
mix1_selfreq[0]=2400;
// Set up a rectangular filter with 2400Hz bandwidth
pnts_per_seg=2400/fft1_hz_per_point;
j=(mix1_size-pnts_per_seg)/2;
k=mix1_size/2;
i=0;
for(i=0; i<j; i++)
  {
  mix1_fqwin[i]=0;
  }
for(i=j; i<=mix1_size/2; i++)
  {
  mix1_fqwin[i]=1;
  }  
j=(mix1_size-mix1_new_points)/2;
k=j;
for(i=0; i<mix1_crossover_points; i++)
  {
  mix1_cos2win[i]=mix1_win[k];
  mix1_sin2win[i]=mix1_win[j];
  k--;
  j++;
  }
if(txtest_no_of_segs <=0)txtest_no_of_segs=1;
txtest_pntinc=(fft1_last_point-fft1_first_point)/txtest_no_of_segs;
t1=pnts_per_seg/2+1;
txtest_first_point=fft1_first_point+t1;
txtest_pixinc=txtest_pntinc;
if(wg.xpoints_per_pixel == 0)
  {
  txtest_pixinc=txtest_pntinc*wg.pixels_per_xpoint;
  t1*=wg.pixels_per_xpoint;
  }
else
  {
  txtest_pixinc=(float)(txtest_pntinc)/wg.xpoints_per_pixel;
  t1/=wg.xpoints_per_pixel;
  if(txtest_pixinc < 1)end_program(1160);
  }  
t1+=wg_first_xpixel;
txtest_spek_no=0;
txtest_spek_p0=0;
while(txtest_first_point-mix1_size/2 < fft1_first_point)
  {
  txtest_first_point+=txtest_pntinc;
  t1+=txtest_pixinc;
  }
txtest_first_xpix=t1+0.5;  
while(txtest_no_of_segs*txtest_pntinc+mix1_size/2+
               txtest_first_point > fft1_last_point) txtest_no_of_segs--;
txtest_yfac=wg_yfac_power*wg.spek_avgnum;
if(genparm[FIRST_FFT_SINPOW] != 0)
  {
  t1=0;
  for(i=0; i<fft1_size/2; i++)t1+=fft1_window[i];
  t1/=fft1_size;
  txtest_yfac/=t1*t1;
  }
max_ijcnt=0;
icnt=0;
jcnt=0;
wg_xpix1=wg_first_xpixel;
txtest_ws_ia=0;
txtest_xpix1=txtest_first_xpix;
txtest_fxpix1=txtest_first_xpix;
txtest_peak_redraw=0;
txtest_ka=WIDTH/(3*NO_REDRAW_BLOCKS*txtest_pixinc);
txtest_saturated=BIG;
}

void txtest_wrerr(char *s)
{
int ix, iy;
ix=(WIDTH>>1)-9*text_width;
iy=wg.ytop+(text_height>>1);
settextcolor(15);
while(iy < wg.ybottom-text_height)
  {
  gl_write(ix,iy,s);
  iy+=1.5*text_height;
  }
settextcolor(7);
}  

void show_adsat(void)
{
int i;
float t1;
char s[80];
t1=0;
for(i=0; i<ui.ad_channels; i++)
  {
  if(t1<ad_maxamp[i])t1=ad_maxamp[i];
  }
if(t1<1)t1=1;
t1/=0x8000;
t1=-20*log10(t1);
if(t1 < 0.0000001) t1=0.0000001;
sprintf(s,"A/D margin %5.2f dB",t1);
xytext(0,HEIGHT/text_height - 2,s);
if(t1 < 0.1 && txtest_saturated > t1)
  {
  txtest_wrerr(s);
  txtest_saturated=t1;
  }
}



void update_txtest(int xpix,int new_y2 )
{
int new_y1, old_y1, old_y2;
int ia, ib;
char bkg_color;
old_y2=fft1_spectrum[xpix];
fft1_old_spectrum[xpix]=old_y2;
if(new_y2 < 0)new_y2=0;
new_y2=wg.ybottom-1-new_y2;
if(new_y2 < wg.yborder+1)new_y2=wg.yborder+1;
if(xpix != wg_first_xpixel)
  {
  new_y1=fft1_spectrum[xpix-1];
  old_y1=fft1_old_spectrum[xpix-1];
  }
else
  {
  new_y1=new_y2;
  old_y1=new_y2;
  }
if( new_y2 != old_y2 || new_y1 != old_y1)
  {
  if(old_y1 > old_y2)
    {
    ia=old_y2;
    ib=old_y1;
    }
  else
    {
    ib=old_y2;
    ia=old_y1;
    }
  while(ia <= ib)
    {  
    bkg_color=wg_background[ia];
    gl_setpixel(xpix,ia,bkg_color);
    ia++;
    } 
  if(new_y1 > new_y2)
    {
    ia=new_y2;
    ib=new_y1;
    }
  else
    {
    ib=new_y2;
    ia=new_y1;
    }
  while(ia <= ib)
    {  
    gl_setpixel(xpix,ia,TXTEST_NARROW_COLOR);
    ia++;
    }
  fft1_spectrum[xpix]=new_y2;
  }
}

void show_txtest_spectrum(void)
{
int i,m,mm, new_y, xpix;
int k, wg_xpix2;
float yval,der,t1;
// fft1_spectrum is the y-value currently on screen.
// behind_fft1_spectrum is the colour behind the current data point.
// Calculate updated values and move the corresponding bright 
// pixel in case the value has changed
if(wg_xpix1 >= wg_last_xpixel)wg_xpix1=wg_first_xpixel;
i=wg.pixels_per_xpoint;
if(i==0)i=1;
k=WIDTH/(NO_REDRAW_BLOCKS*i);
k*=i;
wg_xpix2=wg_xpix1+k;
if(wg_xpix2 >  wg_last_xpixel)wg_xpix2= wg_last_xpixel;
if(  mouse_x+mouse_cursize >= wg_xpix1 &&
     mouse_x-mouse_cursize < wg_xpix2 &&
     mouse_y+mouse_cursize > wg.yborder &&
     mouse_y-mouse_cursize < wg.ybottom)hide_mouse();
if(wg.xpoints_per_pixel == 1 || wg.pixels_per_xpoint == 1)
  {
  i=wg.first_xpoint+wg_xpix1-wg_first_xpixel;
  for(xpix=wg_xpix1; xpix < wg_xpix2; xpix++)
    {
    update_txtest(xpix,(int)(wg_yfac_log*log10(fft1_slowsum[i]*wg_yfac_power)));
    i++;
    }
  }    
else
  {
  if(wg.xpoints_per_pixel == 0)
    {
// There are more pixels than data points so we must interpolate.
    i=wg.first_xpoint+(wg_xpix1-wg_first_xpixel)/wg.pixels_per_xpoint;
    new_y=wg_yfac_log*log10(fft1_slowsum[i]*wg_yfac_power);
    yval=new_y;
    if(i == wg.first_xpoint)
      {
      update_txtest(wg_xpix1,new_y);
      }
    i++;
    for(xpix=wg_xpix1; 
        xpix<wg_xpix2; xpix+=wg.pixels_per_xpoint)
      {
      t1=wg_yfac_log*log10(fft1_slowsum[i]*wg_yfac_power);
      der=(t1-yval)/wg.pixels_per_xpoint;
      m=xpix+wg.pixels_per_xpoint/2;  
      for(k=xpix+1; k<=m; k++)
        {
        yval=yval+der;
        new_y=yval;
        update_txtest(k,new_y);
        }  
      mm=xpix+wg.pixels_per_xpoint;  
      for(k=m+1; k<=mm; k++)
        {
        yval=yval+der;
        new_y=yval;
        update_txtest(k,new_y);
        }  
      i++;
      yval=t1;
      }      
    }
  else
    {
// There is more than one data point for each pixel.
// Slide a triangular filter across the spectrum to make it
// smoother before resampling.
    i=wg.first_xpoint+(wg_xpix1-wg_first_xpixel)*wg.xpoints_per_pixel;
    for(xpix=wg_xpix1; xpix<wg_xpix2; xpix++)
      {
      t1=fft1_slowsum[i]*wg.xpoints_per_pixel;
      for(k=1; k<wg.xpoints_per_pixel; k++)
        {
        t1+=(fft1_slowsum[i+k]+fft1_slowsum[i-k])*(wg.xpoints_per_pixel-k);
        }
      new_y=wg_yfac_log*log10(t1*wg_yfac_power);
      update_txtest(xpix,new_y);
      i+=wg.xpoints_per_pixel;
      }
    }  
  }
show_mouse();
wg_xpix1=wg_xpix2;
}

void update_txpeak(int xpix,int new_y2 )
{
int new_y1, old_y1, old_y2;
int ia, ib;
char bkg_color;
old_y2=txtest_ypeak[xpix];
txtest_old_ypeak[xpix]=old_y2;
if(new_y2 < 0)new_y2=0;
new_y2=wg.ybottom-1-new_y2;
if(new_y2 < wg.yborder+1)new_y2=wg.yborder+1;
if(old_y2 < new_y2)new_y2=old_y2;
if(xpix == txtest_first_xpix)
  {
  bkg_color=wg_background[old_y2];
  if(old_y2 == txtest_ypeak_decay[xpix])bkg_color=TXTEST_PEAK_DECAY_COLOR;
  gl_setpixel(xpix,old_y2,bkg_color);
  txtest_ypeak[xpix]=new_y2;
  gl_setpixel(xpix,new_y2,TXTEST_PEAK_POWER_COLOR);
  return;
  }
new_y1=txtest_ypeak[xpix-1];
old_y1=txtest_old_ypeak[xpix-1];
if( new_y2 != old_y2 || new_y1 != old_y1 || txtest_peak_redraw ==0)
  {
  if(old_y1 > old_y2)
    {
    ia=old_y2;
    ib=old_y1;
    }
  else
    {
    ib=old_y2;
    ia=old_y1;
    }
  while(ia <= ib)
    {  
    bkg_color=wg_background[ia];
    if(ia == txtest_ypeak_decay[xpix])bkg_color=TXTEST_PEAK_DECAY_COLOR;
    gl_setpixel(xpix,ia,bkg_color);
    ia++;
    } 
  if(new_y1 > new_y2)
    {
    ia=new_y2;
    ib=new_y1;
    }
  else
    {
    ib=new_y2;
    ia=new_y1;
    }
  while(ia <= ib)
    {  
    gl_setpixel(xpix,ia,TXTEST_PEAK_POWER_COLOR);
    ia++;
    }
  txtest_ypeak[xpix]=new_y2;
  }
txtest_peak_redraw++;
if(txtest_peak_redraw > PEAK_REDRAW_COUNT)
  {
  txtest_peak_redraw=0;
  }
}

void update_txpeak_decay(int xpix,int new_y2 )
{
int new_y1, old_y1, old_y2;
int ia, ib;
char bkg_color;
old_y2=txtest_ypeak_decay[xpix];
txtest_old_ypeak_decay[xpix]=old_y2;
if(new_y2 < 0)new_y2=0;
new_y2=wg.ybottom-1-new_y2;
if(new_y2 < wg.yborder+1)new_y2=wg.yborder+1;
if(xpix != txtest_first_xpix)
  {
  new_y1=txtest_ypeak_decay[xpix-1];
  old_y1=txtest_old_ypeak_decay[xpix-1];
  }
else
  {
  new_y1=new_y2;
  old_y1=new_y2;
  }
if( new_y2 != old_y2 || new_y1 != old_y1)
  {
  if(old_y1 > old_y2)
    {
    ia=old_y2;
    ib=old_y1;
    }
  else
    {
    ib=old_y2;
    ia=old_y1;
    }
  while(ia <= ib)
    {  
    bkg_color=wg_background[ia];
    if(ia == txtest_ypeak[xpix])bkg_color=TXTEST_PEAK_POWER_COLOR;
    gl_setpixel(xpix,ia,bkg_color);
    ia++;
    } 
  if(new_y1 > new_y2)
    {
    ia=new_y2;
    ib=new_y1;
    }
  else
    {
    ib=new_y2;
    ia=new_y1;
    }
  while(ia <= ib)
    {  
    gl_setpixel(xpix,ia,TXTEST_PEAK_DECAY_COLOR);
    ia++;
    }
  txtest_ypeak_decay[xpix]=new_y2;
  }
}



void update_txavg(int xpix,int new_y2 )
{
int new_y1, old_y1, old_y2;
int ia, ib;
char bkg_color;
old_y2=txtest_yavg[xpix];
txtest_old_yavg[xpix]=old_y2;
if(new_y2 < 0)new_y2=0;
new_y2=wg.ybottom-1-new_y2;
if(new_y2 < wg.yborder+1)new_y2=wg.yborder+1;
if(xpix != txtest_first_xpix)
  {
  new_y1=txtest_yavg[xpix-1];
  old_y1=txtest_old_yavg[xpix-1];
  }
else
  {
  new_y1=new_y2;
  old_y1=new_y2;
  }
if( new_y2 != old_y2 || new_y1 != old_y1)
  {
  if(old_y1 > old_y2)
    {
    ia=old_y2;
    ib=old_y1;
    }
  else
    {
    ib=old_y2;
    ia=old_y1;
    }
  while(ia <= ib)
    {  
    bkg_color=wg_background[ia];
    if(ia == txtest_ypeak_decay[xpix])bkg_color=TXTEST_PEAK_DECAY_COLOR;
    if(ia == txtest_ypeak[xpix])bkg_color=TXTEST_PEAK_POWER_COLOR;
    gl_setpixel(xpix,ia,bkg_color);
    ia++;
    } 
  if(new_y1 > new_y2)
    {
    ia=new_y2;
    ib=new_y1;
    }
  else
    {
    ib=new_y2;
    ia=new_y1;
    }
  while(ia <= ib)
    {  
    gl_setpixel(xpix,ia,TXTEST_WIDE_AVERAGE_COLOR);
    ia++;
    }
  txtest_yavg[xpix]=new_y2;
  }
}


void make_txtest_wide_spectra(void)
{
int i, j, k;
float t1;
float *z;
z=&txtest_power[txtest_spek_p0];
// Make the average spectrum.
for(i=0; i<txtest_no_of_segs; i++)
  {
  t1=0;
  k=0;
  for(j=0; j<wg.spek_avgnum; j++)
    {
    t1+=txtest_power[k+i];
    k+=txtest_no_of_segs;
    }
  txtest_powersum[i]=t1/wg.spek_avgnum;  
  }
}


void show_txtest_wide_spectra(void)
{
int j, k, ia, ib;
float yval, der, fxpix;
float yval_avg, der_avg, next_y, next_y_avg;
float yval_decay, der_decay, next_y_decay;
int m, xpix, txtest_xpix2;
float *z;
z=&txtest_power[txtest_spek_p0];
// Make the average spectrum.
txtest_last_xpix=txtest_first_xpix+(txtest_no_of_segs-1)*txtest_pixinc;  
if(txtest_xpix1 >= txtest_last_xpix)
  {
  txtest_xpix1=txtest_first_xpix;
  txtest_fxpix1=txtest_first_xpix;
  txtest_ws_ia=0;
  }
txtest_xpix2=txtest_xpix1+txtest_ka;
if(txtest_xpix2 >  txtest_last_xpix)txtest_xpix2= txtest_last_xpix;
fxpix=txtest_fxpix1;
xpix=txtest_xpix1;
if(  mouse_x+mouse_cursize >= txtest_xpix1 &&
     mouse_x-mouse_cursize < txtest_xpix2 &&
     mouse_y+mouse_cursize > wg.yborder &&
     mouse_y-mouse_cursize < wg.ybottom)hide_mouse();
// There are typically more pixels than data points so we must interpolate.
ia=txtest_ws_ia;
yval=wg_yfac_log*log10(txtest_peak_power[ia]);
yval_decay=wg_yfac_log*log10(txtest_peak_power_decay[ia]);
yval_avg=wg_yfac_log*log10(txtest_powersum[ia]);
if(ia==0)
  {
  if(txtest_decayfac1 < 0.5)
    {
    update_txpeak_decay(xpix,yval_decay);
    }
  update_txavg(xpix,yval_avg);
  update_txpeak(xpix,yval);
  }
ib=ia+txtest_ka;
while(ia < ib && xpix < txtest_xpix2)
  {
  ia++;
  m=fxpix+txtest_pixinc/2;  
  fxpix+=txtest_pixinc;
  j=fxpix;
  next_y=wg_yfac_log*log10(txtest_peak_power[ia]);
  der=(next_y-yval)/(j-xpix);
  next_y_decay=wg_yfac_log*log10(txtest_peak_power_decay[ia]);
  der_decay=(next_y_decay-yval_decay)/(j-xpix);
  next_y_avg=wg_yfac_log*log10(txtest_powersum[ia]);
  der_avg=(next_y_avg-yval_avg)/(j-xpix);
  for(k=xpix; k<m; k++)
    {
    yval=yval+der;
    yval_avg=yval_avg+der_avg;
    update_txavg(k,yval_avg);
    if(txtest_decayfac1 < 0.5)
      {
      yval_decay=yval_decay+der_decay;
      update_txpeak_decay(k,yval_decay);
      }
    update_txpeak(k,yval);
    }  
  xpix=fxpix;  
  yval=next_y-der*(xpix-m);
  yval_avg=next_y_avg-der_avg*(xpix-m);
  yval_decay=next_y_decay-der_decay*(xpix-m);
  for(k=m; k<xpix; k++)
    {
    yval=yval+der;
    yval_avg=yval_avg+der_avg;
    update_txavg(k,yval_avg);
    if(txtest_decayfac1 < 0.5)
      {
      yval_decay=yval_decay+der_decay;
      update_txpeak_decay(k,yval_decay);
      }
    update_txpeak(k,yval);
    }  
  yval=next_y;
  yval_avg=next_y_avg;
  yval_decay=next_y_decay;
  }  
txtest_ws_ia=ia;
txtest_xpix1=xpix;
txtest_fxpix1=fxpix;
show_mouse();
}



void txtest(void)
{
int i, j, k;
int ia, ib;
int old_fft1_nx, old_fft1_px;
char s[80];
int lirsta, latest_task, wideflag;
int usercontrol_counter, idle_ensure_counter;
audio_buf_info ad_info;
float t1, pwr1, pwr2;
float *z;
if(  (ui.input_mode&IQ_DATA) != 0)
  {
  ui.ad_channels=2;
  } 
else
  {
  ui.ad_channels=1;
  }
genparm[AFC_ENABLE]=0;
genparm[SECOND_FFT_ENABLE]=0;
get_buffers(0);
if(lir_status != LIR_OK)return;
open_audio_input(); 
// Open the graph windows on the screen.
// Note that they must be opened in this order because
// Each window is placed outside the previous ones and the init
// routines assumes the order below and does not check for all possible
// conflicts.
read_freq_control_data();
init_wide_graph();
if(lir_status != LIR_OK)
  {
txt_exit:;
  close_all_except_vga();
  return;
  }
for(i=0; i<txtest_no_of_segs*wg.spek_avgnum; i++)txtest_power[i]=0;
init_freq_control();
show_name_and_size();
open_mouse();
settextcolor(7);
fflush(NULL);
sync();
ensure_valid_input();
new_baseb_flag=-1;
usercontrol_counter=0;
idle_ensure_counter=0;
ampinfo_flag=1;
clear_workload_timers();
lirsta=RX_LOW_PRIO;
latest_task=0;
wideflag=0;
// *******************************************************
//                   MAIN TX TEST LOOP
// Check if we need to fetch data from the input or if we have to
// write to the output.
// If no immediate action is needed, do whatever the state 
// variable lirsta indicates.
// *******************************************************
rx_loop:
if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1079);
if(ad_info.fragments >= overrun_limit) overrun_flag++;
if(overrun_flag != 0)
  {
  close_audio_input();
  overrun_count+=overrun_flag;
  overrun_flag=0;
  sprintf(s,"OVERRUN ERRORS %d [%d] ",overrun_count,latest_task);
  wg_error(s,0);
  open_audio_input();
// Clear disk I/O in case debug messages were causing audio error
  fflush(NULL);
  sync();
  ensure_valid_input();
  }
if(ad_info.bytes >= ad_read_bytes)
  {
  newdat_flag|=ND_HWARE;
  lirsta=RX_GET_RAW_DATA;
  }
else
  {
  newdat_flag&=-1-ND_HWARE;
  }      
switch(lirsta)
  {
// ***************************   RX_GET_RAW_DATA   ****************************
// Read from the device driver.
// A failure to get here in time will cause an overrun error.
// ************************************************************************* 
  case RX_GET_RAW_DATA:
// Make sure the user has a chance now and then even on a slow computer
  usercontrol_counter++;
  if( (fullbuf_flag&ND_TIMF1) != 0)
    {
    goto fft1_first;
    }
  read_rawdat(0);
latest_task=1;
  if( (newdat_flag & ND_TIMF1) != 0)
    {
    lirsta=RX_FFT1_FIRST;
    }
  else
    {  
    lirsta=RX_FFT1_SECOND;
    }
  if(usercontrol_counter>MAX_USER)goto user_control;    
  break;

  case RX_FFT1_FIRST:
fft1_first:;
  if( (fullbuf_flag & ND_FFT1) != 0)goto fft1_second;
  fft1_b(0);
latest_task=2;
  if( (newdat_flag & ND_TIMF1) == 0)lirsta=RX_FFT1_SECOND;
  break; 

  case RX_FFT1_SECOND:
fft1_second:;
  if(fft1_na != fft1_nb)
    {
    while( fft1_na != fft1_nb)
      {
      fft1_c();
latest_task=3;
      }
    while(fft1_sumsq_pws != fft1_sumsq_pa )
      {
      update_fft1_slowsum();
latest_task=4;
      }
    }
  fft1_waterfall();
latest_task=5;
  if(wideflag == 1)
    {
    lirsta=RX_SHOW_WIDE;
    }
  else
    {  
    lirsta=RX_MIX1_BLOCK_TXTEST;
    }
  break;

  case RX_MIX1_BLOCK_TXTEST:
// ###########################################################
// Third processing block.
// evaluate peak voltage in 2.4kHz bandwidth.
// ###########################################################  
  i=(fft1_pa-fft1_px+fft1_mask+1)&fft1_mask;
  if(txtest_ka <= 1)newdat_flag &=-1-ND_FFT1;
  if( (newdat_flag & ND_FFT1) != 0 && i>2*mix1_size)
    {
    z=&txtest_power[txtest_spek_p0];
    old_fft1_nx=fft1_nx;
    old_fft1_px=fft1_px; 
    mix1_point[0]=txtest_first_point;
    for(i=0; i<txtest_no_of_segs; i++)
      {
      pwr1=pwr2;
      fft1_nx=old_fft1_nx;
      fft1_px=old_fft1_px;
      for(j=0; j<2*timf3_block; j++)
        {
        timf3_float[j]=0;     
        }
      timf3_pa=0;  
      fft1_mix1_fixed();
// In case the window is sine squared, we have to divide by it.
// because the standard routine we use does not divide for us.
      ia=(mix1_interleave_points+mix1_crossover_points)/2;
      if(mix1_interleave_points == mix1_new_points)
        {
        ia=mix1_size/4;
        ib=3*mix1_size/4;
        j=mix1_size/2+1;
        k=mix1_size/2-1; 
        while(k >= mix1_size/5)
          {
          timf3_float[2*k]*=mix1_win[k];     
          timf3_float[2*k+1]*=mix1_win[k];     
          timf3_float[2*j]*=mix1_win[k];     
          timf3_float[2*j+1]*=mix1_win[k];     
          j++;
          k--;
          }
        }
      else
        {
        if(mix1_interleave_points == 0)
          {
          ia=0;
          ib=mix1_size;
          }
        else
          {
          ia=0;
          ib=mix1_new_points+1;
          }
        }
      pwr2=0;
      pwr1=0;
      mailbox[0]=-1;if(i==49)mailbox[0]=0;
      for(j=ia; j<ib; j++)
        {
        t1=timf3_float[2*j  ]*timf3_float[2*j  ]+
           timf3_float[2*j+1]*timf3_float[2*j+1];
        pwr1+=t1;
        if(pwr2 < t1)pwr2=t1;
        }
      pwr1*=txtest_yfac;
      pwr2*=txtest_yfac;
      z[i]=pwr1/(ib-ia);
      txtest_peak_power[i]=pwr2;
      txtest_peak_power_decay[i]=
      txtest_peak_power_decay[i]*txtest_decayfac2+
      txtest_peak_power[i]*txtest_decayfac1;
      mix1_point[0]+=txtest_pntinc;
      }
    txtest_show_p0=txtest_spek_p0;  
    txtest_spek_no++;
    txtest_spek_p0+=txtest_no_of_segs;
    if(txtest_spek_no >=wg.spek_avgnum)
      {
      txtest_spek_no=0;
      txtest_spek_p0=0;
      }
latest_task=6;
    if( (newdat_flag & ND_FFT1) != 0)goto rx_loop;
    }
  if(wg_waterf_ymove_ptr > 0)
    {
    lirsta=RX_WATMO;
    }
  else
    {  
    if(wideflag == 1)
      {
      lirsta=RX_SHOW_WIDE;
      }
    else
      {  
      if(icnt >= max_ijcnt)
        {
        lirsta=RX_SHOW_FFT1;
        icnt=0;
        }
      else
        {
        icnt++;
        lirsta=RX_PREPARE_WIDE;
        }
      }
    }  
  break;

  case RX_SHOW_FFT1:
  show_txtest_spectrum();
  latest_task=7;
  if(txtest_ka > 1)
    {
    if(wg_xpix1 == wg_last_xpixel)
      {
      lirsta=RX_PREPARE_WIDE;
      }
    }
  else
    {
    lirsta=RX_WATMO;
    }           
  break;

  case RX_PREPARE_WIDE:
  if(jcnt >= max_ijcnt)
    {
    make_txtest_wide_spectra();
latest_task=81;
    lirsta=RX_SHOW_WIDE;
    jcnt=0;
    }
  else
    {
    lirsta=RX_WATMO;
    jcnt++;
    }
  break;

  case RX_SHOW_WIDE:
  wideflag=1;
  show_txtest_wide_spectra();
latest_task=82;
  if(txtest_xpix1 >= txtest_last_xpix)
    {
    lirsta=RX_WATMO;
    wideflag=0;
    }
  break;


  case RX_WATMO:
  if(wg_waterf_ymove_ptr > 0)
    { 
watmo:;
    move_wg_waterfall(); 
latest_task=9;
    }
  else
    {
    lirsta=RX_USER_CONTROL;
    }
  break;

  case RX_USER_CONTROL:
//**************************************************************
// Lower priority actions.
//**************************************************************
user_control:;
  check_mouse_actions();
latest_task=10;
  vga_inkey=vga_getkey();
  if(vga_inkey != 0)
    {
    clean_vga_inkey();
    if(numinput_flag != 0)
      {
      get_numinput_chars();
      }
    else
      {  
      if(vga_inkey == 'X') goto txt_exit;
      if(vga_inkey == 'Z')
        {
        if(txtest_saturated < 0.1)
          {
          for(i=0; i<19; i++)s[i]=' ';
          s[19]=0;
          txtest_wrerr(s);
          txtest_saturated=BIG;
          }
        clear_workload_timers();
        hide_mouse();
        for(i=wg_first_xpixel+1; i<wg_last_xpixel; i++)
          {
          ia=txtest_ypeak[i];
          ib=txtest_ypeak[i-1];
          if(ia > ib)
            {
            ib=txtest_ypeak[i  ];
            ia=txtest_ypeak[i-1];            
            }
          while(ia <= ib)
            {  
            gl_setpixel(i,ia,wg_background[ia]);
            ia++;
            } 
          }
        for(i=wg_first_xpixel; i<wg_last_xpixel; i++)
          {
          txtest_ypeak[i]=wg.ybottom-1;
          }
        show_mouse();
        clear_workload_timers();
        }
      }
    }  
  if(usercontrol_counter == 0 || lirsta == RX_USER_CONTROL)lirsta=RX_LOW_PRIO;
  usercontrol_counter=0;  
  break;

  case RX_LOW_PRIO:
  idle_ensure_counter++;
  if(idle_ensure_counter >100)
    {
    idle_ensure_counter=0;
    modesub_idle(0);
    show_adsat();
    }
  else  
    {
    if( ( newdat_flag & ND_HWARE) != 0)
      {
      lirsta=RX_GET_RAW_DATA;
      }
    else
      {
      if( (newdat_flag&ND_TIMF1) != 0)goto fft1_first;
      if(wg_waterf_ymove_ptr != 0)goto watmo; 
      }
    modesub_idle(5);
    show_adsat();
    }
  if(workload_percentage>65)
    {
    if(max_ijcnt < 16)max_ijcnt++;
    }
  else
    {
    if(workload_percentage<35)
      {
      if(max_ijcnt > 0)max_ijcnt--;
      }
    }  
  break;
  }
goto rx_loop;
}


