// Prototype for user defined hardware drive routines.
// These routines will replace the routines in hwaredriver.c
// in case a routine users_hwaredriver.c is present when configure is run.
// The Linrad package does not contain a users_hwaredriver.c file
// so it is safe to copy a nev version on top of an old one.
// The users_hwaredriver.c file will not be overwritten.

void users_eme(void)
{
// Called each time moon position is updated.
// data in degrees is available in:
// float moon_az;
// float moon_el;
}

void hware_interface_test(void)
{
// This routine is called from the main menu.
// Here you can put test routines for your hardware interface.
}


void userdefined_u(void)
{
// This routine is called when the 'U' key is pressed
}

void userdefined_q(void)
{
// This routine is called when the 'Q' key is pressed
}


void users_open_devices(void)
{
// Open ports as required
// set allow_parport=1 if the parallel port is allowed.
// We read this port now and then if allowed to in order to
// allow Linrad run without having apm shutting down the computer.
}

void set_hardware_rx_gain(int band_no)
{
// Do something with band_no to keep the compiler happy.
int dummy;
dummy=band_no;
// Input data:
// int fg.gain
// int band_no
// The argument band_no will allow different hardware on different
// bands. 
// Set hardware gain
// This is the data entered in the rx gain control box.
// This routine is responsible for checking that it is compatible to
// hardware capabilities and adjusting it accordingly.
// this routine is also responsible for sending the control information
// to the hardware. Either by setting hware_flag=1 which will
// cause calls to control_hware() until hware_flag is cleared.
// (look at hwaredriver.c to see how it works to make serial data
// on the parallel port).
// In case the hardware can be controlled by the serial port, 
// just write the control informatio to the serial port here.
//************************************************************
// This routine is responsible for setting
// int fg.gain_increment = The increment in dB for clicking
// the arrows in the gain control box.
}

void set_hardware_rx_frequency(void)
{
// Input data:
// double fg.passband_center = The midpoint of the waterfall in MHz.
// This is the data entered in the rx frequency control box.
// This routine is responsible for checking that it is compatible to
// hardware capabilities and adjusting it accordingly.
// this routine is also responsible for sending the control information
// to the hardware. Either by setting hware_flag=1 which will
// cause calls to control_hware() until hware_flag is cleared.
// (look at hwaredriver.c to see how it works to make serial data
// on the parallel port).
// In case the hardware can be controlled by the serial port, 
// just write the control informatio to the serial port here.
//************************************************************
// This routine is responsible for setting
// double fg.passband_increment = The increment in MHz for clicking
// the arrows in the rx control box.
//************************************************************
// This routine is responsible for setting
// int fg.passband_direction = -1 or +1 depending on whether the
// hardware local oscillator is above or below the desired signal. 
//************************************************************
// This routine is responsible for setting
// int fft1_direction = fg.passband_direction
}

void clear_hware_data(void)
{
// This routine can be used to stop the serial data on the
// parallel port so one does not have to wait for it to complete
// if a new frequency has been entered soon after a change.
// See hwaredriver.c
}

void control_hware(void)
{
// This routine is responsible to set
// int hware_flag=0 when it does not want to be called any more.
// If you use this routine to read data from your hardware
// to update fg.passband_center or fg.passband_direction,
// call show_center_frequency() to make Linrad update the frequency scales
// on the screen.  
}

